/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TotalMessage.
 *
 * The Initial Developer of the Original Code is
 *   alta88 <alta88@gmail.com>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2015
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var {classes: Cc, interfaces: Ci, utils: Cu} = Components;

// modules that are generic
Cu.import("resource:///modules/gloda/log4moz.js");
Cu.import("resource:///modules/gloda/mimemsg.js");
Cu.import("resource:///modules/gloda/utils.js");
Cu.import("resource:///modules/mailServices.js");
Cu.import("resource:///modules/StringBundle.js");
Cu.import("resource://gre/modules/FileUtils.jsm");

//****************************************************************************//
// Create headers and content.

var TotalMessage = {

//  Vars: {
  v: {
    // The displayed message.
    msgHdr: null,
    currentHeader: null,
    briefHeaderBuilt: false,
    basicHeaderBuilt: false,
    fullHeaderBuilt: false,

    // Indices of headerValue array.
    kWrapLines:   0,
    kEllipsis:    1,
    kFormat:      2,
    kDisplayName: 3,
    kMapped:      4,

    headerMapMailDefault: {
      // Note: 'compose' is universal, exists only in Mail map.
      // headerValue = [Wraplines, Ellipsis, Format, Display Name Override, (Mapped)]
      "brief":  {"row1":{"subject":      [0, true,  "text",      ""],
                         "from":         [0, true,  "email",     ""],
                         "date":         [0, false, "text",      ""]},
                 "row2":{}},
      "basic":  {"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"from":         [0, true,  "email",     ""],
                         "reply-to":     [0, true,  "email",     ""],
                         "date":         [0, false, "text",      ""]},
                 "row3":{"to":           [0, true,  "email",     ""],
                         "cc":           [0, true,  "email",     ""],
                         "bcc":          [0, true,  "email",     ""]},
                 "row4":{"message-id":   [0, true,  "messageid", ""],
                         "references":   [0, true,  "messageid", ""]},
                 "row5":{"organization": [0, true,  "text",      ""],
                         "user-agent":   [0, true,  "text",      ""]},
                 "row6":{"tags":         [1, true,  "text",      ""]},
                 "row7":{}},
      "compose":{"row1":{"to":           [0, true,  "",          ""]},
                 "row2":{"cc":           [0, true,  "",          ""]},
                 "row3":{"bcc":          [0, true,  "",          ""]},
                 "row4":{"reply-to":     [0, true,  "",          ""]},
                 "row5":{"followup-to":  [0, true,  "",          ""]},
                 "row6":{"newsgroups":   [0, true,  "",          ""]},
                 "row7":{}},
      "forward":{"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"to":           [0, true,  "text",      ""]},
                 "row5":{"cc":           [0, true,  "text",      ""]},
                 "row6":{}},
      "print":  {"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"to":           [0, true,  "text",      ""]},
                 "row5":{"cc":           [0, true,  "text",      ""]},
                 "row6":{}},
      "list":   {"row1":{"subject":      [0, false, "", ""]},
                 "row2":{}},
      "listmr": {"row1":{"subject":      [0, false, "", ""]},
                 "row2":{}},
    },

    headerMapNewsDefault: {
      "brief":  {"row1":{"subject":      [0, true,  "text",      ""],
                         "from":         [0, true,  "email",     ""],
                         "date":         [0, false, "text",      ""]},
                 "row2":{}},
      "basic":  {"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"from":         [0, true,  "email",     ""],
                         "reply-to":     [0, true,  "email",     ""],
                         "date":         [0, false, "text",      ""]},
                 "row3":{"to":           [0, true,  "email",     ""],
                         "cc":           [0, true,  "email",     ""],
                         "bcc":          [0, true,  "email",     ""]},
                 "row4":{"message-id":   [0, true,  "messageid", ""],
                         "in-reply-to":  [0, true,  "messageid", ""],
                         "references":   [0, true,  "messageid", ""]},
                 "row5":{"newsgroups":   [0, true,  "newsgroup", ""],
                         "followup-to":  [0, true,  "newsgroup", ""]},
                 "row6":{"tags":         [1, true,  "text",      ""]},
                 "row7":{}},
      "forward":{"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"reply-to":     [0, true,  "text",      ""]},
                 "row5":{"cc":           [0, true,  "text",      ""]},
                 "row6":{"newsgroups":   [0, true,  "text",      ""]},
                 "row7":{"references":   [0, true,  "text",      ""]},
                 "row8":{}},
      "print":  {"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"reply-to":     [0, true,  "text",      ""]},
                 "row5":{"cc":           [0, true,  "text",      ""]},
                 "row6":{"newsgroups":   [0, true,  "text",      ""]},
                 "row7":{}},
      "list":   {"row1":{"subject":      [0, false, "", ""]},
                 "row2":{}},
      "listmr": {"row1":{"subject":      [0, false, "", ""]},
                 "row2":{}},
    },

    headerMapFeedDefault: {
      "brief":  {"row1":{"subject":      [0, true,  "link",      ""],
                         "from":         [0, true,  "text",      ""],
                         "date":         [0, false, "text",      ""]},
                 "row2":{}},
      "basic":  {"row1":{"subject":      [0, true,  "link",      ""]},
                 "row2":{"from":         [0, true,  "text",      ""],
                         "date":         [0, false, "text",      ""]},
                 "row3":{"content-base": [1, true,  "link",      ""]},
                 "row4":{"message-id":   [0, true,  "messageid", ""]},
                 "row5":{"tags":         [1, true,  "text",      ""]},
                 "row6":{}},
      "forward":{"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"content-base": [0, true,  "link",      ""]},
                 "row5":{}},
      "print":  {"row1":{"subject":      [0, true,  "text",      ""]},
                 "row2":{"date":         [0, true,  "text",      ""]},
                 "row3":{"from":         [0, true,  "text",      ""]},
                 "row4":{"content-base": [0, true,  "link",      ""]},
                 "row5":{}},
      "list":   {"row1":{"subject":      [0, false, "", ""],
                         "date":         [0, false, "", "📅"]},
                 "row2":{}},
      "listmr": {"row1":{"subject":      [0, false, "", ""]},
                 "row2":{}},
    },

    // Indices of commonHeaders map. A null kColId means the header value is
    // derived from the message headers for list.
    kColId:      0,
    kLabelId:    1,
    kLabelValue: 2,

    commonHeadersDefault: new Map([
      ["subject",        ["subjectCol", "expandedsubjectLabel",""]],
      ["from",           ["senderCol", "expandedfromLabel",""]],
      ["sender",         [null, "expandedsenderLabel",""]],
      ["date",           ["dateCol", "expandeddateLabel",""]],
      ["reply-to",       [null, "expandedreply-toLabel",""]],
      ["in-reply-to",    [null, "expandedin-reply-toLabel",""]],
      ["to",             ["recipientCol", "expandedtoLabel",""]],
      ["cc",             [null, "expandedccLabel",""]],
      ["bcc",            [null, "expandedbccLabel",""]],
      ["followup-to",    [null, "expandedfollowup-toLabel",""]],
      ["newsgroups",     [null, "expandednewsgroupsLabel",""]],
      ["organization",   [null, "expandedorganizationLabel",""]],
      ["message-id",     [null, "expandedmessage-idLabel",""]],
      ["references",     [null, "expandedreferencesLabel",""]],
      ["content-base",   [null, "expandedcontent-baseLabel",""]],
      ["user-agent",     [null, "expandeduser-agentLabel",""]],
      ["tags",           ["tagsCol", "expandedtagsLabel",""]],
    ]),

    otherValuesDefault: new Map([
      ["starred",        ["flaggedCol", "flaggedCol",""]],
      ["attachments",    ["attachmentCol", "attachmentCol",""]],
      ["unread",         ["unreadButtonColHeader", "unreadButtonColHeader",""]],
      ["junk-status",    ["junkStatusCol", "junkStatusCol",""]],
      ["received-date",  ["receivedCol", "receivedCol",""]],
      ["correspondents", ["correspondentCol", "correspondentCol",""]],
      ["status",         ["statusCol", "statusCol",""]],
      ["size",           ["sizeCol", "sizeCol",""]],
      ["account",        ["accountCol", "accountCol",""]],
      ["priority",       ["priorityCol", "priorityCol",""]],
      ["location",       ["locationCol", "locationCol",""]],
      ["index",          ["idCol", "idCol",""]],
      ["x-message",      [null, null, ""]],
    ]),

    composeHeadersDefault: new Set([
      "to", "cc", "bcc", "reply-to", "followup-to", "newsgroups"
    ]),

    addressingHeaders: new Set([
      "to", "cc", "bcc", "from", "sender", "reply-to", "mail-reply-to", "mail-followup-to",
      "resent-to", "resent-cc", "resent-bcc", "resent-from", "resent-sender", "resent-reply-to",
      "approved", "disposition-notification-to", "delivered-to", "return-receipt-to",
      "followup-to", "newsgroups"
    ]),

    XUL_NS:  "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
    HTML_NS: "http://www.w3.org/1999/xhtml",
    headersUrl:            "chrome://totalmessage/content/headers.xhtml",
    globalStyleSheet:      "chrome://totalmessage/skin/tm.css",
    scrollbarStyleSheet:   "chrome://totalmessage/skin/tmscrollbar.css",
    composeStyleSheet:     "chrome://totalmessage/skin/tmcompose.css",
    stringBundleHeaders:   "chrome://totalmessage/locale/tm.properties",
    autoCollectActivePref: "extensions.totalmessage.headerAutoCollectActive",
    headerSeparatorPref:   "extensions.totalmessage.headerSeparator",
    headerWrapLinesPref:   "extensions.totalmessage.headerWrapLines",
    headerListEnabled:     "extensions.totalmessage.headerListEnabled",

    headerMapMailPref:     "extensions.totalmessage.headerMapMail2",
    headerMapNewsPref:     "extensions.totalmessage.headerMapNews2",
    headerMapFeedPref:     "extensions.totalmessage.headerMapFeed2",
    commonHeadersPref:     "extensions.totalmessage.commonHeaders",
    collectedHeadersPref:  "extensions.totalmessage.collectedHeaders",
    customHeadersPref:     "extensions.totalmessage.customHeaders",
    otherValuesPref:       "extensions.totalmessage.otherValues2",
    javascriptEnabled:     "javascript.enabled",
    allowJSPref:           "extensions.totalmessage.allow.JS",
    editingMode:           false,
    inDragStart:           false,
    activeEditorCount:     0,
    mouseDragDropMode:     false,
    keyDragDropMode:       false,
    saveShowHeaders:       null,

    HEADER: {
      0: "none",
      1: "brief",
      2: "basic",
      3: "full",
      4: "compose",
      5: "forward",
      6: "print",
      7: "list",
      8: "listmr",
      NONE:    0,
      BRIEF:   1,
      BASIC:   2,
      FULL:    3,
      TOGGLE_TYPES_COUNT: 3,
      COMPOSE: 4,
      FORWARD: 5,
      PRINT:   6,
      LIST:    7,
      LISTMR:  8,
      FORMAT_TYPES:   new Set(["text", "email", "messageid", "newsgroup", "link"]),
      MSG_TYPE_MAIL: "Mail",
      MSG_TYPE_NEWS: "News",
      MSG_TYPE_FEED: "Feed",
      MSG_TYPES:     ["Mail", "News", "Feed"]
    },

    Editor: {}
  }, // Vars

  onLoad: function() {
TM.u.log.debug("TotalMessage.onLoad: START ");
    window.removeEventListener("load", TM.onLoad, true);

    if (TM.u.winType == "Mail:Preferences")
      return;

    // Browser controller.
    window.controllers.appendController(TM.Browser.Controller);

    // Threadpane setup.
    if (TM.u.winType == "mail:3pane")
      TM.List.onLoad();
  },

  onStartHeaders: function() {
    if (TM.v.toggleRemoteContentAllow)
      return;
TM.u.log.debug("TotalMessage.onStartHeaders: START ");
    TM.resetMsgLoadFlags();
  },

  onEndHeaders: function(aWho) {
TM.u.log.debug("TotalMessage.onEndHeaders: START");
    Services.prefs.setIntPref("mail.show_headers", TM.v.saveShowHeaders);
    if (!gFolderDisplay.treeSelection || gFolderDisplay.treeSelection.count == 0 ||
        TM.v.toggleRemoteContentAllow)
      return;

TM.u.log.debug("TotalMessage.onEndHeaders: currentHeaderData "+
               currentHeaderData.toSource());
    TM.v.msgHdr = gDBView.hdrForFirstSelectedMessage;
    TM.v.currentHeader = currentHeaderData;
    TM.Headers.toggle(aWho ? aWho : "select");

    TM.v.headersBuilt = true;

    // See if there are any new headers for our collection.
    TM.u.collectHeaders(TM.v.currentHeader);
  },

  onEndAttachments: function() {
TM.u.log.debug("TotalMessage.onEndAttachments: START");
    TM.Attachments.displayAttachmentsForExpandedViewExtra();
  },

  resetHeaderFlags: function() {
TM.u.log.debug("TotalMessage.resetHeaderFlags: START ");
    TM.v.headersBuilt = false;
    TM.v.briefHeaderBuilt = false;
    TM.v.basicHeaderBuilt = false;
    TM.v.fullHeaderBuilt = false;
  },

  resetMsgLoadFlags: function() {
TM.u.log.debug("TotalMessage.resetMsgLoadFlags: START ");
    TM.v.msgHdr = null;
    TM.v.currentHeader = null;
    this.resetHeaderFlags();

    // Reset editor if coming from edit mode.
    if (TM.v.editingMode) {
      // Clear possible dragRow if changing folders; it's used to select otherwise.
      TM.v.Editor.dragRow = null;
      TM.Editor.onHeaderEditorCommand(null, "exit");
    }
  },

  createHeader: function(aType) {
TM.u.log.debug("TotalMessage.createHeader: START type "+aType);
    let messageHeader = TM.u.messageHeader;
    let headerBody = TM.u.messageHeaderBody;
    let headerDeck = TM.u.c(messageHeader, "headerDeck");

    if (TM.v[aType + "HeaderBuilt"] || aType == "none" || !TM.v.currentHeader) {
TM.u.log.debug("TotalMessage.createHeader: EXIT typeBuilt:!TM.v.curr - "+
               TM.v[aType + "HeaderBuilt"]+":"+(!TM.v.currentHeader));
      return true;
    }

    // Clear old message.
    let oldNodes = headerBody.getElementsByClassName(aType + "HeaderRow");
    while (oldNodes.length)
      oldNodes[0].remove();

    if (aType == "full")
      return TotalMessage.createHeaderFull();

    let rowsTR = this.createHeaderRows(aType, TM.u.MessageType(TM.v.msgHdr), TM.v.currentHeader);
    for (let row of rowsTR)
      headerDeck.appendChild(row);

    TM.v[aType + "HeaderBuilt"] = true;
    return true;

    // Highlight search text, if any.
//    var win = TM.u.p("messageHeader").contentWindow;
//    TotalMessageHighlight.doHighlight("basicheaders", win);

//    return true;
  },

  createHeaderRows: function(aType, aMessageType, aCurrentHeaderData) {
TM.u.log.debug("TotalMessage.createHeaderRows: START - "+aMessageType+":"+aType);
    let colIndex, headerRow1, headerRow, headerRowLabel, headerRowLabelSep,
        headerField, headerRowValue, headerRowValueDiv, headerRowValueA,
        headerTd, headerTable, headerEllipsis;
    let rowIndex = 1;
    let headerMap = TM.u.getHeaderMap(aMessageType);
    let rowsTR = [];
    let labelTag = aType == TM.v.HEADER[TM.v.HEADER.FORWARD] ||
                   aType == TM.v.HEADER[TM.v.HEADER.PRINT] ? "th" : "td";
    let sepL = "", sepR = "";
    TM.u.isRtl ? sepL = TM.u.HEADER_SEP : sepR = TM.u.HEADER_SEP;

    for (let [row, headerEntry] in Iterator(headerMap[aType])) {
      colIndex = 1;

      headerRow1 = document.createElementNS(TM.v.HTML_NS, "tr");
      headerRow1.classList.add(aType + "HeaderRow");

      for (let [headerName, headerValue] in Iterator(headerEntry)) {
TM.u.log.trace("TotalMessage.createHeaderRows: Map="+aType+", "+row+
                                            ", headerName="+headerName+
                                            ", wrap="+headerValue[TM.v.kWrapLines]+
                                            ", ellipse="+headerValue[TM.v.kEllipsis]+
                                            ", format="+headerValue[TM.v.kFormat]);
        // Headers from mime message.
        headerField = aCurrentHeaderData[headerName] ?
                        (aCurrentHeaderData[headerName].headerValue || "").trim() :
                        null;

        if (!headerField)
          // See if the header is a column/computed value.
          headerField = this.getOtherValues(null, headerName, false);

        if (headerName == "subject" && !headerField)
          // Fix raw header if blank subject.
          aCurrentHeaderData[headerName] = {headerName: "Subject"};

        if (headerName == "tags" && !headerField &&
            aCurrentHeaderData["x-mozilla-keys"])
          // Fix tags placeholder for print/compose.
          headerField = aCurrentHeaderData["x-mozilla-keys"].headerValue.trim();

        if (!headerField && headerName != "subject")
          continue;

        if (headerName == "date" && aCurrentHeaderData["x-mozilla-localizeddate"])
          // Use the localized date in brief/basic header date display.
          headerField = aCurrentHeaderData["x-mozilla-localizeddate"].headerValue;

        let count, href;
        let wrap = headerValue[TM.v.kWrapLines];
        let ellipsis = headerValue[TM.v.kEllipsis];
        let format = headerValue[TM.v.kFormat];
        let displayName = headerValue[TM.v.kDisplayName];
        let displayLabel = "";

        headerRowLabel = document.createElementNS(TM.v.HTML_NS, labelTag);
        headerRowLabel.id = aType + "Label_" + headerName;
        headerRowLabel.classList.add("headerLabel");
        if (aType != TM.v.HEADER[TM.v.HEADER.FORWARD] &&
            aType != TM.v.HEADER[TM.v.HEADER.PRINT])
          // Allow for 2nd and greater rows to align with 1st row's leaving
          // room for left side buttons.
          headerRowLabel.setAttribute("colspan", rowIndex == 1 ? "2" : "3");
        else {
          // Align and nowrap labels.
          headerRowLabel.setAttribute("align", "right");
          headerRowLabel.setAttribute("valign", "baseline");
          headerRowLabel.setAttribute("nowrap", "");
        }

        // 1)Display name from customize, 2)special override, 3)dtd name,
        // 4) name as received.  If null, user explicitly saved an empty value.
        if (displayName !== null) {
          displayLabel = displayName ||
                         this.headerDisplaySpecial(aMessageType, headerName) ||
                         TM.u.commonHeaders.get(headerName)[TM.v.kLabelValue] ||
                         aCurrentHeaderData[headerName].headerName;
        }

        // Only add the sep if a displayName has been customized.
        if (displayName === "")
          displayLabel = sepL + displayLabel + sepR;

        if (wrap)
          headerRowLabel.setAttribute("autowrap", true);

        headerRowLabel.textContent = displayLabel;
        if (colIndex == 1) {
          headerRow1.appendChild(headerRowLabel);
          headerTd = document.createElementNS(TM.v.HTML_NS, "td");
          headerTd.classList.add("headerData");
          headerTd.setAttribute("colspan", rowIndex == 1 ? "2" : "3");
          headerTable = document.createElementNS(TM.v.HTML_NS, "table");
          headerTable.style.fontSize = "inherit"
          headerTable.classList.add("messageTable");
          if (aType == TM.v.HEADER[TM.v.HEADER.FORWARD] ||
              aType == TM.v.HEADER[TM.v.HEADER.PRINT])
            headerTable.border = headerTable.cellSpacing = headerTable.cellPadding = 0;
          headerRow = document.createElementNS(TM.v.HTML_NS, "tr");
          headerRow.classList.add(aType + "HeaderRowData");

          headerTable.appendChild(headerRow);
          headerTd.appendChild(headerTable);
          headerRow1.appendChild(headerTd);
        }
        else {
          headerRow.appendChild(headerRowLabel);
        }

        headerRowValue = document.createElementNS(TM.v.HTML_NS, "td");
        headerRowValue.id = aType + "Value_" + headerName;
        headerRowValue.classList.add("headerValue");
        if (aType == TM.v.HEADER[TM.v.HEADER.FORWARD] ||
            aType == TM.v.HEADER[TM.v.HEADER.PRINT]) {
          headerRowValue.setAttribute("valign", "baseline");
          headerRowValue.style.paddingRight = "8px";
          headerRowValue.style.paddingLeft = "2px";
        }

        headerRowValueDiv = document.createElementNS(TM.v.HTML_NS, "div");
        // Height dynamically set based on headerWrapLines pref.
        headerRowValueDiv.classList.add("wrapLinesHt");
        // No tabstops.  Perhaps on action items?
        headerRowValueDiv.setAttribute("tabindex", "-1");

        if (format == "email") {
          // A header containing email addresses.
          [count, headerRowValueA, ] =
              this.OutputEmailAddresses(headerField, headerName);
          if (count > 1)
            headerRowLabel.textContent =
                "(" + count + ")" + headerRowLabel.textContent;
          headerRowValueDiv.classList.add("multiItemContainer");
        }
        else if (format == "messageid") {
          [count, headerRowValueA] =
              this.OutputMessageIds(headerName, headerField);
          if (count > 1)
            headerRowLabel.textContent =
                "(" + count + ")" + headerRowLabel.textContent;
          headerRowValueDiv.classList.add("multiItemContainer");
        }
        else if (TM.v.isNews && format == "newsgroup") {
          [count, headerRowValueA] =
              this.OutputNewsGroups(headerName, headerField);
          if (count > 1)
            headerRowLabel.textContent =
                "(" + count + ")" + headerRowLabel.textContent;
          headerRowValueDiv.classList.add("multiItemContainer");
        }
        else if (headerName == "tags") {
          let raw = aType == TM.v.HEADER[TM.v.HEADER.FORWARD] ||
                    aType == TM.v.HEADER[TM.v.HEADER.PRINT];
          [count, headerRowValueA] =
              this.OutputPrettyTags(headerName, headerField, raw);
          if (count > 1)
            headerRowLabel.textContent =
                "(" + count + ")" + headerRowLabel.textContent;
          headerRowValueDiv.classList.add("multiItemContainer");
        }
        else {
          headerRowValueDiv.classList.add("container");
          headerRowValueA = document.createElementNS(TM.v.HTML_NS, "a");
          headerRowValueA.textContent = headerField === true ? "" : headerField;

          if (format == "link") {
            href = aCurrentHeaderData[headerName].headerValue;
            if (/*TM.v.isFeed && */headerName == "subject") {
              if ("content-base" in aCurrentHeaderData) {
                href = aCurrentHeaderData["content-base"].headerValue;
                TM.u.safelySetURIAttribute(headerRowValueA, "href", href);
                headerRowValueA.classList.add("linkified");
              }
            }
            else {
              TM.u.safelySetURIAttribute(headerRowValueA, "href", href);
              headerRowValueA.classList.add("linkified");
            }
TM.u.log.trace("TotalMessage.createHeaderRows: linkify, header - "+headerName);
          }
        }

        headerRowValueDiv.appendChild(headerRowValueA);
        headerRowValue.appendChild(headerRowValueDiv);
        headerRow.appendChild(headerRowValue);

        if (!ellipsis)
          headerRowValue.classList.add("noEllipsis");

        colIndex++;
      }

      if (headerRow1.hasChildNodes())
        rowsTR.push(headerRow1);
      rowIndex++;
    }

    return rowsTR;
  },

  createHeaderFull: function() {
TM.u.log.debug("TotalMessage.createHeaderFull: START ");
    let fullHeaderTable = TM.u.c(TM.u.messageHeader, "fullHeaderTable");
    if (!fullHeaderTable || TM.v.fullHeaderBuilt)
      return;

    let headerToParse = TM.v.editingMode ? TM.v.Editor.fileHeader :
                                           TM.v.currentHeader;

TM.u.log.debug("TotalMessage.createHeaderFull: CONTINUE ");
    let headerName, headerLabel, headerField, headerRow, charSet,
        headerRowLabel, headerRowLabelSep, headerRowValue, headerRowValueDiv,
        headerRowValueA, headerEllipsis;
//    let sepL = "", sepR = "";
//    TM.u.isRtl ? sepL = TM.u.HEADER_SEP : sepR = TM.u.HEADER_SEP;

    // Iterate through message headers object and create full header.
    for (headerName in headerToParse) {
TM.u.log.trace("TotalMessage.createHeaderFull: name "+
               headerToParse[headerName].toSource());

      if (headerName == "tags")
        continue;

      headerLabel = headerToParse[headerName].headerName;
//      headerLabel = sepL + headerLabel + sepR;
      headerField = headerToParse[headerName].headerValue;
      charSet = headerToParse[headerName].charSet;
      headerRow = document.createElementNS(TM.v.HTML_NS, "tr");
      headerRow.classList.add("fullHeaderRow");
      headerRow.classList.add(headerName);

      headerRowLabel = document.createElementNS(TM.v.HTML_NS, "td");
      headerRowLabel.classList.add("headerLabel");
      headerRowLabel.textContent = headerLabel;
      headerRow.appendChild(headerRowLabel);
/**/
      headerRowLabelSep = document.createElementNS(TM.v.HTML_NS, "td");
      headerRowLabelSep.classList.add("headerSep");
      headerRowLabelSep.textContent = TM.u.HEADER_SEP;
      headerRow.appendChild(headerRowLabelSep);

      headerRowValue = document.createElementNS(TM.v.HTML_NS, "td");
      headerRowValue.classList.add("headerValue");
      if (charSet)
        headerRowValue.setAttribute("charSet", charSet);
      headerRowValueDiv = document.createElementNS(TM.v.HTML_NS, "div");
      headerRowValueDiv.classList.add("container");
      headerRowValueDiv.setAttribute("tabindex", "-1");
      if (TM.u.isStructuredAddressingHeader(headerName, headerField)) {
        let [,, fullNamesArray] = this.OutputEmailAddresses(headerField, headerName);
TM.u.log.debug("TotalMessage.createHeaderFull: fullNames - "+fullNamesArray);
        headerField = fullNamesArray.join(", ");
      }
      headerRowValueA = document.createElementNS(TM.v.HTML_NS, "a");
      headerRowValueA.textContent = headerField;

      headerRowValueDiv.appendChild(headerRowValueA);
      headerRowValue.appendChild(headerRowValueDiv);
      headerRow.appendChild(headerRowValue);
      fullHeaderTable.appendChild(headerRow);
    }

    TM.v.fullHeaderBuilt = true;

    // Highlight search text, if any.
//    var win = TM.u.p("messageHeader").contentWindow;
//    TotalMessageHighlight.doHighlight("headers", win);
TM.u.log.debug("TotalMessage.createHeaderFull: done");
  },

  /**
   * Create an html tbody containing header rows, for map types forward or print
   * in compose or print/print preview.
   * 
   * @param   object aCurrentHeader - headers in currentHeaderData format.
   * @param   string aMessageType   - message type, in MSG_TYPES.
   * @param   string aType          - map type.
   * @param   bool aIsHtml          - true if html result is wanted.
   * @returns html <tbody> node.
   */
  createHeaderDraft: function(aCurrentHeader, aMessageType, aType, aIsHtml) {
TM.u.log.debug("TM.createHeaderDraft: START msgType:type:html? - "+
                aMessageType+":"+aType+":"+aIsHtml);
    // Get the mapped rows, there may be none.
    let rowsTR = this.createHeaderRows(aType, aMessageType, aCurrentHeader);
    let tableBody = aIsHtml && rowsTR.length ?
                      document.createElementNS(TM.v.HTML_NS, "tbody") : "";
    // Iterate through message headers object and create custom header.
    for (let row of rowsTR) {
      if (aIsHtml)
        tableBody.appendChild(row);
      else {
        tableBody += row.innerHTML.replace(/<\/th>/g, " <\/th>")
                                  .replace(/<\/td>/g, "<\/td>    ")
                                  .replace(/<[^>]+>/g, "");
        tableBody = tableBody.trimRight() + "<br>";
      }
    }

TM.u.log.debug("TotalMessage.createHeaderDraft: done");
    TM.v[aType + "HeaderBuilt"] = true;
    return tableBody;
  },

  /**
   * Create an html td containing a linkified list of email addresses.
   * 
   * @param   string emailAddresses  - comma separated string.
   * @param   string headerName      - headerName.
   * @returns [count of emailAddresses, html <div> node, decoded fullnames array]
   */
  OutputEmailAddresses: function (emailAddresses, headerName) {
    let index = 0;
    let headerRowValueContainer, headerRowValueItemA, headerRowValueItemCard;

    headerRowValueContainer = document.createElementNS(TM.v.HTML_NS, "div");
    headerRowValueContainer.classList.add("emailAddressContainer");
    if (!emailAddresses)
      return [index, headerRowValueContainer];

    let addresses = {};
    let fullNames = {};
    let names = {};
    let numAddresses = 0;
    let fullNamesArray = [];

    numAddresses =
        MailServices.headerParser.parseHeadersWithArray(emailAddresses,
                                                        addresses,
                                                        names,
                                                        fullNames);
    while (index < numAddresses) {
      let address = {};
      address.emailAddress = addresses.value[index];
      address.fullAddress = fullNames.value[index];
      address.displayName = names.value[index];

      headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
      headerRowValueItemA.classList.add("emailAddress");
      headerRowValueItemA.classList.add(headerName);
      headerRowValueItemA.classList.add("multiItem");
      TM.u.topWindow.updateEmailAddressNode(headerRowValueItemA, address);
      // Don't leave <> due to jsmime non handling of empty addr-spec.
      headerRowValueItemA.textContent = fullNames.value[index].replace(/\s<>$/, "");
      fullNamesArray.push(fullNames.value[index].replace(/\s<>$/, ""));

      headerRowValueItemCard = document.createElementNS(TM.v.HTML_NS, "a");
      headerRowValueItemCard.classList.add("abCard");
      headerRowValueItemCard.title = headerRowValueItemA.getAttribute("tooltipstar");

      headerRowValueContainer.appendChild(headerRowValueItemA);
      headerRowValueContainer.appendChild(headerRowValueItemCard);
      index++;

      if (index < numAddresses) {
        // Add the separator.
        headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
        headerRowValueItemA.classList.add("multiItemSeparator");
        headerRowValueItemA.textContent = ", ";
        headerRowValueContainer.appendChild(headerRowValueItemA);
      }
    }

    return [index, headerRowValueContainer, fullNamesArray];
  },

  /**
   * Create an html td containing a linkified list of message ids.
   * 
   * @param   string headerName   - headerName.
   * @param   string headerValue  - comma separated string of message ids.
   * @returns [count of messageIds, html <div> node]
   */
  OutputMessageIds: function (headerName, headerValue) {
TM.u.log.trace("TotalMessage.OutputMessageIds: linkify news "+ headerValue);
    let messageIdArray, id, browserUrl, href;
    let count = 0;
    let headerRowValueContainer, headerRowValueItemA, headerRowValueItemGo;

    headerRowValueContainer = document.createElementNS(TM.v.HTML_NS, "div");
    headerRowValueContainer.classList.add("messageIdsContainer");
    // No tabstops.
    headerRowValueContainer.setAttribute("tabindex", "-1");

    if (!headerValue)
      return [count, headerRowValueContainer];

    // Need to split on '> ' as a message id can have a space within <>, and
    // remove the last non value item.
    messageIdArray = headerValue.split(/>\s*/);
    messageIdArray.pop();
    browserUrl = Services.prefs.getComplexValue("mailnews.messageid_browser.url",
                                                Ci.nsIPrefLocalizedString).data;

    for (let messageId of messageIdArray) {
      count++;
      headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
      headerRowValueItemA.classList.add(headerName);
      headerRowValueItemA.classList.add("multiItem");
      headerRowValueItemA.classList.add("messageIdDisplayButton");
      headerRowValueItemA.textContent = messageId + ">";
      // Remove < (the > has been removed in the split).
      id = messageId.substring(1, messageId.length);
      headerRowValueItemA.setAttribute("messageId", id);

      headerRowValueContainer.appendChild(headerRowValueItemA);

      if (TM.v.isNews) {
        // Add a go to indicator.
        headerRowValueItemGo = document.createElementNS(TM.v.HTML_NS, "a");
        headerRowValueItemGo.classList.add("goToMessage");
        headerRowValueItemGo.title = TM.u.strings.get("goToMessage");
        href = browserUrl.replace(/%mid/, id);
        TM.u.safelySetURIAttribute(headerRowValueItemGo, "href", href);

        headerRowValueContainer.appendChild(headerRowValueItemGo);
      }

      if (count < messageIdArray.length) {
        // Add the separator.
        headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
        headerRowValueItemA.classList.add("multiItemSeparator");
        headerRowValueItemA.textContent = ", ";

        headerRowValueContainer.appendChild(headerRowValueItemA);
      }
    }

    return [count, headerRowValueContainer];
  },

  /**
   * Create an html td containing a linkified list of newsgroups.
   * 
   * @param   string headerName   - headerName.
   * @param   string headerValue  - comma separated string of newsgroups.
   * @returns [count of newsgroups, html <div> node]
   */
  OutputNewsGroups: function (headerName, headerValue) {
    let newsgroupsArray;
    let count = 0;
    let headerRowValueContainer, headerRowValueItemA, headerRowValueItemCard;

    headerRowValueContainer = document.createElementNS(TM.v.HTML_NS, "div");
    headerRowValueContainer.classList.add("multiItemSeparator");
    if (!headerValue)
      return [count, headerRowValueContainer];

    newsgroupsArray = headerValue.split(",");

    for (let newsgroup of newsgroupsArray) {
      count++;
      headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
      headerRowValueItemA.classList.add("newsgroup");
      headerRowValueItemA.classList.add(headerName);
      headerRowValueItemA.classList.add("multiItem");
      headerRowValueItemA.setAttribute("newsgroup", newsgroup);
      headerRowValueItemA.textContent = newsgroup;

      headerRowValueContainer.appendChild(headerRowValueItemA);

      if (count < newsgroupsArray.length) {
        // Add the separator.
        headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, "a");
        headerRowValueItemA.classList.add("newsgroupSeparator");
        headerRowValueItemA.textContent = ", ";

        headerRowValueContainer.appendChild(headerRowValueItemA);
      }
    }

    return [count, headerRowValueContainer];
  },

  /**
   * Create an html td containing a pretty list of tags.
   * 
   * @param   string headerName   - headerName.
   * @param   string headerValue  - comma separated string of tags.
   * @param   bool raw            - true if print/forward, for display.
   * @returns [count of tags, html <div> node]
   */
  OutputPrettyTags: function (headerName, headerValue, raw) {
    let tagsArray, tagName, tagColor;
    let count = 0;
    let headerRowValueContainer, headerRowValueItemA, headerRowValueItemCard;
    let htmlTag = raw ? "a" : "button";

    headerRowValueContainer = document.createElementNS(TM.v.HTML_NS, "div");
    headerRowValueContainer.classList.add("multiItemSeparator");
    if (!headerValue)
      return [count, headerRowValueContainer];

    tagsArray = headerValue.split(/\s+/);

    for (let tagKey of tagsArray) {
      try {
        // If we got a bad tag name, getTagForKey will throw an exception,
        // skip it and go to the next one.
        tagName = MailServices.tags.getTagForKey(tagKey);
      }
      catch (ex) {}

      if (!tagName)
        continue;

      count++;
      tagColor = MailServices.tags.getColorForKey(tagKey);
      headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, htmlTag);
      headerRowValueItemA.classList.add(headerName);
      headerRowValueItemA.classList.add("multiItem");
      headerRowValueItemA.setAttribute("label", tagName);
      headerRowValueItemA.setAttribute("tabindex", "-1");
//      headerRowValueItemA.dataset.tagName = tagName;
      if (tagColor)
        headerRowValueItemA.style.color = "#" + tagColor.substr(1);

      headerRowValueItemA.appendChild(document.createTextNode(tagName));
      headerRowValueContainer.appendChild(headerRowValueItemA);

      if (raw && count < tagsArray.length) {
        headerRowValueItemA = document.createElementNS(TM.v.HTML_NS, htmlTag);
        headerRowValueItemA.appendChild(document.createTextNode(", "));
        headerRowValueContainer.appendChild(headerRowValueItemA);
      }
    }

    return [count, headerRowValueContainer];
  },

  headerDisplaySpecial: function(aMessageType, aHeaderName) {
    if (aMessageType == TM.v.HEADER.MSG_TYPE_FEED && aHeaderName == "from")
      return TM.u.getString("h_feed_" + aHeaderName);
    if (aHeaderName == "x-message")
      return TM.u.getString("h_" + aHeaderName);
    if (TM.u.otherValues.has(aHeaderName))
      return TM.u.otherValues.get(aHeaderName)[TM.v.kLabelValue];

    return null;
  },

  getOtherValues: function(aRow, aHeaderName, aBools) {
    let dbView = TM.u.mainWindow ? TM.u.mainWindow.gDBView : null;
    let threadTree = TM.u.mainWindow ? TM.u.mainWindow.GetThreadTree() : null;
    let colId = TM.u.otherValues.has(aHeaderName) ?
                  TM.u.otherValues.get(aHeaderName)[TM.v.kColId] : null;
    let row = aRow != null ? aRow : dbView ? dbView.currentlyDisplayedMessage : null;
    if (row == null || row == -1 || row == TM.Editor.nsMsgKey_None)
      return null;

    if (aHeaderName == "x-message")
      return TM.List._fetch(row, aHeaderName, TM.v.msgHdr);

    let col = threadTree ? threadTree.columns[colId] : null;
    if (!col)
      return null;

    let val = gDBView.getCellText(row, col);
//TM.u.log.trace("TM.getOtherValues: aHeaderName:colId:CellText - "+
//               aHeaderName+":"+col.id+":"+val);
    if (val || TM.List.columnsCustomMap.has(col.id)) {
      if (aBools && col.id == "junkStatusCol")
        return val && val != "0";
//TM.u.log.trace("TM.getOtherValues: aHeaderName:colId:CellProp - "+
//               aHeaderName+":"+col.id+":"+gDBView.getCellProperties(row, col));
      return val;
    }
    val = gDBView.getCellValue(row, col);
//TM.u.log.trace("TM.getOtherValues: aHeaderName:colId:CellValue - "+
//               aHeaderName+":"+col.id+":"+val);
      if (aBools || val == col.id)
        return val && val != col.id;
      return val;
  },


  //**************************************************************************//
  //  List Headers functions (threadpane).

  List: {

    columnsMap: new Map([
      ["dateCol",       ["byDate"]],
      ["receivedCol",   ["byReceived"]],
      ["senderCol",     ["byAuthor"]],
      ["recipientCol",  ["byRecipient"]],
      ["correspondentCol", ["byCorrespondent"]],
      ["subjectCol",    ["bySubject"]],
      ["locationCol",   ["byLocation"]],
      ["accountCol",    ["byAccount",]],
      ["unreadButtonColHeader", ["byUnread"]],
      ["statusCol",     ["byStatus"]],
      ["sizeCol",       ["bySize"]],
      ["priorityCol",   ["byPriority"]],
      ["flaggedCol",    ["byFlagged"]],
      ["threadCol",     ["byThread"]],
      ["tagsCol",       ["byTags"]],
      ["junkStatusCol", ["byJunkStatus"]],
      ["idCol",         ["byId"]],
      ["attachmentCol", ["byAttachments"]],
    ]),

    // Columns excluded from handling.
    columnsExMap: new Map([
      ["totalCol",  ["byNone"]],
      ["unreadCol", ["byUnread"]],
      ["idCol",     ["byId"]],
    ]),

    kSortType: 0,
    kCustColHandler: 1,
    columnsCustomMap: new Map(),

    Sorting: false,
    PaintAfterSortDelay: 40,
    _cache: {},
    _initialized: false,

    onLoad: function() {
      if (!TM.u.HEADER_LIST_ENABLED)
        return;
TM.u.log.debug("TM.List.onLoad: START ");
      let tmSplitter = TM.u.d("subjectCol").nextElementSibling;
      tmSplitter.id = "tmBar";
      tmSplitter.appendChild(TM.u.d("tmBarObserves"));
      tmSplitter.appendChild(TM.u.d("tmBox"));
      TM.u.d("tmBox").removeAttribute("hidden");
      TM.u.d("threadCols").setAttribute("headerlistenabled", true);
      GetThreadTree().setAttribute("hidecolumnpicker", true);

      Services.obs.addObserver(TM.List.Observer, "MsgCreateDBView", false);
//      TM.u.d("tabmail").registerTabMonitor(TM.List.TabMonitor);
//      TM.List.threadTreeChildren.addEventListener("mousemove", TM.List.onMouseMove, false);
//      TM.List.threadTreeChildren.addEventListener("mouseleave", TM.List.onMouseLeave, false);

TM.u.log.debug("TM.List.onLoad: FINISHED  ");
    },

    // Observer for custom column.
    Observer: {
      observe: function(aMsgFolder, aTopic, aData) {
TM.u.log.debug("TM.List.observe: START aTopic - "+aTopic);

        setTimeout(function() {
          if (TM.List._initialized) {
TM.u.log.debug("TM.List.observe: post init aTopic - "+aTopic);
            TM.List.setColumnHandlers();
            TM.List.resetView(aMsgFolder);
            return;
          }

TM.u.log.debug("TM.List.observe: init aTopic - "+aTopic);
          TM.List.setCustomColumns();
          TM.List._initialized = true;
          TM.List.setColumnHandlers();
          // Init.
          UpdateSortIndicators(gDBView.sortType, gDBView.sortOrder);
          // Update.
          UpdateSortIndicators(gDBView.sortType, gDBView.sortOrder);

          TM.List.setMutationObserver();
          TM.List.setTotalMessageBar();
          TM.List.resetView(aMsgFolder);

        }, 20);
      }
    },

    TabMonitor: {
      monitorName: "TotalMessage",
      onTabTitleChanged: function() {},
      onTabSwitched: function(aTab, aOldTab) {
       if ("folderDisplay" in aTab) {
TM.u.log.debug("TM.List.onTabSwitched: title - "+aTab.title);
         TM.List.resetView(aTab.folderDisplay.displayedFolder);
       }
      }
    },

    TreeView: {
      isEditable:          function(aRow, aCol) { return false; },
      isSelectable:        function(aRow, aCol) {/*not implemented in core*/},
      isString:            function()           { return false; },
      getSortStringForRow: function(aMsgHdr)    { return ""; },
      getSortLongForRow:   function(aMsgHdr)    { return 0; },
      getImageSrc:         function(aRow, aCol) { return null; },
      cycleCell:           function(aRow, aCol) { return; },
      getRowProperties:    function(aRow)       {
        if (TM.List.isGroupedBySortHdr(aRow))
          return "dummy";
        return "";
      },
      getCellProperties:   function(aRow, aCol) {
        if (TM.List.isGroupedBySortHdr(aRow)) {
          return "dummy";
        }
        return "";
      },
      getCellText:         function(aRow, aCol) {
        if (TM.List.isGroupedBySortHdr(aRow))
          return TM.List.getGroupedBySortHeader(aRow, aCol.id,
                                                gDBView.getMsgHdrAt(aRow),
                                                null, null);
        return TM.List.populateRow(aRow, 0, gDBView.getMsgHdrAt(aRow));
      },
    }, // TreeView

    resetView: function(aMsgFolder) {
TM.u.log.debug("TM.List.resetView: aMsgFolder - "+
               (aMsgFolder && aMsgFolder.name ? aMsgFolder.name : null));
      TM.List.Sorting = false;
      GetThreadTree().treeBoxObject.invalidate();
    },
  
    setMutationObserver: function() {
TM.u.log.debug("TM.List.setMutationObserver: START ");
      let mo = new MutationObserver(m => TM.List.onMutations(m));
      mo.observe(TM.u.d("subjectCol"),
                 { attributes: true, attributeFilter: ["width"] });
      mo.observe(TM.u.d("tmBar"),
                 { attributes: true, attributeFilter: ["ordinal"] });
      mo.observe(TM.u.d("threadCols"),
                 { attributes: true, childList: true, subtree: true,
                   attributeFilter: ["hidden","dragging"] });
      mo.observe(TM.u.d("folderPaneBox"),
                 { attributes: true, attributeFilter: ["width"] });
      mo.observe(TM.u.d("threadContentArea"),
                 { attributes: true, attributeFilter: ["width"] });
      mo.observe(TM.u.d("messengerWindow"),
                 { attributes: true, attributeFilter: ["width"] });
      mo.observe(TM.u.d("mailContent"),
                 { attributes: true, attributeFilter: ["layout"] });
      TM.List.MutationObserver = mo;
    },

    onMutations: function(aMutations) {
//      aMutations.forEach(function(mutation) {
      for (let mutation of aMutations) {
        if (mutation.attributeName == "width" ||
            mutation.attributeName == "layout" ||
            (mutation.target.localName == "treecol" &&
             mutation.attributeName == "hidden")) {
TM.u.log.debug("TM.List.onMutations: mutation.id:attributeName - "+
               mutation.target.id+":"+mutation.attributeName);
//          setTimeout(function() {
            TM.List.setTotalMessageBar();
//          }, 0);
        }
        if (mutation.target.localName == "treecol" &&
            mutation.attributeName == "dragging") {
TM.u.log.debug("TM.List.onMutations: dragging id:attributeName:oldValue - "+
               mutation.target.id+":"+mutation.attributeName+":"+mutation.oldValue);
          let buttons = TM.u.d("tmScrollBox").querySelectorAll("toolbarbutton");
          for (let button of buttons) {
            if (mutation.target.hasAttribute("dragging"))
              button.setAttribute("disabled", true);
            else
              button.removeAttribute("disabled");
          }
        }
        if (mutation.attributeName == "ordinal") {
TM.u.log.debug("TM.List.onMutations: ordinal mutation.id:attributeName - "+
               mutation.target.id+":"+mutation.attributeName);
//      setTimeout(function() {
          let subjectOrdinal = parseInt(TM.u.d("subjectCol").ordinal);
          let tmBarOrdinal = parseInt(mutation.target.ordinal);
TM.u.log.debug("TM.List.onMutations: ordinal tmBarOrdinal:subjectOrdinal - "+
               tmBarOrdinal+":"+subjectOrdinal);
          if (tmBarOrdinal - 1 != subjectOrdinal)
            mutation.target.ordinal = subjectOrdinal + 1;
//      }, 0);
        }
      };
    },
  
    setTotalMessageBar: function(aReset) {
      let s = TM.u.d("tmScrollBox");
      s.width = TM.u.d("subjectCol").boxObject.width;
      s.setAttribute("style", "margin-left: -" + s.width + "px;");
TM.u.log.debug("TM.List.setTotalMessageBar: set scrollbox width to - "+ s.width);
    },

    setColumnHandlers: function() {
TM.u.log.debug("TM.List.setColumnHandlers: START ");
      gDBView.addColumnHandler("subjectCol", TM.List.TreeView);
    },
  
    setCustomColumns: function(aReset) {
TM.u.log.debug("TM.List.setCustomColumns: START ");
      let popup = TM.u.d("selectColumnsPopupTM"), mi;
      let cols = this.getSortedCols();
      cols.forEach(col => {
//TM.u.log.debug("TM.List.setCustomColumns: column - "+col.id);
        if (!col.element.hasAttribute("ignoreincolumnpicker")) {
          mi = document.createElement("menuitem");
          mi.setAttribute("type", "checkbox");
          let columnName = col.element.getAttribute("display") ||
                           col.element.getAttribute("label");
          mi.setAttribute("label", columnName);
          mi.setAttribute("colindex", col.index);
          mi.setAttribute("colid", col.id);
          mi.setAttribute("closemenu", "none");
          if (col.element.getAttribute("hidden") != "true")
            mi.setAttribute("checked", "true");
          if (col.primary)
            mi.setAttribute("disabled", "true");
          popup.appendChild(mi);
        }

        if (TM.List.columnsMap.has(col.id) || TM.List.columnsExMap.has(col.id) ||
            TM.List.columnsCustomMap.has(col.id))
          return;

        let custColHandler;
        try { custColHandler = gDBView.getColumnHandler(col.id); }
        catch (ex) { return; }

TM.u.log.debug("TM.List.setCustomColumns: add to columnsCustomMap - "+col.id);
        TM.List.columnsCustomMap.set(col.id, ["byCustom", custColHandler]);
        mi.setAttribute("iscustcol", true);
      });
TM.u.log.debug("TM.List.setCustomColumns: columnsCustomMap - "+JSON.stringify([...TM.List.columnsCustomMap]));
      TM.u.otherValues;
    },

    selectColumns: function(aEvent) {
      let col = TM.u.d(aEvent.target.getAttribute("colid"));
      col.setAttribute("hidden", !aEvent.target.hasAttribute("checked"));
    },

    selectColumnsPopup: function(aPopup) {
TM.u.log.debug("TM.List.selectColumnsPopup: aPopup - "+aPopup.id);
      for (let mi of aPopup.children) {
        if (TM.u.d(mi.getAttribute("colid")).getAttribute("hidden") == "true")
          mi.removeAttribute("checked");
        else
          mi.setAttribute("checked", true);
      }
    },

    resetColumns: function(aEvent) {
      gFolderDisplay.setColumnStates(gFolderDisplay._getDefaultColumnsForCurrentFolder(), true);
    },

    applyColumns: function(aEvent, aApplyToSubfolders) {
      let destFolder = aEvent.originalTarget._folder;
//      let bundle = document.getElementById("bundle_messenger");
//      let stringBase = "threadPane.columnPicker.confirmFolder." +
//                       (aApplyToSubfolders ? "withChildren." : "noChildren.");
//      let confirmed = Services.prompt.confirm(
//                        null,
//                        bundle.getString(stringBase + "title"),
//                        bundle.getFormattedString(stringBase + "message",
//                                                  [destFolder.prettiestName]));
//      if (!confirmed)
//        return;

      // Get the current folder's column state.
      let propName = gFolderDisplay.PERSISTED_COLUMN_PROPERTY_NAME;
      let dbFolderInfo = gFolderDisplay.displayedFolder.msgDatabase.dBFolderInfo;
      let columnStateString = dbFolderInfo.getCharProperty(propName);
      // Now propagate appropriately...
      if (aApplyToSubfolders) {
        // Generate an observer notification when we have finished configuring
        // all folders.  This is currently done for the benefit of our mozmill
        // tests.
        let observerCallback = function() {
          Services.obs.notifyObservers(gFolderDisplay.displayedFolder,
                                       "msg-folder-columns-propagated", "");
        }
        MailUtils.setStringPropertyOnFolderAndDescendents(propName,
                                                          columnStateString,
                                                          destFolder,
                                                          observerCallback);
      }
      else {
        destFolder.setStringProperty(propName, columnStateString);
        destFolder.msgDatabase = null;
      }
    },

    checkKeyPress: function(aEvent) {
      aEvent.stopPropagation();
      let node = aEvent.target;
      let key = aEvent.which ? aEvent.which : aEvent.keyCode;
      let dir = window.getComputedStyle(node, null).direction;
      let pass = true;
TM.u.log.debug("TM.List.checkKeyPress: node.id:key - "+aEvent.target.id+":"+key);

      if (!key)
        return true;

      if ((key == KeyEvent.DOM_VK_LEFT && dir == "ltr") ||
          (key == KeyEvent.DOM_VK_RIGHT && dir == "rtl")) {
        if (node == node.parentNode.firstChild)
          aEvent.stopImmediatePropagation();
        else if (node.classList.contains("tmSortMenuList"))
          window.document.commandDispatcher.rewindFocus();
      }

      if ((key == KeyEvent.DOM_VK_RIGHT && dir == "ltr") ||
          (key == KeyEvent.DOM_VK_LEFT && dir == "rtl")) {
        if (node == node.parentNode.lastChild || node.nextSibling.disabled)
          aEvent.stopImmediatePropagation();
        else if (!node.classList.contains("XtmSortMenuList"))
          window.document.commandDispatcher.advanceFocus();
      }

      if (key == aEvent.DOM_VK_SPACE || key == aEvent.DOM_VK_RETURN ||
          (aEvent.charCode && !node.classList.contains("tabbable"))) {
        aEvent.stopImmediatePropagation();
        aEvent.preventDefault();
      }
      if (//("button" in aEvent && aEvent.button != 0) ||
          ("keyCode" in aEvent &&
           key != aEvent.DOM_VK_SPACE && key != aEvent.DOM_VK_RETURN))
        pass = false;

      let popup = node.popup || node.menupopup;
      if (popup) {
        if (pass)
          TM.u.openPopup(aEvent);
        else
          popup.hidePopup();
      }

      return pass;
    },

    // Column header click, change primary sortType or sortOrder.
    handleColumnClick: function(aColumnID) {
      let sortType = TM.List.getSortTypeFromColId(aColumnID);
TM.u.log.debug("TM.List.handleColumnClick: START columnId:sortType:curCustCol - "+
               aColumnID+":"+sortType+":"+gDBView.curCustomColumn);
      if (!sortType) {
        TM.u.log.warn("TM.List.handleColumnClick: unsupported sort column: "+
                      aColumnID + " - no custom handler installed.");
        return;
      }

      if (TM.u.d(aColumnID).hasAttribute("nosortby"))
        return;

      if (sortType != "byCustom")
        gDBView.curCustomColumn = null;

      TM.List.setPaintAfterSortDelay();

      if (sortType == "byThread") {
        TM.List.threadBy(null);
        return;
      }

      if ((sortType != "byCustom" && gDBView.sortType == nsMsgViewSortType[sortType]) ||
          (sortType == "byCustom" && gDBView.curCustomColumn == aColumnID))
        TM.List.doCommand(null, "orderBy");
      else {
        TM.u.d("tmSortByMenu").setAttribute("curselected", aColumnID);
        TM.List.doCommand(null, "sortByType");
      }
    },

    // Menulist oncommand.
    doSortCommand: function(aEvent, aCommand) {
      let checked = aEvent.target.hasAttribute("checked");
TM.u.log.debug("TM.List.doSortCommand: id:command:checked - "+
               aEvent.target.id+":"+aCommand+":"+checked);
      if (!this.checkKeyPress(aEvent) || !(aCommand in this))
        return;

      let selId, prevselId, control;
      switch (aCommand) {
        case "threadBy":
          break;
        case "sortBy":
        case "groupBy":
        case "sortByType":
          control = aEvent.target.parentNode.anchorNode;
          selId = aEvent.target.getAttribute("itemid");
          prevselId = control.getAttribute("curselected");
TM.u.log.debug("TM.List.doSortCommand: id:selId:prevselId - "+control.id+":"+selId+":"+prevselId);
          if (selId == prevselId)
            return;
          control.setAttribute("label", aEvent.target.label);
          control.setAttribute("curselected", selId);
          break;
        case "orderBy":
          break;
        default:
          return;
      }

      this.doCommand(aEvent, aCommand);
    },

    doCommand: function(aEvent, aCommand) {
      TM.List.setPaintAfterSortDelay(aCommand);

      this[aCommand](aEvent);

      TM.List.setTotalMessageBar();

      if (TM.List.Sorting)
        setTimeout(function() {
          TM.List.Sorting = false;
          GetThreadTree().treeBoxObject.invalidate();
        }, TM.List.PaintAfterSortDelay);
    },

    tmButtonPopup: function(aPopup) {
      TM.u.d("enableHeaderEditModeTM").disabled = !TM.Headers.isEditable();
    },

    threadBy: function(aEvent) {
//      if (TM.List.simpleColumns())
        MsgToggleThreaded();
//      else if (gFolderDisplay.view.showThreaded)
//        MsgReverseSortThreadPane();
//      else
//        MsgSortByThread();
    },

    threadPopup: function(aPopup) {
TM.u.log.debug("TM.List.threadPopup:   ");
      view_init();
      InitViewSortByMenu();
      InitViewMessagesMenu();

      let reIsFalse = false;
      for (let mi of aPopup.querySelectorAll("[preference]")) {
        let pref = mi.getAttribute("preference");
        let val = Services.prefs.getBoolPref(pref);
        if (pref == "mail.strict_threading" && val)
          reIsFalse = true;
        if (pref != "mail.thread_without_re" || !reIsFalse)
          mi.setAttribute("checked", val);
        else {
          mi.removeAttribute("checked");
          Services.prefs.setBoolPref("mail.thread_without_re", false);
        }
      }
    },

    setThreadingPref: function(aMenuitem) {
TM.u.log.debug("TM.List.setThreadingPref:   ");
      let pref = aMenuitem.getAttribute("preference");
      let checked = aMenuitem.hasAttribute("checked");
      Services.prefs.setBoolPref(pref, checked);

      if (pref == "mail.strict_threading" && checked) {
        TM.u.d("thread_without_reMenuItemTM").removeAttribute("checked");
        Services.prefs.setBoolPref(TM.u.d("thread_without_reMenuItemTM")
                                     .getAttribute("preference"), false);
      }
      if (pref == "mail.thread_without_re" && checked) {
        TM.u.d("strict_threadingMenuItemTM").removeAttribute("checked");
        Services.prefs.setBoolPref(TM.u.d("strict_threadingMenuItemTM")
                                     .getAttribute("preference"), false);
      }
      if (pref == "mailnews.sort_threads_by_root") {
        gFolderDisplay.view.refresh();
      }
      if (pref == "mail.operate_on_msgs_in_collapsed_threads" &&
          gFolderDisplay.treeSelection && gFolderDisplay.treeSelection.count == 1) {
        gFolderDisplay.treeSelection.toggleSelect(gFolderDisplay.treeSelection.currentIndex);
        gFolderDisplay.treeSelection.toggleSelect(gFolderDisplay.treeSelection.currentIndex);
      }
    },

    resetAllThreadingPrefs: function(aMenupopup) {
TM.u.log.debug("TM.List.resetAllThreadingPrefs:   ");
      for (let mi of aMenupopup.querySelectorAll("[preference]")) {
        Services.prefs.clearUserPref(mi.getAttribute("preference"));
        mi.setAttribute("checked", Services.prefs.getBoolPref(mi.getAttribute("preference")));
      }
    },

    allThreads:          function(aExpand) {
TM.u.log.debug("TM.List.allThreads: action - "+(aExpand ? "expand" : "collapse"));
      if (aExpand)
        gFolderDisplay.doCommand(nsMsgViewCommandType.expandAll);
      else {
        gFolderDisplay.selectSelectedThreadRoots();
        gFolderDisplay.doCommand(nsMsgViewCommandType.collapseAll);
      }
      gFolderDisplay.ensureSelectionIsVisible();
//      GetThreadTree().treeBoxObject.invalidate();
    },

    simpleColumns: function() {
      let simpleColumns = false;
      try {
        simpleColumns = Services.prefs.getBoolPref("mailnews.thread_pane_column_unthreads");
      }
      catch (ex) {}

      return simpleColumns;
    },

    setPaintAfterSortDelay: function(aCommand) {
      TM.List.Sorting = gFolderDisplay.view.showGroupedBySort &&
                        aCommand != "orderBy"; // ? true : false;
    },

    sortBy: function(aEvent) {
TM.u.log.debug("TM.List.sortBy: ");
      let sortName = GetThreadTree().getAttribute("sortName");
      let sortDir = GetThreadTree().getAttribute("sortDirection");
      let simpleColumns = this.simpleColumns();
TM.u.log.debug("TM.List.sortBy: sortName:sortDir:simpleCol - "+
               sortName+":"+sortDir+":"+simpleColumns);

      if (simpleColumns && gFolderDisplay.view.showThreaded)
        gFolderDisplay.view.showUnthreaded = true;

      MsgSortThreadPane(sortName);
    },

    groupBy: function(aEvent) {
TM.u.log.debug("TM.List.groupBy: ");
      gFolderDisplay.view._sort = [gFolderDisplay.view._sort[0]];
      MsgGroupBySort();
    },

    sortByType: function(aEvent) {
TM.u.log.debug("TM.List.sortByType: ");
      let columnId = TM.u.d("tmSortByMenu").getAttribute("curselected");
      let sortType1 = columnId == "byNone" ? nsMsgViewSortType["byNone"] :
                      nsMsgViewSortType[this.getSortTypeFromColId(columnId)] ||
                      nsMsgViewSortType["byDate"];
      if (sortType1 == nsMsgViewSortType["byCustom"] && gDBView.db)
        gDBView.db.dBFolderInfo.setProperty("customSortCol",
                                            gFolderDisplay.view.curCustomColumn = columnId);
      let sortDir1 = TM.u.d("tmSortOrder").getAttribute("sortDirection");
      sortDir1 = sortDir1 == "ascending" ? nsMsgViewSortOrder.ascending :
                                           nsMsgViewSortOrder.descending

      columnId = TM.u.d("tmSortBy2Menu").getAttribute("curselected");
      let sortType2 = columnId == "byNone" ? nsMsgViewSortType["byNone"] :
                      nsMsgViewSortType[this.getSortTypeFromColId(columnId)] ||
                      nsMsgViewSortType["byDate"];
      let sortDir2 = TM.u.d("tmSortOrder2").getAttribute("sortDirection2");
      sortDir2 = sortDir2 == "ascending" ? nsMsgViewSortOrder.ascending :
                                           nsMsgViewSortOrder.descending

TM.u.log.debug("TM.List.sortByType: sortType1:sortDir1:sortType2:sortDir2:custCol - "+
               sortType1+":"+sortDir1+":"+sortType2+":"+sortDir2+":"+
               gFolderDisplay.view.curCustomColumn);

      let grouped = gFolderDisplay.view.showGroupedBySort;
      if (this.simpleColumns() && gFolderDisplay.view.showThreaded)
        gFolderDisplay.view.showUnthreaded = true;

      if (grouped) {
        sortType2 = sortDir2 = null;
        gFolderDisplay.view.showGroupedBySort = false;
      }

      gFolderDisplay.view.sort(sortType1, sortDir1, sortType2, sortDir2);

      if (grouped) {
        TM.List.Sorting = false;
        gFolderDisplay.view.showGroupedBySort = true;
      }
    },

    orderBy: function(aEvent) {
TM.u.log.debug("TM.List.orderBy:");
      let orderBy2 = aEvent ? aEvent.target.getAttribute("sortDirection2") : null;
      if (orderBy2) {
        orderBy2 = orderBy2 == "ascending" ? "descending" : "ascending";
        GetThreadTree().setAttribute("sortDirection2", orderBy2);
        this.sortByType();
      }
      else {
        if (gFolderDisplay.view.showGroupedBySort)
          this.allThreads(false);

        MsgReverseSortThreadPane();

        if (gFolderDisplay.view.showGroupedBySort) {
          this.allThreads(true);
        }
      }
    },

    getSortNameFromType: function(aSortType) {
      if (!this._sortTypeNameMap) {
        this._sortTypeNameMap = new Map();
        for (let [name, type] in Iterator(nsMsgViewSortType))
          this._sortTypeNameMap.set(type, name);
      }
      let sortName = "byDate";
      if (this._sortTypeNameMap.has(aSortType))
        sortName = this._sortTypeNameMap.get(aSortType);
TM.u.log.trace("TM.List.getSortNameFromType: aSortType:sortName - "+aSortType+":"+sortName);

      return sortName;
    },

    getSortTypeFromColId: function(aColId) {
      if (this.columnsMap.has(aColId))
        return this.columnsMap.get(aColId)[this.kSortType];
      try {
        gDBView.getColumnHandler(aColId);
        return "byCustom";
      }
      catch (ex) {
        return null;
      }
    },

    isGroupedBySortHdr: function(aRow) {
      return gFolderDisplay.view.showGroupedBySort &&
             (gDBView.isContainer(aRow) || gDBView.getLevel(aRow) == 0);
    },

    getGroupedBySortHeader: function(aRow, aColId, aMsgHdr, aValidate, aSortType) {
//      let sortType = gFolderDisplay.view._sort[0][0];
      let sortType = aSortType || gFolderDisplay.view.primarySortType;
      switch (sortType) {
        case nsMsgViewSortType.byReceived:
        case nsMsgViewSortType.byDate:
        case nsMsgViewSortType.byAuthor:
        case nsMsgViewSortType.byStatus:
        case nsMsgViewSortType.byTags:
        case nsMsgViewSortType.byPriority:
        case nsMsgViewSortType.byAccount:
        case nsMsgViewSortType.byRecipient:
        case nsMsgViewSortType.byAttachments:
        case nsMsgViewSortType.byFlagged:
        case nsMsgViewSortType.byLocation:
          if (aValidate)
            return true;
          return gDBView.cellTextForColumn(aRow, aColId);
        case nsMsgViewSortType.bySubject:
          if (aValidate)
            return true;
          return this._fetchSubject(aMsgHdr);
        case nsMsgViewSortType.byCustom:
//          if (aValidate) break;
//          if (gDBView.curCustomColumn &&
//              this.columnsCustomMap.has(gDBView.curCustomColumn))
//            headerStr = this.columnsCustomMap
//                            .get(gDBView.curCustomColumn)[this.kCustColHandler]
//                            .getSortStringForRow(msgHdr);
//TM.u.log.info("TM.List.getGroupedBySortHeader: byCustom colId:curCustCol:headerStr - "+
//              aColId+":"+gDBView.curCustomColumn+":"+headerStr);
//          break;
        default:
          if (aValidate)
            return false;
          return "No sort by group for sort type " + sortType;
      }
    },

    sortGroupPopup: function(aPopup) {
TM.u.log.debug("TM.List.sortGroupPopup: ");
      let tree = GetThreadTree();
      let button = TM.u.d("tmSortGroupMenu");
      let bysort = TM.u.d("bySort");
      let bygroup = TM.u.d("byGroup");
      let grouped = gFolderDisplay.view.showGroupedBySort;
      let valid = this.getGroupedBySortHeader(null, null, null, true, null);
      let sel;

      if (gFolderDisplay.view.showThreaded) {
        tree.setAttribute("threaded", true);
        TM.u.d("threadCol").setAttribute("currentView", "threaded");
      }
      else {
        tree.removeAttribute("threaded");
        TM.u.d("threadCol").setAttribute("currentView", "unthreaded");
      }

      if (grouped) {
        sel = bygroup;
        tree.setAttribute("grouped", true);
      }
      else {
        sel = bysort;
        tree.removeAttribute("grouped");
      }
      sel.setAttribute("checked", true);
      button.label = sel.label;
      button.setAttribute("curselected", sel.getAttribute("itemid"));
      button.disabled = !valid;
      bygroup.hidden = bygroup.disabled = !valid;
    },
  
    getSortedCols: function() {
      let col = GetThreadTree().columns.getFirstColumn();
      let cols = [];
      for (col; col; col = col.getNext())
        cols.push(col);
      cols.sort((a, b) => {
        let c1 = (a.element.getAttribute("display") ||
                  a.element.getAttribute("label")).toLowerCase();
        let c2 = (b.element.getAttribute("display") ||
                  b.element.getAttribute("label")).toLowerCase();
        return c1.localeCompare(c2);
      });

      return cols;
    },
  
    sortByPopup: function(aPopup, aSortType, aSortOrder) {
TM.u.log.debug("TM.List.sortByPopup: START  aSortType:aSortOrder - "+
               aSortType+":"+aSortOrder);
      let tree = GetThreadTree();

      // Build the list initially.
      if (aPopup.childElementCount < 3) {
TM.u.log.debug("TM.List.sortByPopup: init popup - "+aPopup.id);
        let cols = this.getSortedCols();
        cols.forEach(col => {
          // Construct an entry for each sortable column.
          col = col.element;
          if (!col.hasAttribute("nosortby")) {
            let mi = document.createElement("menuitem");
            mi.setAttribute("type", "radio");
            mi.setAttribute("name", "sortby");
            mi.setAttribute("itemid", col.id);
            let columnName = col.getAttribute("display") ||
                             col.getAttribute("label");
            if (TM.List.columnsCustomMap.has(col.id))
              mi.setAttribute("iscustcol", true);
            // Binding label property setter is borked.
            mi.label = columnName;
            mi.setAttribute("label", columnName);
            aPopup.appendChild(mi);
          }
        })

        return;
      }

TM.u.log.debug("TM.List.sortByPopup: update popup - "+aPopup.id);
      let grouped = gFolderDisplay.view.showGroupedBySort;
      let threaded = tree.hasAttribute("threaded");
      let secondary = aPopup.hasAttribute("secondary");
      let custCol = gFolderDisplay.view.dbView.curCustomColumn;
      let sortCol = aSortType == nsMsgViewSortType.byNone ? "byNone" :
                    aSortType == nsMsgViewSortType.byCustom ? custCol :
                    ConvertSortTypeToColumnID(aSortType);
      let sortName, sortDirection, sortDirTooltip, button;
      let sortOrder = TM.u.d("tmSortOrder");
      let sortOrder2 = TM.u.d("tmSortOrder2");

      if (secondary) {
        button = TM.u.d("tmSortBy2Menu");
        let sortCol1Id = sortCol;
        let sortType2 = gDBView.secondarySortType;
        if (grouped && aSortType != nsMsgViewSortType.byNone) {
          sortType2 = nsMsgViewSortType.byDate;
          sortCol = ConvertSortTypeToColumnID(sortType2);
        }
        else {
          sortCol = aSortType == nsMsgViewSortType.byDate ? "idCol" :
                    sortType2 == aSortType ||
                      aSortType == nsMsgViewSortType.byId ||
                      aSortType == nsMsgViewSortType.byNone ||
                      sortType2 == nsMsgViewSortType.byCustom ? "byNone" :
                    ConvertSortTypeToColumnID(sortType2);
        }
TM.u.log.debug("TM.List.sortByPopup: secondary sortType2:sortOrder2:sortCol - "+
               sortType2+":"+gDBView.secondarySortOrder+":"+sortCol);
        for (let mi of aPopup.childNodes) {
          if (mi.classList.contains("menuitem-non-iconic"))
            continue;
          let sortCol2Id = mi.getAttribute("itemid");
          let sortCol2IsCustom = TM.List.columnsCustomMap.has(sortCol2Id);
          if (sortCol2Id == sortCol) {
            mi.setAttribute("checked", true);
            button.label = mi.label;
            button.setAttribute("curselected", sortCol);
          }
          else {
            mi.removeAttribute("checked");
          }
          if (sortCol2Id == sortCol1Id || sortCol2IsCustom ||
              sortCol2Id == "threadCol")
            mi.setAttribute("disabled", true);
          else
            mi.removeAttribute("disabled");
        }

        sortName = TM.List.getSortNameFromType(sortType2);
        tree.setAttribute("sortName2", sortName);
        sortDirection = sortCol == "byNone" ? "none" :
                        gDBView.secondarySortOrder == nsMsgViewSortOrder.ascending ?
                          "ascending" : "descending";
        sortDirTooltip = sortDirection == "none" ? "none" :
                         sortDirection == "ascending" ?
                           "tooltiptextasc" : "tooltiptextdes";
        tree.setAttribute("sortDirection2", sortDirection);
        button.disabled = grouped ||
                          aSortType == nsMsgViewSortType.byId ||
                          aSortType == nsMsgViewSortType.byDate ||
                          aSortType == nsMsgViewSortType.byNone;
        sortOrder2.disabled = button.disabled || sortCol == "byNone";
        sortOrder2.setAttribute("tooltiptext", sortOrder.getAttribute(sortDirTooltip));
      }
      else {
        button = TM.u.d("tmSortByMenu");
        for (let mi of aPopup.childNodes) {
          if (mi.classList.contains("menuitem-non-iconic"))
            continue;
          let sortCol1Id = mi.getAttribute("itemid");
          if (sortCol1Id == sortCol) {
            mi.setAttribute("checked", true);
            button.label = mi.label;
            button.setAttribute("curselected", sortCol);
          }
          else {
            mi.removeAttribute("checked");
          }
          let sortType = nsMsgViewSortType[this.getSortTypeFromColId(sortCol1Id)] ||
                         nsMsgViewSortType.byNone;
          if ((grouped && !this.getGroupedBySortHeader(null, null, null, true, sortType)) ||
               sortCol1Id == "threadCol")
            mi.setAttribute("disabled", true);
          else
            mi.removeAttribute("disabled");
        }

        sortName = TM.List.getSortNameFromType(aSortType);
        tree.setAttribute("sortName", sortName);
        sortDirection = sortCol == "byNone" ? "none" :
                        aSortOrder == nsMsgViewSortOrder.ascending ?
                          "ascending" : "descending";
        sortDirTooltip = sortDirection == "none" ? "none" :
                         sortDirection == "ascending" ?
                          "tooltiptextasc" : "tooltiptextdes";
        tree.setAttribute("sortDirection", sortDirection);
        sortOrder.disabled = sortCol == "byNone";
        sortOrder.setAttribute("tooltiptext", sortOrder.getAttribute(sortDirTooltip));
      }
TM.u.log.debug("TM.List.sortByPopup: buttonId:label:sortDirection:sortName:sortCol:custCol - "+
               button.id+":"+button.label+":"+sortDirection+":"+
               sortName+":"+sortCol+":"+custCol);
    },

    populateRow1: function(aRow, aRowN, aMsgHdr) {
      let msgKey = aMsgHdr.messageKey;
      let folderKey = aMsgHdr.folder.URI;
      if (!this._cache[folderKey])
        this._cache[folderKey] = {};
      if (!this._cache[folderKey][msgKey])
        this._cache[folderKey][msgKey] = {};

      let key = this._cache[folderKey][msgKey];
      if (key[aRowN] !== undefined)
        return key[aRowN];

      return key[aRowN] = this.getRowStr(aRow, aRowN, aMsgHdr);
    },

    populateRow: function(aRow, aRowN, aMsgHdr) {
//TM.u.log.trace("TM.List.populateRow: row:map - "+
//               aRowN+":"+this.getMap()[aRowN].toSource());
      let rowStr = "";
      let listType = aRowN == 0 ? TM.v.HEADER[TM.v.HEADER.LIST] :
                                  TM.v.HEADER[TM.v.HEADER.LISTMR];
      let rowN = "row" + (aRowN < 2 ? 1 : aRowN);
      let map = TM.u.getHeaderMap(TM.u.MessageType(aMsgHdr));
      let count = 0;
      let sepL = "", sepR = "";
      TM.u.isRtl ? sepL = TM.u.HEADER_SEP : sepR = TM.u.HEADER_SEP;
      for (let [name, value] in Iterator(map[listType][rowN])) {
        let headerField = TM.List._fetch(aRow, name, aMsgHdr);
        if (!headerField)
          continue;
//TM.u.log.debug("TM.List.populateRow: row:name:value - "+ aRowN+":"+name+":"+value);
        rowStr += count ? " " : "";
        let displayName = value[TM.v.kDisplayName];
        let displayLabel = "";
        if (displayName !== null) {
          displayLabel = displayName ||
                         TM.headerDisplaySpecial(TM.u.MessageType(aMsgHdr), name) ||
                         TM.u.commonHeaders.get(name)[TM.v.kLabelValue] ||
                         name;
          rowStr += displayName || sepL + displayLabel + sepR;
        }

        rowStr += headerField === true ? "" : headerField;
        count++;
      }

      return rowStr;
    },

    _fetch: function(aRow, aName, aMsgHdr) {
      let ccolId = TM.u.commonHeaders.has(aName) ?
                    TM.u.commonHeaders.get(aName)[TM.v.kColId] : null;
      let ocolId = TM.u.otherValues.has(aName) ?
                    TM.u.otherValues.get(aName)[TM.v.kColId] : null;
//if (aName == "x-message")
//TM.u.log.debug("TM.List._fetch: aRow:colId:aName - "+aRow+":"+(ccolId || ocolId)+":"+aName);
//      if (TM.u.columnsCustom.has(aName))
//        return this.columnsCustom.get(ocolId)[kCustColHandler]
//                   .getCellText(aRow, { id:ocolId });

      /* Col data is not cached.*/
      if (ccolId)
        return gDBView.cellTextForColumn(aRow, ccolId);
      if (ocolId) {
        let headerField = TM.getOtherValues(aRow, aName, true);
        if (ocolId == "statusCol" && headerField)
          headerField = true;
        return headerField;
      }

      let msgKey = aMsgHdr.messageKey;
      let folderKey = aMsgHdr.folder.URI;
      if (!this._cache[folderKey])
        this._cache[folderKey] = {};
      if (!this._cache[folderKey][msgKey])
        this._cache[folderKey][msgKey] = {};

      let key = this._cache[folderKey][msgKey];
      if (key[aName] !== undefined)
        return key[aName];

TM.u.log.debug("TM.List._fetch: hdr not cached aRow:aName - "+aRow+":"+aName);
/*
      if (ccolId)
        return key[aName] = gDBView.cellTextForColumn(aRow, ccolId);
      if (ocolId) {
        let headerField = TM.getOtherValues(aRow, aName, true);
        if (ocolId == "statusCol" && headerField)
          headerField = true;
        return key[aName] = headerField;
      }
*/
      if (aName == 'x-message')
        return this._fetchMessage(aRow, aMsgHdr, key, aName);

      return this._fetchHeader(aRow, aMsgHdr, key, aName);
    },

    Re: "Re: ",
    _fetchSubject: function(aMsgHdr) {
      let Re = aMsgHdr.flags & Ci.nsMsgMessageFlags.HasRe ? this.Re : "";
      return Re + aMsgHdr.mime2DecodedSubject;
    },

    get imapAllowDownload()  { return true; },
    get saneBodySize()       { return true; },
    get imapPartsOnDemand()  { return false; },
    // Throws for nntp if false.
    get encryptedParts()     { return true; },

    _fetchHeader: function(aRow, aMsgHdr, aKey, aHdrName) {
      let msgKey = aMsgHdr.messageKey;
      try {
        MsgHdrToMimeMessage(aMsgHdr, null, function(aMsgHdr, aMimeMsg) {
          if (aMimeMsg && aMimeMsg.headers[aHdrName] && aMimeMsg.headers[aHdrName][0])
            aKey[aHdrName] = aMimeMsg.headers[aHdrName][0];
          else
            // No such header name, always true for nntp.
            aKey[aHdrName] = null;
          GetThreadTree().treeBoxObject.invalidateRow(aRow);
TM.u.log.debug("TM.List._fetchHeader: aRow:header:value - "+
               aRow+":"+aHdrName+":"+aKey[aHdrName]);
        },
        this.imapAllowDownload,
        { saneBodySize:          this.saneBodySize,
          partsOnDemand:         this.imapPartsOnDemand,
          examineEncryptedParts: this.encryptedParts });
      }
      catch (ex) {
        // Couldn't get the headers.
TM.u.log.error("TM.List._fetchHeader: row:header:error - "+aRow+":"+aHdrName+":"+ex);
        return "??";
      }
  
      return "";
    },

    _fetchMessage: function(aRow, aMsgHdr, aKey, aHdrName) {
      let msgKey = aMsgHdr.messageKey;

      if (aMsgHdr.folder.server.type == "nntp")
        return aKey[aHdrName] = "?";
      let urlListener = null; // TM.List;
      let outAsync = {};
      let previewText = "??";
      try {
        outAsync.value = aMsgHdr.folder.fetchMsgPreviewText([msgKey], 1, false,
                                                            urlListener);
        previewText = aMsgHdr.getStringProperty("preview");
TM.u.log.debug("TM.List._fetchMessage: previewText raw - "+ previewText);
        let converter = TM.u.converter;
        converter.charset = "UTF-8";
        previewText = converter.ConvertToUnicode(previewText);
        // Simple no Content-Type mbox parsing bug for text < 255, doesn't
        // detect next 'From - ' boundary.
        previewText = previewText.split(/ From - /)[0];
TM.u.log.debug("TM.List._fetchMessage: msgKey:outAsync:text - "+
               msgKey+":"+outAsync.toSource() +":"+previewText);
      }
      catch (ex) {
        // Couldn't get the headers.
TM.u.log.error("TM.List._fetchMessage: error - "+ex);
      }
  
      return aKey[aHdrName] = previewText;
    },
/*
    OnStartRunningUrl: function(url) {
TM.u.log.debug("TM.List.OnStartRunningUrl: url - "+url);
    },
    OnStopRunningUrl: function(url, exitCode) {
TM.u.log.debug("TM.List.OnStopRunningUrl: url:exitCode - "+url+":"+exitCode);
    },
*/
  },


  //**************************************************************************//
  // Messagepane Headers functions.

  Headers: {

    headerIndex: 1,
    toolbarInitialized: false,

    onLoad: function() {
TM.u.log.debug("TM.Headers.onLoad: START ");
      if (gMessageListeners.indexOf(TotalMessage) == -1)
        // Not registered.
        gMessageListeners.push(TotalMessage);
      else
        // Already registered; if reload on a layout change, toggle select.
        this.toggle("select");

      Services.obs.addObserver(TM.Editor.Observer, "MsgCreateDBView", false);
      if (gDBView)
        gDBView.addColumnHandler("threadCol", TM.Editor.threadPaneDND);

      TM.u.g("messagepane").addEventListener("DOMContentLoaded", TM.Browser.onLoad);

      TM.v.saveShowHeaders = Services.prefs.getIntPref("mail.show_headers");
      TM.Headers.headerIndex = parseInt(TM.u.viewHeader.getAttribute("index"));

      let box = TM.u.winType == "mail:messageWindow" ? TM.u.messagepanebox :
                                                       TM.u.singlemessage;
      let bar = TM.u.g("msgNotificationBar");
      let splitter = TM.u.g("header-splitter");
      box.insertBefore(splitter, TM.u.msgHeaderView.nextElementSibling);
      box.insertBefore(bar, box.firstChild);

      let headerView = TM.v.HEADER[TM.Headers.headerIndex];
      TM.u.g("messagepanebox").setAttribute("header", headerView);
      TM.u.g("msgHeaderView").height = TM.u.viewHeader.getAttribute(headerView);

      // Wait a bit for header toolbars to construct first. Mostly due to ABP
      // and its legacy toolbarbutton handling.
      setTimeout(function() { TM.u.initializeToolbars(); }, 5000);
TM.u.log.debug("TM.Headers.onLoad: FINISHED ");
    },

    onKeyPressIndicator: function(aEvent) {
      if (aEvent.keyCode != KeyEvent.DOM_VK_RETURN)
        return;

TM.u.log.debug("TM.Headers.onKeyPressIndicator: CONTINUE "+aEvent.target.id);
      TM.u.messagepanebox.setAttribute("keyfocus", true);
      TM.u.c(TM.u.messageHeader, "headerButton").focus();
    },

    onBlurToggle: function(aEvent) {
      if (TM.Headers.headerIndex || !TM.u.messagepanebox.hasAttribute("keyfocus"))
        // Not toggled off or key focused.
        return;

TM.u.log.debug("TM.Headers.onBlurToggle: continue "+aEvent.target.id);
      window.document.commandDispatcher.rewindFocus();
      TM.u.messagepanebox.removeAttribute("keyfocus");
      window.document.commandDispatcher.advanceFocus();
      window.document.commandDispatcher.rewindFocus();
    },

    onKeyPress: function(aEvent) {
      let docURL = aEvent.target.ownerDocument.URL;
TM.u.log.debug("TM.Headers.onKeyPress: START id:docURL - "+
               aEvent.target.id +":"+docURL);

      if (docURL == TM.v.headersUrl)
        // In headers browser.
TM.u.log.debug("TM.Headers.onKeyPress: START return "+aEvent.keyCode);
//      if (aEvent.keyCode == KeyEvent.DOM_VK_RETURN ||
//          aEvent.keyCode == KeyEvent.DOM_VK_TAB)
//        return;
    },

    onButtonKeyPress: function(aEvent) {
TM.u.log.debug("TM.Headers.onButtonKeyPress: START "+aEvent.keyCode);
      if (aEvent.keyCode == KeyEvent.DOM_VK_RETURN ||
          aEvent.keyCode == KeyEvent.DOM_VK_TAB)
        return;

      // Remove focus from tabbable buttons for back/forward.
      // XXX: due to the button binding, an arrow key while the header button is
      // focused will throw once; no event cancelling functions cancel the event.
      aEvent.target.blur();
//    window = aEvent.target.ownerDocument.defaultView.gBrowserWindow;
    },

    onResize: function() {
TM.u.log.debug("TM.Headers.onResize: START ");
      this.setWrapState();
      this.resizeFullHeader();
    },

    resizeFullHeader: function() {
TM.u.log.debug("TM.Headers.resizeFullHeader: START ");
      let headerIndex = TM.Headers.headerIndex;
      if (headerIndex != TM.v.HEADER.FULL)
        return;

      let headerBody = TM.u.messageHeaderBody;
      let headerDeck = TM.u.c(TM.u.messageHeader, "headerDeck");
      let fullHeaderDiv = TM.u.c(TM.u.messageHeader, "fullHeaderDiv");
      let editingTable = TM.u.c(TM.u.messageHeader, "editingTable");
      // Set full header height for scrollbars to appear correctly.
      let height = headerBody.parentElement.clientHeight -
                     headerDeck.clientHeight -
                     editingTable.clientHeight;
TM.u.log.debug("TM.Headers.resizeFullHeader: height - "+height);
      if (height > 0)
        fullHeaderDiv.style.height = height + "px";

      // Adjust search markers, if any.
//    TotalMessageHighlight.highlightSetMarkers();
    },

    toggle: function(aType) {
      let headerIndex = TM.Headers.headerIndex;
      let headerBody = TM.u.messageHeaderBody;
      if (!headerBody || !TM.v.msgHdr || (aType == "key" && TM.v.editingMode))
        // No message or a key toggle in editing mode.
        return;
TM.u.log.debug("TM.Headers.toggle: START type:headerIndex - "+aType+" : "+headerIndex);
TM.u.log.debug("TM.Headers.toggle: editingMode:dragRow - "+
               TM.v.editingMode+" : "+(TM.v.Editor.dragRow ? TM.v.Editor.dragRow : "null"));

      // Build brief header always, if not already built for current message.
      TotalMessage.createHeader(TM.v.HEADER[TM.v.HEADER.BRIEF]);

      if (aType == "toggle" || aType == "key") {
        // Toggle to next.
        headerIndex = ++headerIndex > TM.v.HEADER.TOGGLE_TYPES_COUNT ?
                        0 : headerIndex;
        TM.Headers.headerIndex = headerIndex;
      }

      TM.u.viewHeader.setAttribute("index", headerIndex);
      TM.u.viewHeader.setAttribute("header", TM.v.HEADER[headerIndex]);
      TM.u.viewHeader.height = TM.u.viewHeader.getAttribute(TM.v.HEADER[headerIndex]);

      // Room for page curl indicator.
      if (TM.u.messagepaneContentDocument && TM.u.messagepaneContentDocument.body)
        TM.u.messagepaneContentDocument.body.style.marginTop =
          TM.Headers.headerIndex ? "8px" : "24px";

      if (headerIndex == TM.v.HEADER.BASIC)
        // Build basic header, if not already built for current message.
        TotalMessage.createHeader(TM.v.HEADER[TM.v.HEADER.BASIC]);

      if (headerIndex == TM.v.HEADER.FULL) {
        // Ensure basic is built, build full header.
        TotalMessage.createHeader(TM.v.HEADER[TM.v.HEADER.BASIC]);
        TotalMessage.createHeader(TM.v.HEADER[TM.v.HEADER.FULL]);
      }

      headerBody.setAttribute("header", TM.v.HEADER[headerIndex]);
      // Clear editing mode bar if not editing.
      headerBody.setAttribute("editingMode", TM.v.editingMode);

      if (!TM.v.editingMode) {
        // Update hidden and disabled states of any toolbarbutton and
        // toolbaritem to reflect state in the Tb header toolbar, which has
        // just been updated.
        TM.Headers.updateHeaderButtonState();
        TM.Headers.setToolbarState();
      }
      else if (TM.v.mouseDragDropMode || TM.v.keyDragDropMode) {
        // In a dnd rethread.  Don't need to show fullheader rows.
        let editingTable = TM.u.c(TM.u.messageHeader, "editingTable");
        TM.u.viewHeader.setAttribute("height", editingTable.clientHeight);
      }


      if (aType == "select" || aType == "edit")
        // Make sure select results in nice headers wrapped content flow.
        TM.Headers.onResize();
    },


    toggleToolbar: function(aChecked) {
TM.u.log.debug("TM.Headers.toggleToolbar: START "+aChecked);
      let viewHeadersToolbar = TM.u.g("viewHeadersToolbar");
      let toolbox = TM.u.c(TM.u.messageHeader, "header-view-toolbox")
      viewHeadersToolbar.setAttribute("checked", aChecked ? true : false);
      toolbox.hidden = aChecked ? false : true;
      TM.Headers.setToolbarState();
    },

    toggleRemoteContent: function(aEvent) {
TM.u.log.debug("TM.Headers.toggleRemoteContent: START ");
      TM.v.toggleRemoteContentAllow = true;
      TM.v.msgHdr.setUint32Property("remoteContentPolicy", kAllowRemoteContent);
      gFolderDisplay.view.dbView.reloadMessage();
    },

    toggleWrap: function(aEvent) {
TM.u.log.debug("TM.Headers.toggleWrap: START ");
      let headerBody = TM.u.messageHeaderBody;
      let viewHdrWrap = TM.u.g("viewHeaderWrap");
      let checked = TM.u.c(TM.u.messageHeader, "hdrWrapButton").checked;

      if (checked) {
        headerBody.classList.add("wrap");
        TM.u.viewHeader.height = headerBody.scrollHeight;
      }
      else {
        headerBody.classList.remove("wrap");
        TM.u.viewHeader.height =
          TM.u.viewHeader.getAttribute(TM.v.HEADER[TM.Headers.headerIndex]);
      }

      viewHdrWrap.setAttribute("checked", checked);
    },

    toggleWrapHeader: function(aHeaderLabelNode) {
TM.u.log.debug("TM.Headers.toggleWrapHeader: START ");
      this.wrapHeader(aHeaderLabelNode, false);
      // Wait for css.
      setTimeout(function() {
        if (TM.Headers.isWrapped(TM.u.c(TM.u.messageHeader, "headerDeck"))) {
          // Reset header height to flow wrapped content.
          TM.u.viewHeader.height = TM.u.messageHeaderBody.scrollHeight;
        }
        else {
          // Restore user set height if nothing wraps, or on init and hover.
          TM.u.viewHeader.height =
            TM.u.viewHeader.getAttribute(TM.v.HEADER[TM.Headers.headerIndex]);
        }

        TM.Headers.resizeFullHeader();
      }, 20);
    },

    wrapHeader: function(aHeaderLabelNode, aAutoWrap) {
TM.u.log.debug("TM.Headers.wrapHeader: START auto - "+aAutoWrap);
      let headerValueNode;
      let headerLabelNode = aHeaderLabelNode;
      let headerBody = TM.u.messageHeaderBody;

      // Get the data containing node.  Normally, the headerValueNode is
      // 1 nextSibling from the headerLabelNode.
      if (headerLabelNode.nextSibling) {
        if (headerLabelNode.nextSibling.classList.contains("headerSep"))
          // In full headers/edit mode, there's a separator in between.
          headerValueNode = headerLabelNode.nextSibling.nextSibling;
        else
          headerValueNode = headerLabelNode.nextSibling;
        if (!headerValueNode.classList.contains("headerValue"))
          // For column 1.
          headerValueNode = headerValueNode.firstChild.firstChild.firstChild;
      }
      else {
        // For row 1 column 1.
        let row1col1Node = headerLabelNode.parentNode.parentNode.parentNode;
        headerValueNode = row1col1Node.nextSibling.firstChild.firstChild.firstChild;
      }

      if (!headerBody.classList.contains("wrap") &&
          (headerLabelNode.classList.contains("wrappable") ||
           headerLabelNode.classList.contains("wrapped"))) {
        // Set wrap and resize only if global wrap not set.
        headerValueNode.classList.toggle("wrap");

TM.u.log.debug("TM.Headers.wrapHeader: labelNode.id:valueNode.id:classList - "+
               headerLabelNode.id+":"+headerValueNode.id+":"+headerValueNode.classList);
        // Break on separator in multi item header if wrap, otherwise
        // remove <br>.
        let container = headerValueNode.firstChild;
        let childrenWidth = 0;
        let separatorNode;
        if (container.classList.contains("multiItemContainer")) {
          for (let child of container.firstChild.childNodes) {
            childrenWidth += child.offsetWidth;
            if (separatorNode && child.classList.contains("multiItem")) {
              if (headerValueNode.classList.contains("wrap") || aAutoWrap) {
                if (container.clientWidth < childrenWidth)
                  separatorNode.appendChild(document.
                                            createElementNS(TM.v.HTML_NS, "br"));
              }
              else if (separatorNode.lastElementChild &&
                       separatorNode.lastElementChild.tagName == "br")
                separatorNode.lastChild.remove();
            }
            if (child.classList.contains("multiItemSeparator"))
              separatorNode = child;
          }
        }
      }
    },

    updateHeaderButtonState: function() {
TM.u.log.debug("TM.Headers.updateHeaderButtonState: START ");
      let toolbarTB  = TM.u.g("header-view-toolbar");
      let toolbar = TM.u.c(TM.u.messageHeader, "header-view-toolbar");

      let elementsTB = toolbarTB.querySelectorAll("toolbarbutton, menuitem");
      for (let nodeTB of elementsTB) {
        let node = nodeTB.id ? TM.u.c(TM.u.messageHeader, nodeTB.id) : null;
        if (!node)
          continue;
        node.hidden = node.tagName == "toolbarbutton" &&
                      (nodeTB.hidden || nodeTB.disabled);
        node.disabled = nodeTB.disabled;
TM.u.log.trace("TM.Headers.updateHeaderButtonState: id:hidden "+node.id+":"+node.hidden);
      }
    },

    setToolbarState: function() {
      // Set header toolbar widths, need to be visible first. Since buttons
      // are shown variably based on message, do this every time. Do this
      // before calculating wrap state.
      let headerToolbarRow = TM.u.c(TM.u.messageHeader, "headerToolbarRow");
      let toolbar = TM.u.c(TM.u.messageHeader, "header-view-toolbar");
      let toolbarTD = TM.u.c(TM.u.messageHeader, "headerToolbarBoxRight");
      if (!toolbar || !toolbarTD)
        return;

      headerToolbarRow.setAttribute(
        "flatstyle", TM.u.g("header-toolbar-buttonstyle").hasAttribute("checked"));

//      setTimeout(function() {
        toolbarTD.style.width = "";
        toolbarTD.style.width = toolbar.clientWidth + "px";
TM.u.log.debug("TM.Headers.setToolbarState: right width - "+toolbar.clientWidth);

        TM.Headers.setWrapState();
//      }, 20);
    },

    setWrapState: function() {
TM.u.log.debug("TM.Headers.setWrapState: START");
      // Get the document line height.
      let headerBody = TM.u.messageHeaderBody;
      let dataCStyle = window.getComputedStyle(headerBody, null);
      let dataLnHt = dataCStyle.getPropertyValue("line-height").replace(/px/, "");

      // Set wrap state on header labels.
      let node, labelNode, headerData, wrappable;
      let nodes = TM.u.messageHeaderContentDocument
                    .querySelectorAll(".headerValue");
      for (let node of nodes) {
        wrappable = false;
        if (node.previousSibling)
          if (node.previousSibling.classList.contains("headerSep"))
            // In full headers/edit mode, there's a separator in between.
            labelNode = node.previousSibling.previousSibling;
          else
            // Usually the headerLabel is a 1 previousSibling of headerValue,
            // except for column 1.
            labelNode = node.previousSibling;
        else
          // For column 1.  In this case, walk up the very specific structure
          // to get the headerLabel node.
          labelNode = node.parentNode.parentNode.parentNode.previousSibling;

        let container = node.firstChild;
        if (container.classList.contains("container")) {
TM.u.log.trace("TM.Headers.setWrapState: container headerValue.id - "+
               (node.id ? node.id : "(fullheader)"));
          if (node.classList.contains("wrap") ||
              node.clientWidth < container.firstChild.offsetWidth ||
              (node.classList.contains("noEllipsis") && node.offsetWidth > 110))
            wrappable = true;
        }
        else if (container.classList.contains("multiItemContainer")) {
TM.u.log.trace("TM.Headers.setWrapState: multiItemContainer headerValue.id - "+
               (node.id ? node.id : "(fullheader)"));
          let childrenWidth = 0;
          for (let child of container.firstChild.childNodes) {
            childrenWidth += child.offsetWidth;
TM.u.log.trace("TM.Headers.setWrapState: multiItemContainer child:width "+
               child.className+":"+child.offsetWidth);
            if (container.clientWidth < childrenWidth) {
              wrappable = true;
              break;
            }
          }
        }

        if (wrappable) {
TM.u.log.trace("TM.Headers.setWrapState: WRAPPABLE labelNode.id - "+
               (labelNode.id ? labelNode.id : "(fullheader)"));
          labelNode.classList.add("wrappable");

          if (labelNode.hasAttribute("autowrap")) {
            this.wrapHeader(labelNode, true);
            labelNode.removeAttribute("autowrap");
          }
        }
        else
          labelNode.classList.remove("wrappable");

        // Need the height of the <a> content as the containing <td> sizes
        // to match any other <td> in the row that has expanded to fit
        // wrapped content.
        let dataHt = node.firstChild.firstChild.scrollHeight;
        if (Math.round(dataHt / dataLnHt) > 1)
          labelNode.classList.add("wrapped");
        else
          labelNode.classList.remove("wrapped");
      }
    },

    isWrapped: function(aNode) {
      // Is any child value in the passed node wrapped?  Not just toggled to
      // wrap, but also flowed to multiple lines.
      let oneWrapped = false;
      let wrappedNodes = aNode.querySelectorAll(".wrap");
      for (let node of wrappedNodes) {
        let dataCStyle = window.getComputedStyle(node, null)
        let dataHt = dataCStyle.getPropertyValue("height");
        let dataLnHt = dataCStyle.getPropertyValue("line-height");
TM.u.log.trace("TM.Headers.isWrapped: node.id:ht:lnht - "+
               node.id+":"+dataHt+":"+dataLnHt);
        if (dataHt != dataLnHt) {
          oneWrapped = true;
          break;
        }
      }
TM.u.log.trace("TM.Headers.isWrapped: "+oneWrapped);

      return oneWrapped;
    },

    onClick: function(aEvent) {
      if (aEvent.button != 0 || aEvent.detail == 2 || !aEvent.target.className)
        return;
TM.u.log.debug("TM.Headers.onClick: START type:detail - "+aEvent.type+":"+ aEvent.detail);
TM.u.log.debug("TM.Headers.onClick: START node.className: "+ aEvent.target.className);

      let node = aEvent.target;
      if (node.classList.contains("headerLabel")) {
        // Clicked on a header label.
        this.toggleWrapHeader(node);
      }

      if (node.classList.contains("emailAddress")) {
        // Clicked on an email address.
        TM.u.g("emailAddressPopup").openPopup(node, "after_start",
                                              0, 0, false, false);
      }

      if (node.classList.contains("newsgroup")) {
        // Clicked on a newsgroup.
        TM.u.g("newsgroupPopup").openPopup(node, "after_start",
                                           0, 0, false, false);
      }

      if (node.classList.contains("abCard")) {
        // Clicked on the add/edit card image.
TM.u.log.trace("TM.Headers.onClick: abCard: "+
               node.previousElementSibling.className);
TM.u.log.trace("TM.Headers.onClick: abCard: "+
               node.previousElementSibling.cardDetails.toSource());
        TM.u.topWindow.onClickEmailStar(aEvent, node.previousElementSibling);
      }

      if (node.hasAttribute("messageId")) {
        // Clicked on the go to message image (for newsgroups).
        let messageId = node.getAttribute("messageId");
TM.u.log.trace("TM.Headers.onClick: goToMessage: "+messageId);
//        TM.resetMsgLoadFlags();
        try {
          TM.u.topWindow.OpenMessageForMessageId(messageId);
        }
        catch (ex) {
          // Tb doesn't catch bad msgDatabase in CheckForMessageIdInFolder()...
          if (TM.u.winType == "mail:messageWindow") {
            // And OpenMessageByHeader() is also broken in standalone message
            // window.  But don't put up a misleading message.
          }
          else {
            let messageIdStr = "<" + messageId + ">";
            let bundle = TM.u.g("bundle_messenger");
            let errorTitle =
                bundle.getString("errorOpenMessageForMessageIdTitle");
            let errorMessage =
                bundle.getFormattedString("errorOpenMessageForMessageIdMessage",
                                          [messageIdStr]);
            let promptService = TM.u.topWindow.Services.prompt;
            promptService.alert(TM.u.topWindow, errorTitle, errorMessage);
         }

          TM.u.topWindow.setCursor("auto");
          TM.u.topWindow.ReloadMessage();
        }
      }

      if (node.classList.contains("linkified")) {
        // Clicked on a linkified subject or website/content-base link.
        aEvent.preventDefault();
        aEvent.stopImmediatePropagation();
        if (TM.u.g("mailContext-openLinkInMessagePane").hasAttribute("checked")) {
          let messagepane = TM.u.messagepane;
          if (messagepane.currentURI.scheme.startsWith("http")) {
            TM.u.topWindow.ReloadMessage();
          }
          else {
            let href = node.getAttribute("href");
TM.u.log.debug("TM.Headers.onClick: load messagepane, href - "+href);
            let loadFlag = messagepane.webNavigation.LOAD_FLAGS_NONE;
            messagepane.webNavigation.loadURI(href, loadFlag, null, null, null);
          }
        }
        else {
          contentAreaClick(aEvent);
        }
      }
    },

    onToolbarContext: function(aEvent) {
TM.u.log.debug("TM.Headers.onToolbarContext: START "+aEvent.target.id);
     let node = aEvent.target;
     let toolbarContext = TM.u.g("header-toolbar-context-menu");
     toolbarContext.openPopup(node, "after_pointer", 0, 0, true, false);
    },

    isEditable: function() {
      // Enable header editing, one message must be selected. If in
      // conversation mode on collapsed thread, the thread must be open.
      // No edit for newsgroup or imap readonly or trash folders.
      // Also, too many issues with editing in standalone window..
      if (TM.u.winType != "mail:3pane" || !TM.v.msgHdr ||
          GetNumSelectedMessages() != 1 ||
          gDBView.currentlyDisplayedMessage == TM.Editor.nsMsgKey_None ||
          !TM.v.msgHdr.folder.canFileMessages ||
          !TM.v.msgHdr.folder.canDeleteMessages ||
          TM.v.msgHdr.folder.isSpecialFolder(Ci.nsMsgFolderFlags.Trash, true))
        return false;
      return true;
    },

    enableEditMode: function(aEditingMode) {
      if (!this.isEditable())
        return;

      TM.u.g("editorActive").setAttribute("checked", aEditingMode);
      TM.v.editingMode = aEditingMode;
TM.u.log.debug("TM.Headers.enableEditMode: "+TM.v.editingMode);
      if (!TM.v.editingMode) {
        TM.Editor.onHeaderEditorCommand(null, "exit");
        return;
      }

      TM.v.Editor["editorButtonSave"] =
         TM.u.c(TM.u.messageHeader, "editorButtonSave");
      TM.v.Editor.editorButtonSave.setAttribute("disabled", "true");
      TM.v.Editor["editorButtonBackup"] =
         TM.u.c(TM.u.messageHeader, "editorButtonBackup");
      // Save current header toggle state.
      TM.Editor.saveHeaderIndex = TM.Headers.headerIndex;

      function callback() {
        // Show full headers in edit mode, after the message is read/parsed.
        TM.Headers.headerIndex = TM.v.HEADER.FULL;
        TM.v.fullHeaderBuilt = false;
        TM.Headers.toggle("edit");
      }

      TM.Editor.getMessageStream(TM.v.msgHdr, callback);
    },

    onDblClick: function(aEvent) {
      // Start the editor on dblclick on a header value, in fullheader mode.
      if (!TM.v.editingMode || aEvent.button != 0 )
        return;
TM.u.log.debug("TM.Headers.onDblClick: START clicks: "+ aEvent.detail);
      TM.Editor.startEditor(aEvent);
    },

    ThreadPaneOnDragStart: function(aEvent) {
TM.u.log.debug("TM.Headers.ThreadPaneOnDragStart: id - "+ aEvent.target.parentNode.id);
      TM.v.inDragStart = true;
      if (TM.Headers.isEditable() && gFolderDisplay.view.showThreaded) {
        // Enter edit mode if one message selected for dnd in threadpane. Also
        // only enable dnd if in threaded view, too confusing otherwise.
        // Set the drag target row.
        TM.v.Editor.dragRow = gDBView.currentlyDisplayedMessage;
        aEvent.dataTransfer.setData("text/x-moz-message-dragrow", TM.v.Editor.dragRow);
TM.u.log.debug("TM.Headers.ThreadPaneOnDragStart: dragRow - "+ TM.v.Editor.dragRow);
        if (!TM.v.editingMode)
          // Start the editor if not already started.
          TM.Headers.enableEditMode(true);

        TM.v.mouseDragDropMode = true;
        // If already in key dnd, cancel it and go to mouse dnd.
        if (TM.v.keyDragDropMode)
          TM.Editor.enableKeyDragDropInThreadpane(TM.v.keyDragDropMode);
      }

      TM.Headers._ThreadPaneOnDragStart(aEvent);
    },

    ThreadPaneOnDragOver: function(aEvent) {
TM.u.log.trace("TM.Headers.ThreadPaneOnDragOver: id - "+ aEvent.target.parentNode.id);
      if (aEvent.target.parentNode.id == "threadTree" &&
          aEvent.dataTransfer.getData("text/x-moz-message-dragrow") != "") {
        // Got a drag row type, set the drop target while dragging one message
        // onto another.
        let ds = Cc["@mozilla.org/widget/dragservice;1"]
                   .getService(Ci.nsIDragService).getCurrentSession();
        let targetRow = TM.Editor.threadPaneDND.getTargetRow(aEvent);
        let curRow = TM.Editor.threadPaneDND.getCurRow();
        let curRowParent = TM.Editor.threadPaneDND.getCurRowParent(curRow);
        TM.v.Editor.dropRow = null;
        GetThreadTree().treeBoxObject.invalidate();

        if (targetRow == -1 || curRow == targetRow || curRowParent == targetRow) {
          TM.v.Editor.dragRow = null;
          return ds.canDrop = false;
        }

        TM.v.Editor.dragRow = curRow;
        TM.v.Editor.dropRow = targetRow;
        return ds.canDrop = true;
      }

      return TM.Headers._ThreadPaneOnDragOver(aEvent);
    },

    ThreadPaneOnDragEnter: function(aEvent) {
TM.u.log.trace("TM.Headers.ThreadPaneOnDragEnter: id - "+ aEvent.target.parentNode.id);
      if (aEvent.target.parentNode.id == "threadTree" && 
          aEvent.dataTransfer.getData("text/x-moz-message-dragrow") != "" &&
          !TM.v.editingMode) {
        // If re-entering threadpane with a row drag target, go to edit mode.
        TM.v.mouseDragDropMode = true;
        TM.Headers.enableEditMode(true);
      }
    },

    ThreadPaneOnDragLeave: function(aEvent) {
TM.u.log.trace("TM.Headers.ThreadPaneOnDragLeave: id - "+ aEvent.target.parentNode.id);
      if (aEvent.target.parentNode.id == "threadTree" && TM.v.editingMode) {
        // If leaving threadpane, exit edit mode, if enabled.
        TM.Headers.enableEditMode(false);
      }
    },

    ThreadPaneOnDragEnd: function(aEvent) {
TM.u.log.debug("TM.Headers.ThreadPaneOnDragEnd: id - "+ aEvent.target.parentNode.id);
      TM.v.inDragStart = false;
    },

    ThreadPaneOnDrop: function(aEvent) {
TM.u.log.debug("TM.Headers.ThreadPaneOnDrop: id - "+ aEvent.target.parentNode.id);
//      TM.v.inDragStart = false;
      if (aEvent.target.parentNode.id == "threadTree" &&
          aEvent.dataTransfer.getData("text/x-moz-message-dragrow") != "") {
TM.u.log.debug("TM.Headers.ThreadPaneOnDrop: START");
        // Thread a message to another via drag and drop on the target.
        // Dropping the dragged message onto the drop message indicates we want
        // to thread the dragged as a child of the drop. Get the msgHdr and
        // in-reply-to/references headers, update the dropped.
        let targetRow = TM.Editor.threadPaneDND.getTargetRow(aEvent);
        let curRow = TM.Editor.threadPaneDND.getCurRow();
        let curRowParent = TM.Editor.threadPaneDND.getCurRowParent(curRow);
        if (targetRow == -1 || targetRow == curRowParent )
          return;

TM.u.log.debug("TM.Headers.ThreadPaneOnDrop: dragRow:dropRow - "+ TM.v.Editor.dragRow+":"+TM.v.Editor.dropRow);
        TM.Editor.getReferencesForThreading(targetRow);
        return;
      }

      TM.Headers._ThreadPaneOnDrop(aEvent);
    }

  }, // Headers


  //**************************************************************************//
  // Editor functions.

  Editor: {
    CR: "",
    LF: "\n",
    CRLF: "\n",
    MIN_ONE_FIELDS: ["date", "from", "message-id"], //message-id is rfc 'should'..
    MAX_ONE_FIELDS: ["date", "from", "sender", "reply-to", "to", "cc", "bcc",
                     "message-id", "in-reply-to", "references", "subject"],
    TMP_FILENAME: "tmpHdrEditMsg.eml",
    nsMsgKey_None: 0xFFFFFFFF,
    saveHeaderIndex: null,
    CharSet: "UTF-8",
    HEADER_TM_EDITED: "X-TotalMessage-Edited",
    HEADER_ORIG_REFERENCES: "X-TotalMessage-Original-References",
    HEADER_ORIG_INREPLYTO: "X-TotalMessage-Original-In-Reply-To",
    HEADER_ORIG_NONE: "<X-TotalMessage-None>",

    startEditor: function(aEvent) {
      // Start the editor on dblclick on a header value, in fullheader mode.
      if (!TM.v.editingMode || aEvent.button != 0 )
        return;
TM.u.log.debug("TM.Editor.startEditor: START clicks: "+ aEvent.detail);

      let node = aEvent.target;
      let headerValue = node;
      while (headerValue && headerValue.classList &&
             !headerValue.classList.contains("headerValue"))
        headerValue = headerValue.parentNode;
      if (!headerValue.classList ||
          !headerValue.classList.contains("headerValue")) {
TM.u.log.debug("TM.Editor.startEditor: not in headerValue ");
        return;
      }

      if (TM.v.activeEditorCount)
        // Only have one active editor; the open editor header must be
        // 'updated' first before others can be edited.
        return;

      let headerRow = headerValue.parentNode;
      let headerLabel = headerValue.previousSibling.previousSibling;
      let headerNameLC = headerLabel.textContent.toLowerCase();
      if (headerValue.classList.contains("pendingAdd"))
        // Real label is previous hidden node, not the editor node.
        headerLabel = headerLabel.previousSibling;
      let textarea;

      if (headerRow.classList.contains("editing")) {
TM.u.log.debug("TM.Editor.startEditor: editing - "+headerLabel.textContent);
        // Already in an editor.
        textarea = headerValue.getElementsByClassName("editorTextarea")[0];
        if (headerValue.classList.contains("pendingUpdate") ||
            headerValue.classList.contains("pendingAdd") ||
            headerValue.classList.contains("pendingDelete")) {
TM.u.log.debug("TM.Editor.startEditor: pending - "+headerLabel.textContent);
          // Already in an editor with pending changes.
          headerRow.classList.remove("pending");
          headerValue.classList.remove("pendingUpdate");
          headerLabel.classList.remove("pendingDelete");
          headerValue.classList.remove("pendingDelete");
          textarea.removeAttribute("readonly");
          textarea.setSelectionRange(textarea.textLength, textarea.textLength);
          TM.v.activeEditorCount++;
          textarea.blur();
          if (headerValue.classList.contains("pendingAdd")) {
            headerValue.classList.add("pendingNew");
            headerValue.classList.remove("pendingAdd");
            let editor = headerRow.getElementsByClassName("editorName")[0];
            let textareaLabel = editor.getElementsByTagName("textarea")[0];
            textareaLabel.removeAttribute("readonly");
            headerLabel.classList.add("pendingNew");
            headerLabel.classList.remove("pendingAdd");
          }
        }
      }
      else {
TM.u.log.debug("TM.Editor.startEditor: new editor - "+headerLabel.textContent);
        // Start a new editor for this header.
        let editor = TM.u.c(TM.u.messageHeader, "headerEditorTemplate")
                       .cloneNode(true);
        editor.id = "";
        textarea = editor.getElementsByTagName("textarea")[0];
        let curheader = headerValue.firstChild.firstChild;
        textarea.textContent = curheader.textContent;
        headerValue.appendChild(editor);
        textarea.setAttribute("rows", "1");
        headerRow.classList.add("editing");
        TM.v.activeEditorCount++;
        let delButton = editor.getElementsByClassName("editorButtonDelete")[0];
        if (this.MIN_ONE_FIELDS.indexOf(headerNameLC) != -1)
          delButton.setAttribute("disabled", true);
        else
          delButton.removeAttribute("disabled");
      }

      textarea.focus();
    },

    closeHeaderEditorInfo: function() {
      let button = TM.u.c(TM.u.messageHeader, "editingInfoButton");
      button.checked = false;
      this.onHeaderEditorInfo(button);
    },

    onHeaderEditorInfo: function(aButton) {
TM.u.log.debug("TM.Editor.onHeaderEditorInfo: START "+aButton.id);
      let p = TM.u.g("editingInfo");
      let mpBox = TM.u.g("messagepaneboxwrapper").boxObject;
      let winWidth = TM.u.g("messagepaneboxwrapper").ownerDocument.width;
      let top = mpBox.y + aButton.parentNode.clientHeight * 1.5;
      let left = mpBox.x < p.clientWidth / 2            ? 10 :
                 mpBox.x + p.clientWidth / 2 > winWidth ? winWidth - p.clientWidth - 20 :
                                                          mpBox.x - p.clientWidth / 2;
      if (aButton.checked)
        p.setAttribute("style", "top: " + top + "px; " + "left: " + left + "px;");
      else
        p.setAttribute("style", "left: " + left + "px;");
    },

    onHeaderEditorCommand: function(aEvent, aCommand) {
TM.u.log.debug("TM.Editor.onHeaderEditorCommand: START "+aCommand);
      // First see if it's an editing row command.
      if (aCommand == "exit" || aCommand == "done") {
        TM.u.g("editorActive").setAttribute("checked", false);
        TM.Editor.closeHeaderEditorInfo();
        TM.v.editingMode = false;
        TM.v.mouseDragDropMode = false;
        if (TM.v.keyDragDropMode) {
          if (aCommand == "exit" && (TM.v.Editor.dropRow == null ||
              TM.v.Editor.dragRow == TM.v.Editor.dropRow)) {
            // This is from a cancel only; ensure original row is selected.
            gFolderDisplay.treeSelection.select(gDBView.currentlyDisplayedMessage);
          }
          TM.Editor.onKeyDragEnd();
          TM.v.keyDragDropMode = false;
        }
        TM.v.Editor = {};
        TM.v.activeEditorCount = 0;
        TM.u.g("EnableHeaderEditModeMenuItem").removeAttribute("checked");
        TM.List.Sorting = false;
        GetThreadTree().treeBoxObject.invalidate();

        TM.resetHeaderFlags();
        TM.Headers.headerIndex = TM.Editor.saveHeaderIndex;
//TM.u.log.debug("TM.Editor.onHeaderEditorCommand: saveHeaderIndex - "+TM.Editor.saveHeaderIndex);
        TM.Headers.toggle("select");

        return;
      }
      if (aCommand == "save") {
        // Collect the headers from the UI.
        this.createEditorChanges();

        // Write the new headers message to a temp .eml file.
        this.writeEditorChanges();

        // Replace current message with the new headers message temp .eml file;
        // args are for a single message interactive Editor mode transaction.
        let args = {};
        // Backup original in trash if false (button checked).
        args.deleteStorage = TM.v.Editor.editorButtonBackup.getAttribute("check") != "true";
        // Select the message.
        args.selectMsg = true;
        // The msgWindow.
        args.msgwin = TM.u.winType == "mail:messageWindow" ?
                        null : TM.u.topWindow.msgWindow;
        // Is current message being edited in an opened thread or an opened
        // thread root?
        args.openSelectedMsgThread = false;
        // Root thread of the message or the message if not threaded or the root.
        args.origThreadId = null;

        let mainWin = TM.u.mainWindow;
        let topWin = TM.u.topWindow;
        if (mainWin && mainWin.gDBView) {
          mainWin.gFolderDisplay.treeSelection.selectEventsSuppressed = true;
          let dbView = mainWin.gDBView;
          let index = dbView.findIndexOfMsgHdr(TM.v.msgHdr, false);
          let rootMsgKey = dbView.getThreadContainingIndex(index).getChildHdrAt(0).messageKey;
          let rootIndex = dbView.findIndexFromKey(rootMsgKey, false);
          args.openSelectedMsgThread = dbView.isContainerOpen(index);
          args.flags = TM.v.msgHdr.flags;
          args.keywords = TM.v.msgHdr.getStringProperty("keywords");
          args.origThreadId = TM.v.msgHdr.threadId;
          args.origThreadParent = TM.v.msgHdr.threadParent;
          if (args.openSelectedMsgThread) {
            // Close an open root thread message.
            //dbView.toggleOpenState(index);
TM.u.log.debug("TM.Editor.onHeaderEditorCommand: close open root, index:rootIndex:rootMsgKey - "+
               index+":"+rootIndex+":"+rootMsgKey);
            GetThreadTree().changeOpenState(rootIndex, false)
          }
          if (TM.u.winType == "mail:messageWindow") {
            if (mainWin.TM.v.msgHdr &&
                topWin.TM.v.msgHdr.messageKey != mainWin.TM.v.msgHdr.messageKey)
              // If standalone window message not the same as 3pane, no select.
              args.selectMsg = false;
          }
          else
            mainWin.gFolderDisplay.treeSelection.clearSelection();
        }
        // Do it.
        this.replaceEditorChanges(TM.v.msgHdr, args);
        this.closeHeaderEditorInfo();
        return;
      }

      // Check header editor command.
      let headerValue = aEvent ? aEvent.target : null;
      while (headerValue && !headerValue.classList.contains("headerValue"))
        headerValue = headerValue.parentNode;
      if (!headerValue || !headerValue.classList.contains("headerValue")) {
TM.u.log.debug("TM.Editor.onHeaderEditorCommand: no target or headerValue ");
        return;
      }

      let headerRow = headerValue.parentNode;
      let headerLabel = headerValue.previousSibling.previousSibling;
      let headerName;

      let editor = headerValue.getElementsByClassName("editor")[0];
      let textarea = editor.getElementsByTagName("textarea")[0];
      let curheader = headerValue.firstChild.firstChild;
      switch (aCommand) {
        case "update":
          if (textarea.value == curheader.textContent) {
            // If no change, like a cancel.
            if (headerValue.classList.contains("pendingNew")) {
              // Remove the new row if add.
              headerValue.parentNode.remove();
            }
            else {
              editor.remove();
              headerRow.classList.remove("editing");
            }
          }
          else {
            textarea.setAttribute("readonly", true);
            headerRow.classList.add("pending");
            if (headerValue.classList.contains("pendingNew")) {
              // For a new header if add; get the label editor.
              editor = headerRow.getElementsByClassName("editorName")[0];
              let textareaLabel = editor.getElementsByTagName("textarea")[0];
              headerName = textareaLabel.value;
              let headerNameLC = headerName.toLowerCase();
              if (!headerName)
                // No header name entered.
                headerValue.parentNode.remove();
              else if (this.MAX_ONE_FIELDS.indexOf(headerNameLC) != -1 &&
                       headerNameLC in TM.v.Editor.fileHeader) {
                // No dupes for this name.
                textareaLabel.value = "";
                textareaLabel.focus();
                return;
              }
              else {
                headerValue.classList.remove("pendingNew");
                headerValue.classList.add("pendingAdd");
                textareaLabel.setAttribute("readonly", true);
                // Real label is previous hidden node.
                headerLabel = headerLabel.previousSibling;
                headerLabel.classList.remove("pendingNew");
                headerLabel.classList.add("pendingAdd");
                headerLabel.textContent = headerName;
              }
            }
            else
              headerValue.classList.add("pendingUpdate");
          }
          TM.v.activeEditorCount--;
        break;
        case "undo":
TM.u.log.debug("TM.Editor.onHeaderEditorCommand: undo current:prior "+
               textarea.value+":"+curheader.textContent);
          textarea.value = curheader.textContent;
          textarea.focus();
        break;
        case "delete":
          if (headerValue.classList.contains("pendingNew")) {
            headerValue.parentNode.remove();
          }
          else {
            textarea.setAttribute("readonly", true);
            headerRow.classList.add("pending");
            headerLabel.classList.add("pendingDelete");
            headerValue.classList.add("pendingDelete");
          }
          TM.v.activeEditorCount--;
        break;
        case "add":
          // Clone current header, append after current header, close editor
          // for current header.  Append the label editor and hide the header
          // name.
          let newHeaderRow = headerRow.cloneNode(true);
          headerRow.parentNode.insertBefore(newHeaderRow, headerRow.nextSibling)
          editor.remove();
          headerRow.classList.remove("editing");

          headerValue = newHeaderRow.getElementsByClassName("headerValue")[0];
          headerRow = headerValue.parentNode;
          headerLabel = headerRow.getElementsByClassName("headerLabel")[0];

          headerValue.firstChild.firstChild.textContent = "";
          editor = headerValue.getElementsByClassName("editor")[0];
          textarea = editor.getElementsByTagName("textarea")[0];
          textarea.textContent = "";
          headerValue.classList.add("pendingNew");
          headerValue.classList.remove("pendingUpdate");
          let addButton = editor.getElementsByClassName("editorButtonAdd")[0];
          addButton.setAttribute("disabled", true);

          headerLabel.textContent = "";
          headerLabel.setAttribute("hidden", true);
          let editorName = editor.getElementsByClassName("editorName")[0];
          headerRow.insertBefore(editorName, headerLabel.nextSibling)
          textarea = editorName.getElementsByTagName("textarea")[0];
          textarea.setAttribute("rows", "1");
          textarea.focus();
        break;
        case "cancel":
        default:
          if (headerValue.classList.contains("pendingNew")) {
            headerValue.parentNode.remove();
          }
          else {
            editor.remove();
            headerRow.classList.remove("editing");
          }
          TM.v.activeEditorCount--;
      }

      // If any change to be made, remove disable from Save button.
      let fullHeaderTable = TM.u.c(TM.u.messageHeader, "fullHeaderTable");
      let hasChanges = false;
      if (fullHeaderTable.getElementsByClassName("pendingUpdate").length)
        hasChanges = true;
      else if (fullHeaderTable.getElementsByClassName("pendingAdd").length)
        hasChanges = true;
      else if (fullHeaderTable.getElementsByClassName("pendingDelete").length)
        hasChanges = true;
      if (hasChanges)
        TM.v.Editor.editorButtonSave.removeAttribute("disabled");
      else
        TM.v.Editor.editorButtonSave.setAttribute("disabled", "true");
    },

    onKeyPressHeaderEditor: function(aEvent) {
TM.u.log.trace("TM.Editor.onKeyPressHeaderEditor: START char:val - "+
               aEvent.which+":"+aEvent.keyCode);
      // Per RFC5322, header name (email) chars must be printable us-ascii
      // chars d'33' to d'126' minus the colon d'58'.
      aEvent.stopPropagation();
      let charCode = aEvent.which;
      let validChar =
          charCode == 58 || charCode < 33 || charCode > 126 ? false : true;
      if (validChar || aEvent.keyCode)
        return;

      aEvent.preventDefault();
    },

    createEditorChanges: function() {
TM.u.log.debug("TM.Editor.createEditorChanges: START ");
      let fullHeaderTable = TM.u.c(TM.u.messageHeader, "fullHeaderTable");
      let headerRow, headerValue, line, label, labelLC, value;
      let newHeader = "";
      let headerRows = fullHeaderTable.getElementsByClassName("fullHeaderRow");

      let converter = TM.u.converter;
      converter.charset = TM.Editor.CharSet;

      if (TM.v.Editor.FromLine)
        newHeader = TM.v.Editor.FromLine + TM.Editor.CRLF;
      for (let headerRow of headerRows) {
        headerValue = headerRow.getElementsByClassName("headerValue")[0];
        if (!headerValue || headerValue.classList.contains("pendingDelete"))
          continue;

        if (headerValue.classList.contains("pendingUpdate") ||
            headerValue.classList.contains("pendingAdd"))
          value = headerValue.getElementsByTagName("textarea")[0].value;
        else
          value = headerValue.getElementsByClassName("container")[0].firstChild.
                  textContent;
        label = headerRow.getElementsByClassName("headerLabel")[0].textContent;
        labelLC = label.toLowerCase();
        // For folding, indent length of header name for nice alignment, make
        // sure first col never starts with a non blank when folding.
        let pad = " ".repeat(label.length + 2);

        // Need to encode per RFC2047. Only for email, don't pollute feeds/others.
        if (value.match(/[\u0080-\uFFFF]/) &&
            TM.u.MessageType(TM.v.msgHdr) == TM.v.HEADER.MSG_TYPE_MAIL) {
          // True for email addressing headers with <> enclosed addr-spec.
          let structured = TM.u.isStructuredAddressingHeader(labelLC, value);
          // NOTE: If we had an encoding coming in, use it to encode back - deprecated.
          // Now, always encode in UTF-8 regardless of original encoding,
          // despite RFC6532 some clients still requiring RFC2047.
//          let charSet = headerValue.getAttribute("charSet");
//          charSet = charSet || TM.Editor.CharSet;
          let wordSize = Ci.nsIMimeConverter.MIME_ENCODED_WORD_SIZE; // 72
          value = MailServices.mimeConverter
                              .encodeMimePartIIStr_UTF8(value,
                                                        structured,
                                                        TM.Editor.CharSet, // charSet,
                                                        label.length,
                                                        wordSize);
          value = value.replace(/\r/g, TM.Editor.LF);
          value = value.replace(/\n\s*(\S)/gm, TM.Editor.CRLF + pad + "$1");

TM.u.log.debug("TM.Editor.createEditorChanges: mime header:origCharset:isStructured:value - "+
               labelLC+":"+charSet+":"+structured+":"+value);
TM.u.log.debug("TM.Editor.createEditorChanges: mime deMime - "+
               GlodaUtils.deMime(value));
        }
        else {
          value = converter.ConvertFromUnicode(value);

          if (labelLC == "x-mozilla-keys") {
            // Special handling since Tb preallocates space, so we pad out to
            // this amount even if empty or partially filled.
            value = value.trimRight();
            value = value + " ".repeat((80 - value.length) < 0 ? 0 : 80 - value.length);
          }
          else {
            value = value.replace(/\n\s*(\S)/gm, TM.Editor.CRLF + pad + "$1");
          }
        }

        line = label + ": " + value + TM.Editor.CRLF;
        newHeader += line;
      }

      TM.v.Editor["newMessage"] = newHeader + TM.v.Editor.fileMsgBody;
      TM.v.Editor.fileMsgBody = "";
TM.u.log.debug("TM.Editor.createEditorChanges: newHeader -\n"+newHeader);
    },

    writeEditorChanges: function() {
TM.u.log.debug("TM.Editor.writeEditorChanges: START ");
      let tmpFile = FileUtils.getFile("TmpD", [], true);
      tmpFile.append(this.TMP_FILENAME);
      if (tmpFile.exists())
        tmpFile.remove(true);
      tmpFile.create(tmpFile.NORMAL_FILE_TYPE, FileUtils.PERMS_FILE);
      let fos = FileUtils.openSafeFileOutputStream(tmpFile, FileUtils.MODE_WRONLY);
      fos.write(TM.v.Editor.newMessage, TM.v.Editor.newMessage.length);
      FileUtils.closeSafeFileOutputStream(fos);
      TM.v.Editor.newMessage = "";
TM.u.log.debug("TM.Editor.writeEditorChanges: file - "+tmpFile.path);
    },

    replaceEditorChanges: function(aMsgHdr, aArgs) {
      let msgHdr = aMsgHdr;
      let Args = aArgs;
TM.u.log.debug("TM.Editor.replaceEditorChanges: START flags:keys:args - "+
               msgHdr.flags+":"+
               msgHdr.getStringProperty("keywords")+":"+
               Args.toSource());
      if (msgHdr.folder.locked) {
        TM.u.log.info("TM.Editor.replaceEditorChanges: folder is busy, " +
                      "try again later.");
        return this.onHeaderEditorCommand(null, "exit");
      }
      // Get the previously created tmpFile, cannot pass tmpFile.
      let tmpFile = FileUtils.getFile("TmpD", [this.TMP_FILENAME], true);
      if (!tmpFile.exists())
        return this.onHeaderEditorCommand(null, "exit");
      Args.tmpFile = tmpFile;

      // Create a listener for delete message completion.
      let delListener = new this.copyServiceListener("delete", msgHdr, Args);

      // Create a listener for copy message (import new headers message into
      // the folder from file) completion.
      let copyListener = new this.copyServiceListener("copy", msgHdr, Args);
      delListener.copyListener = copyListener;

      // Set up delete.
      let msgHdrArray  = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
      msgHdrArray.appendElement(msgHdr, false);
      // Don't select on delete.
      gFolderDisplay._nextViewIndexAfterDelete = -1;

      // It seems listeners are called multiple times for maildir and Local
      // Folders when using deleteMessages, so don't use one.
      let storeType = msgHdr.folder.msgStore.storeType;
      let listener = (aArgs.deleteStorage && !TM.v.isImap) ||
                       storeType == "maildir" ? null : delListener;

      // Do the delete before the copy.
      gFolderDisplay.hintAboutToDeleteMessages();
      msgHdr.folder.deleteMessages(msgHdrArray,
                                   aArgs.msgwin,
                                   aArgs.deleteStorage,
                                   true, // isMove
                                   listener,
                                   false); // allowUndo

      if (!listener) {
TM.u.log.debug("TM.Editor.replaceEditorChanges: no listener");
        // If not copying to trash, no notification, so run manually; IMAP
        // handles deletes per account prefs. Also, run manually for maildir.
        // TODO: temporarily change account delete model pref to MoveToTrash
        // to handle the delete per button toggle state.  If MarkDeleted or
        // RemoveImmediately are selected, results are not correct; there
        // are longstanding core bugs 399475/370509 on delete model handling.
        // Also 651945 for gmail.
        delListener.OnStopCopy(0);
      }

      return true;
    },

    messageEditFinished: function(aMsgFolder, aArgs) {
      TM.v.editingMode = false;
      let mainWin = TM.u.mainWindow;
      let topWin = TM.u.topWindow;
      let newMsgHdr = aMsgFolder.GetMessageHeader(aArgs.msgKey);
TM.u.log.debug("TM.Editor.messageEditFinished: subject:messageKey:threadParent:threadId:flags:keywords:args - "+
               (newMsgHdr.subject || newMsgHdr.mime2DecodedSubject) +":"+
               newMsgHdr.messageKey +":"+
               newMsgHdr.threadParent +":"+
               newMsgHdr.threadId +":"+
               newMsgHdr.flags +":"+
               newMsgHdr.getStringProperty("keywords") +":"+
               aArgs.toSource());
      let newThreadParent =
          newMsgHdr.threadParent != this.nsMsgKey_None ? newMsgHdr.threadParent :
                                                         aArgs.msgKey;
      let newThreadParentIndex =
          mainWin.gDBView.findIndexFromKey(newThreadParent, true);
      let newThreadRoot = mainWin.gDBView.
                                  getThreadContainingIndex(newThreadParentIndex).
                                  getChildHdrAt(0).
                                  messageKey;
TM.u.log.debug("TM.Editor.messageEditFinished: newThreadRoot key - "+newThreadRoot);
      newThreadRoot = mainWin.gDBView.findIndexFromKey(newThreadRoot, false);
TM.u.log.debug("TM.Editor.messageEditFinished: newThreadRoot idx - "+newThreadRoot);

      let origThreadRoot;
      let origThreadParent =
          aArgs.origThreadParent != this.nsMsgKey_None ? aArgs.origThreadParent :
                                                         null;
      if (origThreadParent) {
        let origThreadParentIndex = mainWin.gDBView.findIndexFromKey(origThreadParent, false);
        origThreadRoot = mainWin.gDBView
                                .getThreadContainingIndex(origThreadParentIndex)
                                .getChildHdrAt(0)
                                .messageKey;
TM.u.log.debug("TM.Editor.messageEditFinished: origThreadRoot key - "+origThreadRoot);
        origThreadRoot = mainWin.gDBView.findIndexFromKey(origThreadRoot, false);
TM.u.log.debug("TM.Editor.messageEditFinished: origThreadRoot idx - "+origThreadRoot);
      }

//TM.u.log.debug("TM.Editor.messageEditFinished: origThreadParentIndex - "+origThreadParentIndex);
      if (mainWin &&
          (gDBView.viewFlags & Ci.nsMsgViewFlagsType.kThreadedDisplay) &&
          origThreadParent &&
          origThreadRoot != this.nsMsgKey_None &&
          mainWin.gDBView.isContainer(origThreadRoot) &&
          origThreadRoot != newThreadRoot) {
TM.u.log.debug("TM.Editor.messageEditFinished: toggle orig thread - "+origThreadRoot);
        // Toggle original thread on unthread, if it is different from the new
        // thread, for redisplay of threadlines.
//        mainWin.gDBView.toggleOpenState(origThreadRoot);
//        mainWin.gDBView.toggleOpenState(origThreadRoot);
        GetThreadTree().changeOpenState(origThreadRoot);
        GetThreadTree().changeOpenState(origThreadRoot);
      }

//TM.u.log.debug("TM.Editor.messageEditFinished: newThreadParentIndex - "+newThreadParentIndex);
      if (mainWin &&
          (gDBView.viewFlags & Ci.nsMsgViewFlagsType.kThreadedDisplay) &&
          newThreadRoot != this.nsMsgKey_None &&
          mainWin.gDBView.isContainer(newThreadRoot) &&
          (mainWin.gDBView.isContainerOpen(newThreadRoot) ||
           aArgs.openSelectedMsgThread)) {
TM.u.log.debug("TM.Editor.messageEditFinished: toggle new thread - "+newThreadRoot);
        // If threaded message, toggle the root for correct redisplay.
//        mainWin.gDBView.toggleOpenState(newThreadRoot);
        GetThreadTree().changeOpenState(newThreadRoot);
      }

//      mainWin.gFolderDisplay.treeSelection.selectEventsSuppressed = false;

      if (aArgs.selectMsg && mainWin) {
TM.u.log.debug("TM.Editor.messageEditFinished: select msgKey - "+aArgs.msgKey);
        // Select the message.
        mainWin.gDBView.selectMsgByKey(aArgs.msgKey);
      }

      if (TM.u.winType == "mail:messageWindow") {
TM.u.log.debug("TM.Editor.messageEditFinished: messageWindow ");
        // Standalone window.
        // XXX: Too many issues: If reusing standalone window, edit of second
        // or more message opened in the window causes an unload event, closing
        // the window after deleting but not copying back the edited message.
        // If deleteStorage is false, ie retain copy in trash, the next message's
        // headers are delivered, despite the select.  Also true for tabs, but
        // edit remains enabled there.
//        TM.resetMsgLoadFlags();
//        TM.Editor.onHeaderEditorCommand(null, "exit");
//TM.u.log.debug("TM.Editor.messageEditFinished: loading msgKey - "+aArgs.msgKey);
        gDBView.selectMsgByKey(aArgs.msgKey);
TM.u.log.debug("TM.Editor.messageEditFinished: loaded msgKey - "+aArgs.msgKey);
//        if (newMsgHdr != mainWin.gFolderDisplay.selectedMessage)
          // Different selection in 3pane than standalone, don't select.
//          aArgs.selectMsg = false;
      }
//      else {
//        newMsgHdr.markRead(true);
//        mainWin.MarkCurrentMessageAsRead();
        TM.Editor.onHeaderEditorCommand(null, "done");
//      }

      gFolderDisplay.ensureSelectionIsVisible();
      mainWin.gFolderDisplay.treeSelection.selectEventsSuppressed = false;
    },

    getReferencesForThreading: function(aTargetRow) {
      let targetMsgKey = gDBView.getKeyAt(aTargetRow);
      let targetMsgHdr = gDBView.db.GetMsgHdrForKey(targetMsgKey);
      let targetMsgMessageId = targetMsgHdr.messageId;
TM.u.log.debug("TM.Editor.getReferencesForThreading: drop targetMsgMessageId - "+targetMsgMessageId);

      try {
        MsgHdrToMimeMessage(targetMsgHdr, null, function(aMsgHdr, aMimeMsg) {
          if (aMimeMsg == null) {
            TM.u.log.error("TM.Editor.setReferencesForThreading: " +
                           "could not get a mimeMsg.");
            return this.onHeaderEditorCommand(null, "exit");
          }

//TM.u.log.debug("TM.Editor.getReferencesForThreading: mime message: "+aMimeMsg.toSource());
          let targetHeaders = aMimeMsg.headers;
          // If no messageId in the eml file , use the one from the msf
          // database as assigned by Tb on receipt.
          let targetMessageId = targetHeaders["message-id"] ||
                                targetMsgMessageId;
          let targetReferences = targetHeaders["references"];
          let targetInReplyTo = targetHeaders["in-reply-to"];

TM.u.log.trace("TM.Editor.setReferencesForThreading: targetMessageId - "+targetMessageId);
TM.u.log.trace("TM.Editor.setReferencesForThreading: targetInReplyTo - "+targetInReplyTo);
TM.u.log.trace("TM.Editor.setReferencesForThreading: targetReferences - "+targetReferences);

          // Per rfc2822, references of the child message are done thusly
          // based on the parent's headers.
          let newReferences = targetReferences ? targetReferences + " " +
                                                 targetMessageId :
                              targetInReplyTo ? targetInReplyTo + " " +
                                                targetMessageId :
                              targetMessageId ? targetMessageId :
                              null;

          if (!newReferences) {
            TM.u.log.warn("TM.Editor.setReferencesForThreading: " +
                          "new parent has no messageid, inreplyto, or references.");
            return this.onHeaderEditorCommand(null, "exit");
          }

          return TM.Editor.setReferencesForThreading(newReferences);
        },
        // imapAllowDownload, {saneBodySize, imapPartsOnDemand}
        true, {saneBodySize: true, partsOnDemand: false});
      }
      catch(ex) {
        TM.u.log.error("TM.Editor.setReferencesForThreading: " +
                       "could not get a mimeMsg. Error: " + ex);
      }

      return true;
    },

    setReferencesForThreading: function(aNewReferences) {
TM.u.log.debug("TM.Editor.setReferencesForThreading: aNewReferences - "+aNewReferences);
      let fullHeaderTable = TM.u.c(TM.u.messageHeader, "fullHeaderTable");

      let messageIdRow = fullHeaderTable.getElementsByClassName("message-id");
      messageIdRow = messageIdRow.length ? messageIdRow[0] : null;

      let referencesRow = fullHeaderTable.getElementsByClassName("references");
      referencesRow = referencesRow.length ? referencesRow[0] : null;
      let referencesOrigRow = fullHeaderTable.
          getElementsByClassName(this.HEADER_ORIG_REFERENCES.toLowerCase());
      referencesOrigRow = referencesOrigRow.length ? referencesOrigRow[0] : null;

      let inreplytoRow = fullHeaderTable.getElementsByClassName("in-reply-to");
      inreplytoRow = inreplytoRow.length ? inreplytoRow[0] : null;
      let inreplytoOrigRow = fullHeaderTable.
          getElementsByClassName(this.HEADER_ORIG_INREPLYTO.toLowerCase());
      inreplytoOrigRow = inreplytoOrigRow.length ? inreplytoOrigRow[0] : null;

      let newReferences = aNewReferences;

      if (!newReferences) {
        // If no newReferences, delete (to unthread) the references field
        // by removing the references editor node.  Must also remove
        // in-reply-to header to unthread.  Further, the only way to not hit
        // Bug 604319 is to rename the messageId of the target, as any messages
        // with nested references to the messageId will keep it threaded.  This
        // also unthreads all children from a root thread, making it no one's
        // parent.
        TM.Editor.resetMessageId(messageIdRow);

//        if (!referencesRow && !inreplytoRow)
          // No references or inreplyto header.
//          return this.onHeaderEditorCommand(null, "exit");

        if (referencesRow) {
          if (referencesOrigRow)
            referencesRow.remove();
          else
            referencesRow.getElementsByClassName("headerLabel")[0]
                         .textContent = this.HEADER_ORIG_REFERENCES;
        }
        if (inreplytoRow) {
          if (inreplytoOrigRow)
            inreplytoRow.remove();
          else
            inreplytoRow.getElementsByClassName("headerLabel")[0]
                        .textContent = this.HEADER_ORIG_INREPLYTO;
        }
      }
      else {
        if (gDBView.currentlyDisplayedMessage < gDBView.rowCount - 1 &&
            gDBView.getLevel(gDBView.currentlyDisplayedMessage) <
            gDBView.getLevel(gDBView.currentlyDisplayedMessage + 1))
          // A message that is a parent (its messageId is a reference for child
          // messages) needs to have the messageId reset for edge rethread cases,
          // such as rethreading it to one of its descendants.
          TM.Editor.resetMessageId(messageIdRow);

        let origReferences;
        if (referencesRow && "references" in TM.v.Editor.fileHeader) {
          // Save original references.
          if (!referencesOrigRow)
            origReferences = referencesRow.getElementsByTagName("a")[0].textContent;

          // Overwrite existing references header to rethread to new parent.
          referencesRow.getElementsByTagName("a")[0].textContent = newReferences;
        }
        else {
          if (!referencesOrigRow)
            // No original references.
            origReferences = this.HEADER_ORIG_NONE;

          // Add new references header.
          referencesRow = fullHeaderTable.getElementsByClassName("fullHeaderRow");
          if (!referencesRow || !referencesRow[0]) {
            TM.u.log.error("TM.Editor.setReferencesForThreading: " +
                           "no fullHeaderRow to clone!");
            return this.onHeaderEditorCommand(null, "exit");
          }

          referencesRow = referencesRow[0].cloneNode(true);
          referencesRow.getElementsByClassName("headerLabel")[0]
                       .textContent = "References";
          referencesRow.getElementsByTagName("a")[0]
                       .textContent = newReferences;
          fullHeaderTable.appendChild(referencesRow);
        }

        if (!referencesOrigRow) {
          // Save original references header or none, first time only.
          referencesOrigRow = fullHeaderTable.firstElementChild.cloneNode(true);
          referencesOrigRow.getElementsByClassName("headerLabel")[0]
                           .textContent = this.HEADER_ORIG_REFERENCES;
          referencesOrigRow.getElementsByTagName("a")[0]
                           .textContent = origReferences;
          fullHeaderTable.appendChild(referencesOrigRow);
        }
      }

      // Save em to file.
      return this.onHeaderEditorCommand(null, "save");
    },

    resetMessageId: function(messageIdRow) {
      if (messageIdRow) {
        // Rename a messageId by prepending 'X-TH#' and a count.
        let messageIdTargetTag = messageIdRow.getElementsByTagName("a")[0];
        let messageIdTarget = messageIdTargetTag.textContent;
        let editCount = messageIdTarget.match(/^<X-TH#(\d+)\..*/);
TM.u.log.debug("TM.Editor.resetMessageId: messageIdTarget:editCount - " +
               messageIdTarget+":"+ editCount);
        if (editCount) {
          editCount = parseInt(editCount[1]) + 1;
          messageIdTarget = messageIdTarget.replace(/(^<X-TH#)(\d+)(\..*)/,
                                                    "$1" + editCount + "$3");
        }
        else
          messageIdTarget = messageIdTarget.replace(/^<(.*)/, "<X-TH#1.$1");

        messageIdTargetTag.textContent = messageIdTarget;
      }
      else {
        // Since Message-ID is 'should', it doesn't need to be there.  And if
        // not then it's not an issue.
        TM.u.log.warn("TM.Editor.resetMessageId: no Message-ID header.");
      }
    },

    unThreadMessage: function() {
      if (TM.v.editingMode && TM.v.keyDragDropMode)
        TM.Editor.setReferencesForThreading(null);
    },

    getMessageStream: function(msgHdr, aCallback) {
      let listener = {
        mStream : null,
        mFileMsg: "",
        mCallback: aCallback,
        QueryInterface: XPCOMUtils.generateQI([Ci.nsISupports,
                                               Ci.nsIStreamListener]),

        onStartRequest: function (aRequest, aContext) {
TM.u.log.debug("TM.Editor.getMessageStream.onStartRequest: START" );
          this.mStream = Cc['@mozilla.org/binaryinputstream;1'].
                         createInstance(Ci.nsIBinaryInputStream);
          let channel = aRequest.QueryInterface(Ci.nsIChannel);
          channel.URI.QueryInterface(Ci.nsIMsgMailNewsUrl);
          channel.URI.msgHeaderSink = null; //this
        },

        onStopRequest: function (aRequest, aContext, aStatusCode) {
          this.mStream = null;
          TM.Editor.processMessageStream(this.mFileMsg, this.mCallback);
          this.mFileMsg = null;
        },

        onDataAvailable: function (aRequest, aContext, aInputStream,
                                   aOffset, aCount) {
          this.mStream.setInputStream(aInputStream);
          let chunk = this.mStream.readBytes(aCount);
          this.mFileMsg += chunk;
        }
      };

      let messenger = Cc["@mozilla.org/messenger;1"].createInstance(Ci.nsIMessenger);
      let uri = msgHdr.folder.getUriForMsg(msgHdr);
      let messengerSvc = messenger.messageServiceFromURI(uri)
                                  .QueryInterface(Ci.nsIMsgMessageService);
      messengerSvc.DisplayMessage(uri, listener, null, null, null, {});
TM.u.log.debug("TM.Editor.getMessageStream.DisplayMessage: START "+uri);
    },

    processMessageStream: function(aFileMsg, aCallback) {
      // If the content has a CRLF to start a line, it uses CR.  This avoids
      // content with no CRLF for lines, but with random CRLF in the doc.
      // This should also take care of Imap servers using CR.
      TM.Editor.CR = aFileMsg.match(/^\r\n/gm) ? "\r" : "";
      TM.Editor.CRLF = TM.Editor.CR + TM.Editor.LF;

TM.u.log.debug("TM.Editor.processMessageStream: START: "+
               (TM.Editor.CR ? "has CR" : "no CR"));

      let headersEnd = aFileMsg.indexOf(TM.Editor.CR + "\n" +
                                        TM.Editor.CR + "\n") +
                                        TM.Editor.CR.length + 1;
      TM.v.Editor["fileMsgHeader"] = aFileMsg.substr(0, headersEnd);
      TM.v.Editor["fileMsgBody"] = aFileMsg.substring(TM.v.Editor.fileMsgHeader.length);
TM.u.log.debug("TM.Editor.processMessageStream: headers:\n"+
               TM.v.Editor.fileMsgHeader+".");
TM.u.log.trace("TM.Editor.processMessageStream: body:\n"+
               TM.v.Editor.fileMsgBody+".");

      let charSet = TM.v.msgHdr.Charset;
      if (!charSet) {
        // No charset for multipart msgHdr, find it in the body.
        charSet = TM.v.Editor.fileMsgBody.split(/charset=/i)[1];
        charSet = charSet ? charSet.split(/[,;\s\"\']/)[0] : TM.Editor.CharSet;
      }

TM.u.log.debug("TM.Editor.processMessageStream: charSet - "+ charSet);
      let converter = TM.u.converter;
      converter.charset = charSet;

      // Create object like currentHeader, to build the fullheaders.  We need
      // to build this from file since the mime emitter does not send blank
      // values (ex subject) and those should be kept in the db.
      let fKey = "empty", fName, fValue, sepIndex, unfoldedValue;
      TM.v.Editor.fileHeader = {"empty":{"headerValue":" ", "headerName":" "}};
      let fileHeaders = TM.v.Editor.fileMsgHeader.split(TM.Editor.CRLF);
      fileHeaders.forEach(function(header, index) {
TM.u.log.debug("TM.Editor.processMessageStream: header - "+header);
        header = header.replace(/\r/gm, "");
        if (header.substr(0, 6) == "From -")
          TM.v.Editor["FromLine"] = header;
        else {
          if (header.substr(0, 1) == " " || header.substr(0, 1) == "\t")
            TM.v.Editor.fileHeader[fKey].headerValue += (TM.Editor.LF + header);
          else {
            // This header is a new header, so the prior has completed being
            // constructed, check the prior's unfolded value here.  The bogus
            // "empty" template makes it work for the first header.
            unfoldedValue = TM.v.Editor.fileHeader[fKey].headerValue;
            let mimeCharSet = /\=\?(\S+)\?[BQ]\?\S+\?\=/.exec(unfoldedValue);
            if (mimeCharSet) {
              // Have an RFC2047 mime header.
TM.u.log.debug("TM.Editor.processMessageStream: RFC2047 charset:unfolded val - "+
                mimeCharSet[1]+":"+unfoldedValue);

              fValue = GlodaUtils.deMime(unfoldedValue);
              TM.v.Editor.fileHeader[fKey].headerValue = fValue;
              TM.v.Editor.fileHeader[fKey].charSet = mimeCharSet[1];
//TM.u.log.debug("TM.Editor.processMessageStream: RFC2047 demimed val -\n"+fValue);
            }
            else {
              fValue = converter.ConvertToUnicode(unfoldedValue);
              TM.v.Editor.fileHeader[fKey].headerValue = fValue;
              TM.v.Editor.fileHeader[fKey].charSet = charSet;
            }

TM.u.log.debug("TM.Editor.processMessageStream: fKey:unfolded:converted header -\n"+
               fKey+":"+unfoldedValue+":"+fValue);

            // Continue with new header.
            sepIndex = header.indexOf(":");
            if (sepIndex > -1) {
              fKey = header.substr(0, sepIndex).toLowerCase();
              fKey = fKey in TM.v.Editor.fileHeader ? fKey + (index++) : fKey;
              fName = header.substr(0, sepIndex);
              fValue = header.substr(sepIndex + 1).trimLeft();
              fValue = fValue ? fValue : " ";
              TM.v.Editor.fileHeader[fKey] = { headerValue: fValue,
                                               headerName: fName };
            }
          }
        }
      }, this);

      // Add tags placeholder.
//      TM.v.Editor.fileHeader["tags"] = {"headerValue":"", "headerName":"tags"};
      delete TM.v.Editor.fileHeader.empty;
TM.u.log.debug("TM.Editor.processMessageStream: fileHeader - "+
               TM.v.Editor.fileHeader.toSource());

      // The callback saves header toggle state, shows full headers.
      if (aCallback)
        aCallback();
    },

    copyServiceListener: function(aWho, aMsgHdr, aArgs) {
      let copyServiceListener = {
        QueryInterface: XPCOMUtils.generateQI([Ci.nsISupports,
                                               Ci.nsIUrlListener,
                                               Ci.nsIMsgCopyServiceListener]),
        // nsIUrlListener
        OnStartRunningUrl: function () {
          TM.u.log.debug("TM.Editor.copyServiceListener.OnStartRunningUrl - "+this.who);
        },
        OnStopRunningUrl: function () {
          // This notification is sent for imap.
          TM.u.log.debug("TM.Editor.copyServiceListener.OnStopRunningUrl - "+this.who);
          if (this.who == "delete")
            // If deleteStorage is true.
            this.OnStopCopy(0);
          if (this.who == "copy") {
            // For imap, CopyFileMessage is a fail for tags, so add them now.
            if (this.Args.keywords) {
              let newMsgHdr = this.msgHdrOriginal.folder.GetMessageHeader(aArgs.msgKey);
              let msgArray = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
              msgArray.appendElement(newMsgHdr, false);
              this.msgHdrOriginal.folder.addKeywordsToMessages(msgArray, this.Args.keywords);
            }

            TM.Editor.messageEditFinished(this.msgHdrOriginal.folder, this.Args)
          }
        },

        // nsIMsgCopyServiceListener
        OnProgress: function (progress, progressMax) { },
        OnStartCopy: function () {
          TM.u.log.debug("TM.Editor.copyServiceListener.OnStartCopy - "+this.who);
        },
        OnStopCopy: function (status) {
          TM.u.log.debug("TM.Editor.copyServiceListener.OnStopCopy: "+
                         this.who+":"+status);
          if (status != 0) {
            TM.u.log.error("TM.Editor.copyServiceListener.OnStopCopy: "+
                           this.who+":"+status);
            return this.onHeaderEditorCommand(null, "exit");
          }

          if (this.who == "copy")
            return true;

          if (this.who == "delete" || this.who == "compact") {
TM.u.log.debug("TM.Editor.copyServiceListener.OnStopCopy: CopyFileMessage -"+
               " Args: "+this.Args.toSource()+
               " folder: "+this.msgHdrOriginal.folder.name);
            MailServices.copy.CopyFileMessage(this.Args.tmpFile,
                                              this.msgHdrOriginal.folder,
                                              null, // this.msgHdrOriginal, //
                                              false, // isDraftOrTemplate
                                              this.Args.flags,
                                              this.Args.keywords,
                                              this.copyListener, //null, //
                                              this.Args.msgwin);
          }

          return true;
        },
        SetMessageKey: function (key) {
TM.u.log.debug("TM.Editor.copyServiceListener.SetMessageKey: "+ this.who+":"+key);
          if (this.who != "copy")
            return true;

          this.Args.msgKey = key;
          if (this.msgHdrOriginal.folder.flags & Ci.nsMsgFolderFlags.ImapBox) {
TM.u.log.debug("TM.Editor.copyServiceListener.SetMessageKey: IMAP - return");
            return true;
          }

TM.u.log.debug("TM.Editor.copyServiceListener.SetMessageKey: non IMAP");
          return TM.Editor.messageEditFinished(this.msgHdrOriginal.folder,
                                               this.Args)
        }
      };

      copyServiceListener.who = aWho;
      copyServiceListener.msgHdrOriginal = aMsgHdr;
      copyServiceListener.Args = aArgs;
      return copyServiceListener;
    },

    // Observer for threadPaneDND.  Attach our dnd system to the threadCol,
    // threadPane is hardcoded to not support canDrop().
    Observer: {
      observe: function(aMsgFolder, aTopic, aData) {
        if (gDBView)
          gDBView.addColumnHandler("threadCol", TM.Editor.threadPaneDND);
      }
    },

    threadPaneDND: {
      useRowProperties:    true,

      isEditable:          function(aRow, aCol)      { return false; },
//      canDrop:             function(row, orient)     { return true; },
      getImageSrc:         function(row, col)        {},
      getCellProperties:   function(row, col)        { return "" },
      getCellText:         function(row, col, value) { return value; },
      isString:            function()                { return true; },
      getSortStringForRow: function(msgHdr)          { return null; },
      getSortLongForRow:   function(msgHdr)          { return null; },
      getRowProperties:    function(row)             {
        if (!this.useRowProperties)
          return null;
        return this.getProperties(row);
      },

      getProperties: function(aRow) {
        if (!TM.v.editingMode)
          return null;
        if (aRow == TM.v.Editor.dropRow) {
TM.u.log.trace("TM.Editor.threadPaneDND: row:dropRow - "+aRow+":"+TM.v.Editor.dropRow);
          return "droppable";
        }
        if (TM.v.Editor.dragRow != null) {
          // Opening/closing folders changes the drag row index..
          TM.v.Editor.dragRow = gDBView.findIndexOfMsgHdr(TM.v.msgHdr, false);
          if (aRow == TM.v.Editor.dragRow) {
TM.u.log.trace("TM.Editor.threadPaneDND: row:dragRow - "+aRow+":"+TM.v.Editor.dragRow);
            return "indrag";
          }
        }
        if (TM.v.keyDragDropMode && aRow == gFolderDisplay.treeSelection.currentIndex) {
TM.u.log.trace("TM.Editor.threadPaneDND: row:keyDropRow - "+
               aRow+":"+gFolderDisplay.treeSelection.currentIndex);
          return "droppable";
        }
        return null;
      },

      getTargetRow: function(aEvent, aRow) {
        let row = aRow;
        if (aEvent instanceof Ci.nsIDOMMouseEvent)
          row = GetThreadTree().treeBoxObject.getRowAt(aEvent.clientX, aEvent.clientY);
        return row;
      },
      getCurRow: function(aRow) {
        let row = aRow;
        if (!row)
          row = gFolderDisplay.treeSelection.currentIndex;
        return row;
      },
      getCurRowParent: function(aRow) {
        let row = aRow;
        if (!(row >= 0 && row < gDBView.rowCount))
          row = null;
        return gDBView.getParentIndex(row);
      }
    },

    onKeyDragEnd: function() {
      if (TM.v.Editor.dropRow != null || TM.v.Editor.dragRow != null) {
TM.u.log.debug("TM.Editor.onKeyDragEnd: START dragRow:dropRow - "+ TM.v.Editor.dragRow+":"+TM.v.Editor.dropRow);
        TM.v.Editor.dragRow = null;
        TM.v.Editor.dropRow = null;
        GetThreadTree().removeAttribute("inkeydragdrop");
        gFolderDisplay.treeSelection.tree.clearStyleAndImageCaches();
        GetThreadTree().treeBoxObject.invalidate();
        gFolderDisplay.treeSelection.selectEventsSuppressed = false;
      }
    },

    enableKeyDragDropInThreadpane: function(aKeyDragDropMode) {
      if (!TM.Headers.isEditable())
        return false;

      if (!TM.v.editingMode)
        // Enable editing mode on key dnd start, if not enabled.
        TM.Headers.enableEditMode(true);

      if (!aKeyDragDropMode) {
        // Start key dnd.
        TM.v.Editor.dragRow = gDBView.currentlyDisplayedMessage;
        TM.v.Editor.dropRow = null;
        GetThreadTree().setAttribute("inkeydragdrop", true);
        GetThreadTree().focus();
        gFolderDisplay.treeSelection.tree.clearStyleAndImageCaches();
        GetThreadTree().treeBoxObject.invalidate(); //Row(TM.v.Editor.dragRow);
        gFolderDisplay.treeSelection.selectEventsSuppressed = true;
TM.u.log.debug("TM.Editor.enableKeyDragDropInThreadpane: START - "+TM.v.Editor.dragRow);
      }
      else {
        // On finish, do the move if drag row is not the drop row or the drag
        // row parent is not the drop row, else exit.
        TM.v.Editor.dropRow = this.threadPaneDND.getCurRow();
        let dragRowParent = this.threadPaneDND.getCurRowParent(TM.v.Editor.dragRow);
TM.u.log.debug("TM.Editor.enableKeyDragDropInThreadpane: STOP dragRow:dropRow:dragRowParent - "+
               TM.v.Editor.dragRow+":"+TM.v.Editor.dropRow+":"+dragRowParent);

        if (TM.v.Editor.dragRow == TM.v.Editor.dropRow ||
            dragRowParent == TM.v.Editor.dropRow) {
          // Cancel and restore selection.
          TM.v.Editor.dropRow = null;
          return this.onHeaderEditorCommand(null, "exit");
        }
        else {
          // Do it.
          return TM.Editor.getReferencesForThreading(TM.v.Editor.dropRow);
        }
      }

      TM.v.keyDragDropMode = !aKeyDragDropMode;
      return true;
    },

    onThreadTreeBlur: function() {
      if (TM.v.editingMode && TM.v.keyDragDropMode) {
TM.u.log.debug("TM.Editor.onThreadTreeBlur: START - "+ TM.v.Editor.dragRow);
        TM.Editor.onHeaderEditorCommand(null, "exit");
      }
    }

  }, // Editor


  //****************************************************************************//
  // Message Browser functions.

  Browser: {
    onLoad: function(aWho) {
TM.u.log.debug("TM.Browser.onLoad: START " + (TM.u.messagepane.currentURI ? TM.u.messagepane.currentURI.spec : null));

      if (gDBView && gDBView.selection && gDBView.selection.count == 1) {
        TM.u.messagepanebox.setAttribute("singleselection", true);
        if (TM.v.toggleRemoteContentAllow) {
          // Temporarily allowed remote content, now reset.
          TM.v.toggleRemoteContentAllow = false;
          TM.v.msgHdr.setUint32Property("remoteContentPolicy", kNoRemoteContentPolicy);
        }
      }
      else {
        TM.u.messagepanebox.removeAttribute("singleselection");
      }

      // Room for page curl indicator.
      if (TM.u.messagepaneContentDocument && TM.u.messagepaneContentDocument.body)
        TM.u.messagepaneContentDocument.body.style.marginTop =
          TM.Headers.headerIndex ? "8px" : "24px";
    },

    Controller: {
      supportsCommand: function (aCommand) {
        switch (aCommand) {
          case "Browser:Back_TM":
          case "Browser:Forward_TM":
            return true;
          default:
            return false;
        }
      },
      isCommandEnabled: function(aCommand) {
        let tabmail = TM.u.g("tabmail");
        let browser = tabmail ? tabmail.getBrowserForSelectedTab() : null;
        if (!browser || !browser.sessionHistory)
          return false;
        switch (aCommand) {
          case "Browser:Forward_TM":
            return browser.canGoForward;
          case "Browser:Back_TM":
            return browser.canGoBack;
          default:
            return false;
        }
      },
      doCommand: function(aCommand) {
        let browser = TM.u.g("tabmail").getBrowserForSelectedTab();
        switch (aCommand) {
          case "Browser:Forward_TM":
            browser.goForward();
            break;
          case "Browser:Back_TM":
            browser.goBack();
            break;
        }
      },
      onEvent: function(aEvent) { }
    }
  }, // Browser


  //**************************************************************************//
  // Attachments functions.

  Attachments: {
    displayAttachmentsForExpandedViewExtra: function() {
      if (!currentAttachments || !currentAttachments.length)
        return;
TM.u.log.debug("TM.Attachments.displayAttachmentsForExpandedViewExtra: START ");

      let attachmentName = TM.u.d("attachmentName");
      let linkmenuitem = TM.u.d("context-attachmentLink");
      let linksep = TM.u.d("context-attachmentLink-sep");
      let isExternalAttachment = currentAttachments[0].isExternalAttachment;
      let isFeedMsg = TM.v.msgHdr.flags & Ci.nsMsgMessageFlags.FeedMsg;

      // Attachment bar single.
      attachmentName.textContent = !isFeedMsg ? "" : currentAttachments[0].url;
      let tooltiptext = isExternalAttachment || currentAttachments[0].isDeleted ?
          "" : attachmentName.getAttribute("tooltiptextsave");
      attachmentName.setAttribute("tooltiptext", tooltiptext);
      let statustext = isExternalAttachment ? "currentAttachments[0].url" : "null";
      attachmentName.setAttribute(
          "onmouseover", "statusFeedback.showStatusString(" + statustext + ")");
      attachmentName.setAttribute(
          "onmouseout", "statusFeedback.showStatusString(null)");
      attachmentName.setAttribute(
          "onfocus", "statusFeedback.showStatusString(" + statustext + ")");
      attachmentName.setAttribute(
          "onblur", "statusFeedback.showStatusString(null)");
      attachmentName.classList.remove("text-link");
      attachmentName.style.textDecoration = "";
      if (isExternalAttachment) {
        if (currentAttachments[0].url.startsWith("file:") &&
            !currentAttachments[0].hasFile) {
          attachmentName.setAttribute(
            "tooltiptext", attachmentName.getAttribute("tooltiptextdetachednotfound"));
          attachmentName.style.textDecoration = "line-through";
        }
        else {
          attachmentName.classList.add("text-link");
        }
      }
      linkmenuitem.hidden = linksep.hidden = !isFeedMsg;

      let attachmentList = TM.u.d("attachmentList");
      attachmentList.setAttribute("tabindex", "-1");

      // Expanded attachment list.
      for (let attachmentitem of attachmentList.children) {
        if (isFeedMsg)
          attachmentitem.textContent = attachmentitem.attachment.url;
        if (attachmentitem.attachment.isExternalAttachment) {
          attachmentitem.setAttribute(
              "onmouseover", "statusFeedback.showStatusString(this.attachment.url)");
          attachmentitem.setAttribute(
              "onmouseout", "statusFeedback.showStatusString(null)");
          attachmentitem.setAttribute(
              "onfocus", "statusFeedback.showStatusString(this.attachment.url)");
          attachmentitem.setAttribute(
              "onblur", "statusFeedback.showStatusString(null)");
          if (attachmentitem.attachment.url.startsWith("file:") &&
              !attachmentitem.attachment.hasFile) {
            attachmentitem.setAttribute(
              "tooltiptext", attachmentName.getAttribute("tooltiptextdetachednotfound"));
            attachmentitem.style.textDecoration = "line-through";
          }
          else {
            attachmentitem.setAttribute("tooltiptext", "");
            let name = attachmentitem.boxObject.firstChild
                                     .getElementsByClassName("attachmentcell-name")[0];
            name.classList.add("text-link");
          }
        }
      }
    },

    /**
     * Link attachments are passed as an array of AttachmentInfo objects. This
     * is meant to download feed enclosure link content, not email.
     *
     * @param array of {AttachmentInfo}.
     */
    saveLinkAttachmentsToFile: function(aAttachmentInfoArray) {
      let URL, Document, DefaultFileName, ContentDisposition,
          ContentType, ShouldBypassCache, FilePickerTitleKey,
          ChosenData, Referrer, InitiatingDocument, SkipPrompt,
          CacheKey;
      // Required for privacy context.
      InitiatingDocument = document;

      for (let attachment of aAttachmentInfoArray) {
        URL = attachment.url;
        DefaultFileName = attachment.name;

TM.u.log.debug("TM.Attachments.saveLinkAttachmentsToFile: filename:url - "+
               DefaultFileName+" : "+URL);

        internalSave(URL, Document, DefaultFileName, ContentDisposition,
                     ContentType, ShouldBypassCache, FilePickerTitleKey,
                     ChosenData, Referrer, InitiatingDocument, SkipPrompt,
                     CacheKey);
      }
    }

  }, // Attachments


  //**************************************************************************//
  // Compose functions.  Add custom headers to addressing widget menulist.

  Compose: {
    onLoad: function() {
TM.u.log.debug("TM.Compose.onLoad: START ");
      // The inline script runs once, just the first time the compose window
      // is opened.
      TM.u.d("msgcomposeWindow").addEventListener("compose-window-init",
                                                  TM.Compose.onInit, false);
      TM.u.d("msgcomposeWindow").addEventListener("compose-window-close",
                                                  TM.Compose.onClose, false);

      // Make headers not crop.
      let col1 = TM.u.d("typecol-addressingWidget");
      col1.style.minWidth = col1.style.width;
      col1.style.width = "";

      let addressColMenu = TM.u.d("addressCol1#1");
      let addressColPopup = addressColMenu.firstChild;
      let addressColInput = TM.u.d("addressCol2#1");
      // Since these are cloned, adding a command listener for each would be
      // more complex; just patch the oncommand attribute.
      TM.Compose._onAddressColCommand = onAddressColCommand;
      addressColMenu.setAttribute("onclick",     "TM.u.openPopup(event);");
      addressColMenu.setAttribute("onkeydown",   "TM.List.checkKeyPress(event);");
      addressColMenu.setAttribute("onmouseover", "TM.u.openPopup(event);");
      addressColMenu.setAttribute("onmouseout",  "TM.u.leaveAnchor(event);");
      addressColMenu.setAttribute("oncommand",   "TM.Compose.onAddressColCommand(this);");

      addressColPopup.setAttribute("onmouseover",    "TM.u.enterPopup(event);");
      addressColPopup.setAttribute("onmouseleave",   "TM.u.leavePopup(event);");
      addressColPopup.setAttribute("onpopuphiding",  "TM.u.popupHiding(event);");
      addressColPopup.setAttribute("onpopupshowing", "TM.Compose.onAddressColPopupShowing(event);");

      if (getPref("mail.autoComplete.highlightNonMatches"))
        addressColInput.setAttribute("highlightnonmatches", true);

      let sep = document.createElementNS(TM.v.XUL_NS, "menuseparator");
      sep.setAttribute("id","customHeaderSep");
      sep.setAttribute("value", "addr_other");
      addressColPopup.appendChild(sep);

      TM.Compose._awRecipientKeyPress = awRecipientKeyPress;
      awRecipientKeyPress = TM.Compose.awRecipientKeyPress;

      // Register the style sheet to force styling.
      let ssSvc = Cc["@mozilla.org/content/style-sheet-service;1"]
                    .getService(Ci.nsIStyleSheetService);
      let styleSheetURI = Services.io.newURI(TM.v.composeStyleSheet, null, null);
      if (!ssSvc.sheetRegistered(styleSheetURI, ssSvc.AGENT_SHEET))
        ssSvc.loadAndRegisterSheet(styleSheetURI, ssSvc.AGENT_SHEET);
    },

    onInit: function() {
      gMsgCompose.RegisterStateListener(TM.Compose);
    },

    onClose: function() {
      gMsgCompose.UnregisterStateListener(TM.Compose);
    },

    awRecipientKeyPress: function(event, element) {
      if (event.keyCode != KeyEvent.DOM_VK_RETURN && event.keyCode != KeyEvent.DOM_VK_TAB)
        return;
TM.u.log.debug("TM.Compose.awRecipientKeyPress: START  ");
      if (element.hasAttribute("customheader") &&
          !TM.v.composeHeadersDefault.has(element.getAttribute("header"))) {
        if (event.keyCode == KeyEvent.DOM_VK_TAB)
          awTabFromRecipient(element, event);
        return;
      }

      TM.Compose._awRecipientKeyPress(event, element);
    },

    onAddressColCommand: function(aAddressWidget) {
      let iscustomheader = aAddressWidget.selectedItem.hasAttribute("customheader");
      aAddressWidget.setAttribute("label", aAddressWidget.label.split(/\s+/)[0]);
      let header = aAddressWidget.selectedItem.getAttribute("header");
      let customval = aAddressWidget.selectedItem.getAttribute("customvalue");
      let wascustomheader = aAddressWidget.hasAttribute("wascustomheader");
      // If there is a custom autofill value for a compose header, fill it.
      let row = awGetRowByInputElement(aAddressWidget);
      let inputElem = awGetInputElement(row);

TM.u.log.debug("TM.Compose.onAddressColCommand: START row:header:iscustom:wascustom:custval - "+
               row+":"+header+":"+iscustomheader+":"+wascustomheader+":"+customval);

      inputElem.blur();
      inputElem.setAttribute("header", header);
      inputElem.removeAttribute("nomatch");
      if (wascustomheader)
        inputElem.value = "";
      if (iscustomheader) {
        aAddressWidget.setAttribute("wascustomheader", true);
        inputElem.setAttribute("customheader", true);
      }
      else {
        aAddressWidget.removeAttribute("wascustomheader");
        inputElem.removeAttribute("customheader");
      }

      if (!header || TM.v.addressingHeaders.has(header)) {
        inputElem.removeAttribute("disableautocomplete");
        inputElem.setAttribute("highlightnonmatches", true);
      }
      else {
        inputElem.setAttribute("disableautocomplete", true);
        inputElem.removeAttribute("highlightnonmatches");
      }

      if (customval) {
        inputElem.value = customval;
        awReturnHit(inputElem);
        if (TM.v.addressingHeaders.has(header))
          awRecipientKeyPress({keyCode:KeyEvent.DOM_VK_RETURN}, inputElem);
        return;
      }

      inputElem.focus();

      TM.Compose._onAddressColCommand(aAddressWidget.id);
    },

    onAddressColPopupShowing: function(aEvent) {
TM.u.log.debug("TM.onPopupShowing: START  "+aEvent.target.tagName);
      let addressColPopup = aEvent.target;
      let row = awGetRowByInputElement(addressColPopup.parentNode);

      let oldNodes = addressColPopup.querySelectorAll("[customheader]");
      for (let oldNode of oldNodes)
        // Clear out old headers so list is dynamic on changes without restart.
        oldNode.remove();

      let headerMap = TM.u.getHeaderMap(TM.v.HEADER.MSG_TYPE_MAIL);
      let headerMapCompose = headerMap[TM.v.HEADER[TM.v.HEADER.COMPOSE]];
      let composeHeaders = {};
      let sepL = "", sepR = "", custL = "", custR = "";
      TM.u.isRtl ? sepL = ":" : sepR = ":";
      let displayName, value, label; //, iscustomheader, isfirstcustomheader;

      for (let [, headerEntry] in Iterator(headerMapCompose)) {
        for (let [headerName, headerValue] in Iterator(headerEntry))
          composeHeaders[headerName] = headerValue;
      }

      for (let node of addressColPopup.childNodes) {
        let headerNameMenuVal = node.getAttribute("value").split("addr_")[1];
        headerNameMenuVal = headerNameMenuVal == "reply" ||
                            headerNameMenuVal == "followup" ? headerNameMenuVal + "-to" :
                                                              headerNameMenuVal;
TM.u.log.debug("TM.messengerCompose: headerNameMenuVal - "+headerNameMenuVal);
        if (TM.v.composeHeadersDefault.has(headerNameMenuVal)) {
          if (headerNameMenuVal in composeHeaders)
            node.removeAttribute("hidden");
          else
            node.setAttribute("hidden", true);
        }
      }

      let composeHeadersCount = 0;
      for (let [headerName, headerValue] in Iterator(composeHeaders)) {
        if ((TM.v.composeHeadersDefault.has(headerName) && !headerValue[TM.v.kFormat]) ||
            (row == 1 && !TM.v.composeHeadersDefault.has(headerName)))
          // Skip default headers if no custom value. Skip non addressing headers
          // if first row.
          continue

        if (TM.v.composeHeadersDefault.has(headerName)) {
TM.u.log.debug("TM.messengerCompose: defaultHeader - "+headerName+":"+headerValue);
          if (!headerValue[TM.v.kFormat])
            // Skip default headers if no custom value.
            continue

          let labelNode = TM.u.d("tm_" + headerName);
          value = labelNode.value;
          label = labelNode.getAttribute("label");
//          iscustomheader = false;
          }
        else {
TM.u.log.debug("TM.messengerCompose: customHeader - "+headerName+":"+headerValue);
          value = "addr_other";
          label = null;
//          iscustomheader = true;
        }

        displayName = sepL + (headerValue[TM.v.kDisplayName] || headerName) + sepR;
        displayName = headerValue[TM.v.kDisplayName] ? displayName :
                      label                          ? label :
                                                       displayName;
TM.u.log.debug("TM.messengerCompose: displayName - "+displayName);

        // Allow spaces; don't trim. ie /,\s+/
        let customvalues = headerValue[TM.v.kFormat].replace(/\\,/g ,"\x01");
        customvalues = customvalues.split(",");
        for (let customvalue of customvalues) {
          if (!customvalue && TM.v.composeHeadersDefault.has(headerName))
            continue

          customvalue = customvalue.replace(/\x01/g, ",");
          TM.u.isRtl ? custL = customvalue + " " : custR = " " + customvalue;
          let menuItem = document.createElementNS(TM.v.XUL_NS, "menuitem");
          menuItem.setAttribute("value", value);
          menuItem.setAttribute("header", headerName);
          menuItem.setAttribute("headerLabel", headerValue[TM.v.kDisplayName] || headerName);
          menuItem.setAttribute("label", custL + displayName + custR);
          menuItem.setAttribute("customheader", true);
          menuItem.setAttribute("customvalue", customvalue);
          addressColPopup.appendChild(menuItem);
          composeHeadersCount++;
        }
      }

      TM.u.d("customHeaderSep").hidden = composeHeadersCount ? false : true;
    },

    setRows: function() {
      setTimeout(function() {
        awFitDummyRows();
        let rows = TM.u.d("addressingWidget").childElementCount - 1;
        Services.prefs.setIntPref("mail.compose.addresswidget.numRowsShownDefault", rows);
        awInitializeNumberOfRowsShown();
TM.u.log.debug("TM.Compose.setRows: rows - "+rows);
      }, 100);
    },

    setForwardHeaders: function(aEvent) {
TM.u.log.debug("TM.Compose.setForwardHeaders: START html? - "+gMsgCompose.composeHTML);
     function callback() {
       let tableBody = TM.createHeaderDraft(TM.v.Editor.fileHeader,
                                            TM.u.MessageType(TM.v.msgHdr),
                                            TM.v.HEADER[TM.v.HEADER.FORWARD],
                                            gMsgCompose.composeHTML);
       let doc = TM.u.d("content-frame").contentDocument;
       if (gMsgCompose.composeHTML) {
         let table = doc.querySelector("table.moz-email-headers-table");
         if (table) {
           if (tableBody) {
             table.firstChild.remove();
             table.appendChild(tableBody);
             table.style.fontSize = "inherit";
             table.classList.add("forward");
           }
           else {
             table.nextElementSibling.remove();
             table.remove();
           }
         }
       }
       else {
         let content = doc.body.innerHTML;
         let headers = content.split("----<br>")[1].split("<br><br>")[0];
         content = content.replace(headers + "<br>", tableBody);
         doc.body.innerHTML = "";
         doc.body.insertAdjacentHTML("beforeend", content);
       }

       TM.v.Editor = {};
     }

     TM.v.msgHdr = gMessenger.msgHdrFromURI(gMsgCompose.originalMsgURI);
     TM.Editor.getMessageStream(TM.v.msgHdr, callback);
    },

    // nsIMsgComposeStateListener
    ComposeProcessDone: function() {},
    SaveInFolderDone: function() {},
    NotifyComposeFieldsReady: function() {},
    NotifyComposeBodyReady: function() {
      if (gComposeType != Ci.nsIMsgCompType.ForwardInline)
        return;

      if (Services.prefs.getIntPref("mail.show_headers") == 1)
        // Forward - set custom headers if show_headers pref is Normal.
        TM.Compose.setForwardHeaders();
      else {
        // Show the default All Headers table.
        let table = TM.u.d("content-frame").contentDocument
                        .querySelector("table.moz-email-headers-table");
        (table && table.classList.add("forward"));
      }
    }

  }, // Compose


  //**************************************************************************//
  // Print functions.  Add custom headers to print/print preview.

  Print: {
    ReplaceWithSelection: function() {
TM.u.log.debug("TM.Print.ReplaceWithSelection: START ");
      Cu.import("resource://gre/modules/XPCOMUtils.jsm");

      let body = window.content.document.querySelector( "body" );
      // There should be only one content containing <div> child of <body>.
      let div = body ? window.content.document.querySelector( "body > div" ) : null;
      if (!body || !div)
        return;

      let selection; // = printOpener.content.getSelection();
      if (printOpener.document && printOpener.document.commandDispatcher &&
          printOpener.document.commandDispatcher.focusedWindow)
        selection = printOpener.document.commandDispatcher.focusedWindow.getSelection();
      if (selection && !selection.isCollapsed) {
        let range = selection.getRangeAt(0);
        let contents = range.cloneContents();
TM.u.log.debug("TM.Print.ReplaceWithSelection: selection contents - "+contents.textContent);
        // Replace the content div with the user's selection.
        div.innerHTML = "";
        div.appendChild(contents);
      }

      // Set custom headers if show_headers pref is Normal.  The custom headers
      // are added only if using Print Preview first.
      if (Services.prefs.getIntPref("mail.show_headers") == 1) {
        let oldHeaders = body.querySelectorAll("body > table");
        for (let oldHeader of oldHeaders) {
          if (oldHeader.classList.contains("header-part1"))
            oldHeader.removeAttribute("width");
          else
            oldHeader.remove();
        }

        TM.Print.setPrintHeaders(body, this.documentURI.spec);
      }
    },

    setPrintHeaders: function(aBody, aMsgURL) {
TM.u.log.debug("TM.Print.setPrintHeaders: START msgURI - "+aMsgURL);
     function callback() {
       let tableBody = TM.createHeaderDraft(TM.v.Editor.fileHeader,
                                            TM.u.MessageType(TM.v.msgHdr),
                                            TM.v.HEADER[TM.v.HEADER.PRINT],
                                            true);
       let table = aBody.querySelector("table.header-part1");
       if (table) {
         if (tableBody) {
           table.firstChild.remove();
           table.appendChild(tableBody);
           table.style.fontSize = "inherit";
           table.classList.add("print");
         }
         else {
           table.nextElementSibling.remove();
           table.remove();
         }
       }

       TM.v.Editor = {};
     }

     // Since aMsgURL is a necko url (mailbox://), convert it to a message header.
     let neckoURL;
     try {
       neckoURL = Services.io.newURI(aMsgURL, null, null);
     }
     catch (e) { }

     if (neckoURL instanceof Ci.nsIMsgMessageUrl) {
       TM.v.msgHdr = neckoURL.messageHeader;
       TM.Editor.getMessageStream(TM.v.msgHdr, callback);
     }
    }

  }, // Print


  //**************************************************************************//
  // Utils.

//  Utils: {
  u: {
    d: function(elementId) {
      return document.getElementById(elementId);
    },

    p: function(elementId) {
      return parent.document.getElementById(elementId);
    },

    c: function(documentId, elementId) {
      return documentId && documentId.contentDocument ?
             documentId.contentDocument.getElementById(elementId) : null;
    },

    g: function(elementId) {
      return this.topWindow.document.getElementById(elementId);
    },

    get converter() {
      return Cc["@mozilla.org/intl/scriptableunicodeconverter"]
               .createInstance(Ci.nsIScriptableUnicodeConverter);
    },

    get log() {
      return Log4Moz.getConfiguredLogger("extensions.TotalMessage");
    },

    get strings() {
      return this._strings ? this._strings :
        this._strings = new StringBundle(TM.v.stringBundleHeaders);
    },

    getString: function(aString) {
      // Why doesn't StringBundle.js handle string not found??
      try { return this.strings.get(aString); } catch (ex) { return null; }
    },

    get loadingMessageStr() {
      // "Loading Message…"
      return this._loadingMessageStr ? this._loadingMessageStr :
        this._loadingMessageStr =
          Services.strings
                  .createBundle("chrome://messenger/locale/messenger.properties")
                  .GetStringFromName("documentLoading");
    },

    get loadingStr() {
      // "Loading…"
      return this._loadingStr ? this._loadingStr :
        this._loadingStr =
          Services.strings
                  .createBundle("chrome://messenger/locale/messenger.properties")
                  .GetStringFromName("loadingTab");
    },

    get ELLIPSIS() {
      if (!this._ellipsis) {
        this._ellipsis = "\u2026";
        try {
          this._ellipsis =
              Services.prefs.getComplexValue("intl.ellipsis",
                                             Ci.nsIPrefLocalizedString).data;
        } catch (e) { }
      }
      return this._ellipsis;
    },

    get AUTOCOLLECT_ACTIVE() {
      return Services.prefs.getBoolPref(TM.v.autoCollectActivePref);
    },

    set AUTOCOLLECT_ACTIVE(val) {
      Services.prefs.setBoolPref(TM.v.autoCollectActivePref, val);
    },

    get HEADER_SEP() {
      return Services.prefs.getComplexValue(TM.v.headerSeparatorPref,
                                            Ci.nsISupportsString).data[0] || "";
    },

    set HEADER_SEP(val) {
      let str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
      str.data = val;
      Services.prefs.setComplexValue(TM.v.headerSeparatorPref,
                                     Ci.nsISupportsString, str);
    },

    get HEADER_WRAP_LINES() {
      return Services.prefs.getIntPref(TM.v.headerWrapLinesPref);
    },

    set HEADER_WRAP_LINES(val) {
      Services.prefs.setIntPref(TM.v.headerWrapLinesPref, val);
    },

    get HEADER_LIST_ENABLED() {
      return Services.prefs.getBoolPref(TM.v.headerListEnabled);
    },

    get javascriptEnabled() {
      return Services.prefs.getBoolPref(TM.v.javascriptEnabled);
    },

    set javascriptEnabled(val) {
      Services.prefs.setBoolPref(TM.v.javascriptEnabled, val);
    },

    get allowJS() {
      return Services.prefs.getBoolPref(TM.v.allowJSPref);
    },

    get mrToggleKey() {
      return this.strings.get("mrToggleKey");
    },

    get viewHeader() {
      return this.g("viewHeader");
    },

    get viewHeadersToolbar() {
      return this.g("viewHeadersToolbar");
    },

    get viewHeaderWrap() {
      return this.g("viewHeaderWrap");
    },

    get TotalMessage() {
      // Look for the object in the main window context.
      return this.topWindow.TotalMessage;
    },

//    get MessageType() {
    MessageType: function(aMsgHdr) {
      let folder = TM.u.topWindow.gFolderDisplay;
      let msgHdr = aMsgHdr || TM.v.msgHdr;
      if (folder && !aMsgHdr) {
        if (folder.selectedCount == 0) {
          if (folder.displayedFolder) {
            TM.v.isImap = folder.displayedFolder.flags & Ci.nsMsgFolderFlags.ImapBox;
            TM.v.isNews = folder.displayedFolder.flags & Ci.nsMsgFolderFlags.Newsgroup;
            TM.v.isFeed = folder.displayedFolder.server.type == "rss";
            TM.v.isTwitter = folder.displayedFolder.server.type == "twitter";
          }
        }
        else {
          TM.v.isImap = folder.selectedMessageIsImap;
          TM.v.isNews = folder.selectedMessageIsNews;
          TM.v.isFeed = folder.selectedMessageIsFeed;
          TM.v.isTwitter = folder.displayedFolder &&
                           folder.displayedFolder.server.type == "twitter";
        }
      }
      else {
        folder = msgHdr.folder;
        TM.v.isImap = folder.flags & Ci.nsMsgFolderFlags.ImapBox;
        TM.v.isNews = folder.flags & Ci.nsMsgFolderFlags.Newsgroup;
        TM.v.isFeed = folder.server.type == "rss" ||
                      msgHdr.flags & Ci.nsMsgMessageFlags.FeedMsg;
        TM.v.isTwitter = folder.server.type == "twitter";
      }
//TM.u.log.trace("TM.Utils.MessageType - "+
//               (TM.v.isImap ? "Imap" : TM.v.isNews ? "News" : TM.v.isFeed ? "Feed" :
//                TM.v.isTwitter ? "Twitter" : "Pop3 or other"));

      return TM.v.isNews ? TM.v.HEADER.MSG_TYPE_NEWS :
             TM.v.isFeed ? TM.v.HEADER.MSG_TYPE_FEED : TM.v.HEADER.MSG_TYPE_MAIL;
    },

    get messagepanebox() {
      // Parent of msgHeaderView in both 3pane and standalone window.
      return this.g("messagepanebox");
    },

    get singlemessage() {
      return this.g("singlemessage");
    },

    get messagepane() {
      // Original messagepane.
      return this.g("messagepane");
    },

    get messagepaneContentDocument() {
      // Original messagepane.
      return this.g("messagepane").contentDocument;
    },

    get messageHeader() {
      return this.g("messageHeaderContainer");
    },

    get messageHeaderContentDocument() {
      return this.messageHeader.contentDocument;
    },

    get messageHeaderBody() {
      return this.c(this.messageHeader, "messageHeaderBody");
    },

    get msgHeaderView() {
      return this.g("msgHeaderView");
    },

    _headerMap: {},
    getHeaderMap: function(aMsgType) {
      // Valid types are 'Mail', 'News', 'Feed'.
      let prefName = "headerMap" + aMsgType + "Pref";
      let str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
      let headerMap, headerMapDefault, defaultMapName;
//TM.u.log.trace("TM.Utils.getHeaderMap: "+prefName);
      try {
        str = Services.prefs.getComplexValue(TM.v[prefName], Ci.nsISupportsString);
        headerMap = JSON.parse(str.data);
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.getHeaderMap: pref reset or error, " +
                      "restoring defaults - " + TM.v[prefName]);
        // Either error or pref has been reset.  Store pref with default.
        defaultMapName = "headerMap" + aMsgType+ "Default";
        headerMap = TM.v[defaultMapName] ? TM.v[defaultMapName] :
                                           TM.v["headerMapMailDefault"];
        str.data = JSON.stringify(headerMap);
        Services.prefs.setComplexValue(TM.v[prefName], Ci.nsISupportsString, str);
        str = Services.prefs.getComplexValue(TM.v[prefName], Ci.nsISupportsString);
        headerMap = JSON.parse(str.data);
      }

      return headerMap;
    },

    setHeaderMap: function(aMsgType, aMap) {
      // Valid types are 'Mail', 'News', 'Feed'.
      let prefName = "headerMap" + aMsgType + "Pref";
      let str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
TM.u.log.debug("TM.Utils.setHeaderMap: "+prefName);

      try {
        str.data = JSON.stringify(aMap);
        Services.prefs.setComplexValue(TM.v[prefName], Ci.nsISupportsString, str);
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.setHeaderMap: error storing pref - " + TM.v[prefName]);
      }
    },

    get commonHeaders() {
      let prefName = "commonHeadersPref";
      let map;
      try {
        map = new Map(JSON.parse(Services.prefs.getCharPref(TM.v[prefName])));
        if (this._commonHeaders)
          return this._commonHeaders;
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.commonHeaders: pref reset or error, " +
                      "restoring defaults - " + TM.v[prefName]);
      }

      if (!map) {
        // Either error or pref has been reset.  Store pref with default.
        map = TM.v["commonHeadersDefault"];
        Services.prefs.setCharPref(TM.v[prefName], JSON.stringify([...map]));
      }

      map.forEach(function(value, name) {
TM.u.log.trace("TM.Utils.commonHeaders: name:value  - "+name+":"+value);
        let element = document.getElementById(value[TM.v.kLabelId]);
        if (element) {
          let str = element.value || element.getAttribute("label") || "";
          value[TM.v.kLabelValue] = str;
        }
TM.u.log.trace("TM.Utils.commonHeaders: name:newVal - "+name+":"+map.get(name));
      })

      return this._commonHeaders = map;
    },

    get collectedHeaders() {
      let prefName = "collectedHeadersPref";
      try {
        return new Set(JSON.parse(Services.prefs.getCharPref(TM.v[prefName])));
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.collectedHeaders: error reading pref - " +
                      TM.v[prefName]);
      }

      return new Set([]);
    },

    set collectedHeaders(aVal) {
      let prefName = "collectedHeadersPref";
      // Store em sorted.
      let sortedHeaders = [...aVal].sort();
      try {
        Services.prefs.setCharPref(TM.v[prefName], JSON.stringify(sortedHeaders));
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.collectedHeaders: error storing pref - " +
                      TM.v[prefName]);
      }
    },

    get customHeaders() {
      let prefName = "customHeadersPref";
      try {
        return new Set(JSON.parse(Services.prefs.getCharPref(TM.v[prefName])));
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.customHeaders: error reading pref - " +
                      TM.v[prefName]);
      }

      return new Set([]);
    },

    set customHeaders(aVal) {
      let prefName = "customHeadersPref";
      try {
        Services.prefs.setCharPref(TM.v[prefName],
                                   JSON.stringify([...aVal]));
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.customHeaders: error storing pref - " +
                      TM.v[prefName]);
      }
    },

    get otherValues() {
      let prefName = "otherValuesPref";
      let map;
      try {
        map = new Map(JSON.parse(Services.prefs.getCharPref(TM.v[prefName])));
        if (this._otherValues)
          return this._otherValues;
      }
      catch (ex) {
        TM.u.log.warn("TM.Utils.otherValues: pref reset or error, " +
                      "restoring defaults - " + TM.v[prefName]);
      }

      if (!map) {
        // Either error or pref has been reset.  Store pref with default.
        map = TM.v["otherValuesDefault"];
        Services.prefs.setCharPref(TM.v[prefName], JSON.stringify([...map]));
      }

      TM.List.columnsCustomMap.forEach(function(value, name) {
        map.set(name.toLowerCase(), [name, name])
      });

      map.forEach(function(value, name) {
TM.u.log.debug("TM.Utils.otherValues: name:value  - "+name+":"+value);
        let element = document.getElementById(value[TM.v.kLabelId]);
        if (element) {
          let str = element.value || element.getAttribute("label") || "";
          value[TM.v.kLabelValue] = str;
        }
TM.u.log.debug("TM.Utils.otherValues: name:newVal - "+name+":"+map.get(name));
      })

      return this._otherValues = map;
    },

    get KEY_ESCAPE() {
      if (this._KEY_ESCAPE)
        return this._KEY_ESCAPE;
      let evt = document.createEvent("KeyboardEvent");
      evt.initKeyEvent("keydown", true, true, null,
                       false, false, false, false,
                       KeyEvent.DOM_VK_ESCAPE, 0);
      return this._KEY_ESCAPE = evt;
    },

    get currentSortIsGlobal() {
      return this.d("sortOrderGlobal").hasAttribute("checked") || false;
    },

    get currentModeIsGlobal() {
      return this.d("mrModeGlobal").hasAttribute("checked") || false;
    },

    get threadTreeChildren() {
      return GetThreadTree().getElementsByTagName("treechildren")[0];
    },

    isStructuredAddressingHeader: function(aHeader, aValue) {
      // Crappy jsmime doesn't handle no addr-spec, so test for it (lamely).
      return TM.v.addressingHeaders.has(aHeader) && aValue.match(/<.+@/) ? true : false;
    },

    /**
     * Add headers to collected list, if not already there.  Store the list as
     * a JSON string in prefs.
     *
     * @param {object} aCurrentHeaderData.
     */
    collectHeaders: function(aCurrentHeaderData) {
TM.u.log.debug("TM.Utils.collectHeaders: START ");
      if (!this.AUTOCOLLECT_ACTIVE) {
TM.u.log.debug("TM.Utils.collectHeaders: autocollect inactive ");
        return;
      }

      let headerNameLC;
      let gotNewHeader = false;
      let currentHeaderData = aCurrentHeaderData;
      let collectedHeaders = this.collectedHeaders;
      for (let [index, headerEntry] in Iterator(currentHeaderData)) {
        headerNameLC = headerEntry.headerName.toLowerCase();
        if (collectedHeaders.has(headerNameLC) ||
            this.commonHeaders.has(headerNameLC) ||
            this.customHeaders.has(headerNameLC))
          continue;

TM.u.log.debug("TM.Utils.collectHeaders: new header to add - "+headerNameLC);
        // Add the new header to the array.
        collectedHeaders.add(headerNameLC);
        gotNewHeader = true;
      }

      if (gotNewHeader)
        // Store the new list.
        this.collectedHeaders = collectedHeaders;
    },

    /**
     * Autopup functions.
     */
    OPENPOPUP_DELAY:  400,
    CLOSEPOPUP_DELAY: 100,
    OPENPOPUP_WHERE: "after_start",

    openPopup: function(aEvent) {
      let anchor = aEvent.target;
      if (anchor.disabled || anchor != aEvent.originalTarget ||
          (("button" in aEvent) && aEvent.button != 0)) {
        return;
      }
      if (aEvent.type == "keypress" && (aEvent.keyCode || aEvent.which)) {
        aEvent.stopImmediatePropagation();
        aEvent.stopPropagation();
        let key = aEvent.keyCode || aEvent.which;
        if (key != KeyEvent.DOM_VK_DOWN)
          return;
      }

      anchor.setAttribute("popupactive", null);
      if (anchor.hasAttribute("open"))
        return;

      let timeout = aEvent.type == "mouseover" ? TM.u.OPENPOPUP_DELAY : 0;
      setTimeout(function() {
        if (anchor.getAttribute("popupactive") == "false")
          return;

        let popup = anchor.popup || anchor.menupopup;
        if (!popup)
          return;

        if (anchor.tagName == "toolbarbutton")
          TM.u.KEY_ESCAPE();
        let where = popup.getAttribute("openwhere") || TM.u.OPENPOPUP_WHERE;
        popup.openPopup(anchor, where);
        anchor.setAttribute("open", true);
//        aEvent.target.focus();
      }, timeout);
    },

    leaveAnchor: function(aEvent) {
      let anchor = aEvent.target;
      let popup = anchor.popup || anchor.menupopup;
      if (!popup)
        return;

      anchor.setAttribute("popupactive", false);
      setTimeout(function() {
        if (anchor.getAttribute("popupactive") == "true" || popup.state != "open")
          return;

        popup.hidePopup();
      }, TM.u.CLOSEPOPUP_DELAY);
    },

    enterPopup: function(aEvent) {
      let popup = aEvent.target;
      let anchor, anchorPopup;
      while (popup && !anchor) {
        anchor = popup.anchor || popup.anchorNode;
        if (anchor) {
          anchorPopup = anchor.popup || anchor.menupopup;
          if (anchorPopup && anchorPopup == aEvent.originalTarget)
            break;
        }
        popup = popup.parentElement;
      }
      if (!anchor)
        return;

      anchor.setAttribute("popupactive", true);

      // Don't close embedded popup's parent popup.
      let parentPopup = anchor.parentElement;
      let parentAnchor;
      while (parentPopup && !parentAnchor) {
        parentAnchor = parentPopup.anchor || popup.anchorNode;
        if (!parentAnchor)
          parentPopup = parentPopup.parentElement;
      }
      if (!parentAnchor)
        return;

      parentAnchor.setAttribute("popupactive", true);
    },

    leavePopup: function(aEvent) {
      let popup = aEvent.target;
      if (popup != aEvent.originalTarget)
        return;

      let anchor;
      while (popup && !anchor) {
        anchor = popup.anchor || popup.anchorNode;
        if (!anchor)
          popup = popup.parentElement;
      }
      if (!anchor)
        return;

      anchor.setAttribute("popupactive", false);
      setTimeout(function() {
        if (anchor.getAttribute("popupactive") != "false")
          return;

        popup.hidePopup();
      }, TM.u.CLOSEPOPUP_DELAY);
    },

    keypress: function(aEvent) {
      if (aEvent.keyCode != KeyEvent.DOM_VK_TAB && !aEvent.altKey &&
          aEvent.keyCode != KeyEvent.DOM_VK_DOWN)
      aEvent.preventDefault();
    },

    popupHiding: function(aEvent) {
      let popup = aEvent.target;
      let anchor = popup.anchor || popup.anchorNode;
      if (!anchor)
        return;

      setTimeout(function() {
//TM.u.log.debug("TM.Utils.popupHiding: anchor.id - "+anchor.id);
        anchor.removeAttribute("open");
        anchor.removeAttribute("popupactive");
      }, 0);
    },

    KEY_ESCAPE: function() {
      let evt = document.createEvent("KeyboardEvent");
      evt.initKeyEvent("keydown", true, true, null,
                       false, false, false, false,
                       KeyEvent.DOM_VK_ESCAPE, 0);
      document.dispatchEvent(evt);
    },

    /**
     * Set up existing toolbox from saved customizations on our header toolbox.
     */
    initializeToolbars: function() {
TM.u.log.debug("TM.Utils.initializeToolbars: START ");
      let mode, iconsize, labelalign;
      let winDoc       = TM.u.topWindow.document.documentElement;
      let toolboxTB    = TM.u.g("header-view-toolbox");
      let toolbarTB    = TM.u.g("header-view-toolbar");
      let menupopupsTM = TM.u.g("menupopupsTM");
      let buttonIdsTB  = [];

      // Get mode and icon size and labelalign per toolbar customization.
      let getToolbarModes = function() {
        mode       = winDoc.getAttribute(toolboxTB.id+"_mode") ||
                     toolboxTB.getAttribute("mode");
        iconsize   = winDoc.getAttribute(toolboxTB.id+"_iconsize") ||
                     toolboxTB.getAttribute("iconsize");
        labelalign = winDoc.getAttribute(toolboxTB.id+"_labelalign") ||
                     toolboxTB.getAttribute("labelalign");
TM.u.log.trace("TM.Utils.initializeToolbars: mode:iconsize:labelalign "+
               mode+":"+iconsize+":"+labelalign);
      };
      getToolbarModes();

      // Headers toolbars.
      toolbarTB.setAttribute("mode", mode);
      toolbarTB.setAttribute("iconsize", iconsize);
      toolboxTB.setAttribute("labelalign", labelalign);
      // Restore currentset.  Uses defaultset until first customization change.
      let savedcurrentset = winDoc.getAttribute(toolbarTB.id+"_currentset");
      toolbarTB.currentSet = savedcurrentset || toolbarTB.getAttribute("defaultset");

      // Get the toolbarbutton elements on the toolbar.
      let elementsTB = toolbarTB.querySelectorAll("toolbarbutton");
      elementsTB = [].map.call(elementsTB, e => e);
      // Get the toolbarbuttons not customized, but in the toolbar's palette.
      for (let button of toolboxTB.palette.childNodes) {
        if (button.localName != "toolbarbutton")
          button = button.querySelector("toolbarbutton");
        if (button)
          elementsTB.push(button);
      }
      for (let button of elementsTB)
        // Make sure all buttons have this class.
        button.classList.add("msgHeaderView-button");

      // Get the menupopup elements on the toolbar.
      elementsTB = toolbarTB.querySelectorAll("menupopup");
      elementsTB = [].map.call(elementsTB, e => e);
      // Get the menupopups of buttons not customized, but in the toolbar's palette.
      for (let button of toolboxTB.palette.childNodes) {
        if (button.localName != "toolbarbutton")
          button = button.querySelector("toolbarbutton");
        if (button) {
          let menupopup = button.querySelector("menupopup");
          if (menupopup)
            elementsTB.push(menupopup);
        }
      }

      // Initialize menupopupTM only once; gather all menupopups there.
      for (let menupopup of elementsTB) {
        let button = menupopup.parentElement;
        let buttonTM = button && button.id == "menupopupsTM";
        if (!button || (button.localName != "toolbarbutton" && !buttonTM))
          continue;

        menupopup.addEventListener("mouseover",   TM.u.enterPopup, false);
        menupopup.addEventListener("mouseleave",  TM.u.leavePopup, false);
        menupopup.addEventListener("popuphiding", TM.u.popupHiding, true);

        if (!buttonTM) {
          menupopupsTM.appendChild(menupopup);
          menupopup.setAttribute("anchorid", button.id);
          button.setAttribute("type", "menu");
//          button.removeAttribute("command");
          button.removeAttribute("tooltiptext");
        }
TM.u.log.debug("TM.Utils.initializeToolbars: menupopups button id - "+button.id);
      }

      let headerToolbarBoxRight = TM.u.c(TM.u.messageHeader, "headerToolbarBoxRight");
      if (!headerToolbarBoxRight)
        // Not ready; this happens in a non classic view layout, which causes a
        // DOM move and unload of our page.  The subsequent load will run this.
        return;
TM.u.log.debug("TM.Utils.initializeToolbars: toolbar.id:currentset "+
               toolbarTB.id+" : "+toolbarTB.getAttribute("currentset"));

      // Clone existing toolbox, if not there; depends on layout type and change.
      let toolbox = TM.u.c(TM.u.messageHeader, "header-view-toolbox");
      if (!toolbox) {
        headerToolbarBoxRight.appendChild(toolboxTB.cloneNode(true));
        toolbox = TM.u.c(TM.u.messageHeader, "header-view-toolbox");
        toolbox.palette = toolboxTB.palette;
        toolbox.hidden = !(TM.u.viewHeadersToolbar.getAttribute("checked") == "true");
        toolbox.addEventListener("customizationchange", TM.Headers.setToolbarState);
        toolbox.addEventListener("contextmenu", TM.Headers.onToolbarContext);
      }

      // Intialize toolbar view and buttons.  Need to reset on layout change etc.
      let setButtonState = function() {
TM.u.log.debug("TM.Utils.initializeToolbars: setButtonState ");
        for (let menupopup of menupopupsTM.childNodes) {
          let buttonId = menupopup.getAttribute("anchorid");
          let button = TM.u.c(TM.u.messageHeader, buttonId);
          if (!button)
            continue;
TM.u.log.trace("TM.Utils.initializeToolbars: setButtonState, id - "+button.id);
          menupopup.anchor = button;
          button.popup = menupopup;
          button.addEventListener("click", TM.u.openPopup, true);
          button.addEventListener("mouseover", TM.u.openPopup, true);
          button.addEventListener("mouseout",  TM.u.leaveAnchor, true);
        }
        // Set global wrap.
        let hdrWrapChecked = TM.u.viewHeaderWrap.getAttribute("checked") == "true";
        let hdrWrapButton = TM.u.c(TM.u.messageHeader, "hdrWrapButton");
        if (hdrWrapButton)
          hdrWrapButton.checked = hdrWrapChecked;
        if (hdrWrapChecked)
          TM.u.messageHeaderBody.classList.add("wrap");
        else
          TM.u.messageHeaderBody.classList.remove("wrap");
      }

      // Intialize buttons in threadpane bar.
      let setButtonStateList = function() {
TM.u.log.debug("TM.Utils.initializeToolbars: setButtonStateList ");
        let menupopupsListTM = TM.u.d("menupopupsListTM");
        for (let menupopup of menupopupsListTM.childNodes) {
          let buttonId = menupopup.getAttribute("anchorid");
          let button = buttonId ? TM.u.d(buttonId) : null;
          if (!button)
            continue;

TM.u.log.debug("TM.Utils.initializeToolbars: setButtonStateList, id - "+button.id);
          button.popup = menupopup;
          button.addEventListener("click",     TM.u.openPopup, false);
          button.addEventListener("keydown",   TM.List.checkKeyPress, true);
          button.addEventListener("mouseover", TM.u.openPopup, true);
          button.addEventListener("mouseout",  TM.u.leaveAnchor, false);
          menupopup.anchor = button;
          menupopup.addEventListener("mouseover",   TM.u.enterPopup, false);
          menupopup.addEventListener("mouseleave",  TM.u.leavePopup, false);
          menupopup.addEventListener("popuphiding", TM.u.popupHiding, false);
        }
      }

      // Initialize buttons on new load.
      setButtonState();
      this.setWrapLinesAndBorder();
      TM.Headers.setToolbarState();
      setButtonStateList();

      // Customize functions.
      toolbox.customizeDone = function(aEvent) {
        let topWin = TM.u.topWindow;
        topWin.MailToolboxCustomizeDone(aEvent, "CustomizeHeaderToolbar");
        toolbox.removeAttribute("doCustomization");
        setButtonState();
        TM.Headers.setToolbarState();
        getToolbarModes();
      };

      toolbox.ownerDocument.persist = function(aId, aWhat) {
        let topWinDoc = TM.u.topWindow.document;
        let value = toolbox.ownerDocument.getElementById(aId).getAttribute(aWhat);
        topWinDoc.documentElement.setAttribute(aId+"_"+aWhat, value);
        topWinDoc.persist(topWinDoc.documentElement.id, aId+"_"+aWhat);
        if (aWhat == "currentset") {
          // Clone to Tb toolbar to reflect current state (as command updater
          // functions run on those buttons and we use that for our own toolbar)
          // and persist the Tb toolbar currentset so the constructor builds it
          // on startup.
          toolboxTB = toolbarTB.parentNode;
          toolbarTB.remove();
          toolbarTB = TM.u.c(TM.u.messageHeader, "header-view-toolbar").cloneNode(true);
          toolbarTB.removeAttribute("customizing");
          toolboxTB.insertBefore(toolbarTB, toolboxTB.lastChild);
          topWinDoc.persist(toolbarTB.id, aWhat);
        }
      };

      // Register the style sheet to force scrollbar styling.
      let ssSvc = Cc["@mozilla.org/content/style-sheet-service;1"]
                    .getService(Ci.nsIStyleSheetService);
      let styleSheetURI = Services.io.newURI(TM.v.scrollbarStyleSheet, null, null);
      if (!ssSvc.sheetRegistered(styleSheetURI, ssSvc.AGENT_SHEET))
        ssSvc.loadAndRegisterSheet(styleSheetURI, ssSvc.AGENT_SHEET);
    },

    setWrapLinesAndBorder: function () {
      // Add a style rule for max container height based on headerWrapLines pref.
      let headerIndex = TM.Headers.headerIndex;
      let headerBody = TM.u.messageHeaderBody;
      if (!headerBody)
        return;
      // Get the document line height.
      let cStyle = TM.u.topWindow.getComputedStyle(headerBody, null)
      let dataLnHt = parseInt(cStyle.getPropertyValue("line-height").replace(/px/, ""));
      // The '4' is added for margin and padding, to prevent scrollbar despite
      // desired wraplines.  The vertical margin/padding is added to the
      // .container and .multiItemContainer css rules.
      let maxHeight = TM.u.HEADER_WRAP_LINES * (dataLnHt + 4) + "px";
      let rule = ".wrapLinesHt { max-height: " + maxHeight + "; }";
      let styleSheets = TM.u.messageHeaderContentDocument.styleSheets;
      for (let styleSheet of styleSheets) {
        if (styleSheet.href == TM.v.globalStyleSheet) {
          styleSheet.insertRule(rule, styleSheet.cssRules.length);
TM.u.log.trace("TM.Utils.setWrapLinesAndBorder: css rule inserted - "+ rule);
        }
      }
    },

    /**
     * Persist user selected height/width on header pane.
     */
    onMouseUpSplitter: function () {
      let headerIndex = TM.Headers.headerIndex;
      let height = TM.u.msgHeaderView.height;
      TM.u.viewHeader.setAttribute(TM.v.HEADER[headerIndex], height);
      TM.u.viewHeader.height = TM.u.viewHeader.getAttribute(TM.v.HEADER[headerIndex]);
TM.u.log.debug("TM.Utils.onMouseUpSplitter: set header height - "+height);

      // Resizing makes something set the height; selecting the next message
      // doesn't remove it.
      setTimeout(function() {
        TM.u.g("msgNotificationBar").removeAttribute("height");
      }, 0);
    },

    /**
     * Safely sets the URI attribute (f.e. "href") on a tag (f.e. the HTML <a>
     * tag), providing the URI specified can be loaded according to the rules.
     *
     * In particular, this prevents sources from linking to javascript: and data:
     * URLs that would run with chrome privileges in our various chrome-privileged
     * views.  It also prevents sources from linking to chrome: URLs.
     *
     * Based on the similar method in FeedWriter.js.
     *
     * @param   aElement     {Element}
     *          the element on which to set the attribute
     *
     * @param   aAttribute   {String}
     *          the name of the attribute to set, f.e. href or src
     *
     * @param   aUriStr      {String}
     *          the URI to which to set the attribute, derive an nsIPrincipal
     *          associated with the URI
     */
    safelySetURIAttribute: function(aElement, aAttribute, aUriStr) {
TM.u.log.debug("TM.Utils.safelySetURIAttribute: aUrl - "+aUriStr);
      let uri, principal;
      let flags = Ci.nsIScriptSecurityManager.DISALLOW_INHERIT_PRINCIPAL;
      try { uri = Services.io.newURI(aUriStr, null, null); } catch (ex) { }
      if (!uri)
        return;

      principal = Services.scriptSecurityManager.getSimpleCodebasePrincipal(uri);
      try {
        // This will throw if the URI shouldn't be loaded, either because the
        // source isn't allowed to load it or per the rules specified in |flags|.
        Services.scriptSecurityManager.checkLoadURIStrWithPrincipal(principal,
                                                                    aUriStr,
                                                                    flags);
      }
      catch(ex) {
        // Threw, so we don't set the attribute.
        return;
      }
  
      aElement.setAttribute(aAttribute, aUriStr);
    },

    get isRtl() {
      return Services.prefs.getIntPref("bidi.direction") == 2;
    },

    get winType() {
      return this.topWindow.document.documentElement.getAttribute("windowtype");
    },

    get mainWindow() {
      return Services.wm.getMostRecentWindow("mail:3pane");
    },

    get messageWindow() {
      return Services.wm.getMostRecentWindow("mail:messageWindow");
    },

    topWindow: window.QueryInterface(Ci.nsIInterfaceRequestor).
                      getInterface(Ci.nsIWebNavigation).
                      QueryInterface(Ci.nsIDocShellTreeItem).
                      rootTreeItem.
                      QueryInterface(Ci.nsIInterfaceRequestor).
                      getInterface(Ci.nsIDOMWindow)
  }, // Utils

//  get v() { return TM.Vars  },
//  get u() { return TM.Utils }

}; // TotalMessage

// TM = TotalMessage.Utils.TotalMessage;
var TM = TotalMessage.u.TotalMessage;
window.addEventListener("load", TM.onLoad, true);
